proc debugPuts { a } {
#     puts "TCL debug: $a"
}
debugPuts "Starting browser.tcl"
#######################################################################
# Sather class browser
# Author: Benedict Gomes <gomes@icsi.berkeley.edu>
# David Stoutamire (many ideas, sather code for graph layout, 
# comment highlighting, compiler support etc.), 
# Matt Kennel (window layout, intelligent color consistency, event flushing, 
# cursor changes, bugs),
# Yon Uriarte (changes to graph layout, portability)
# 
# Submitting changes: Please contact the author when making changes
# to avoid duplication of effort.
# The recommended way to make changes is to redefine procedures in your
# .bs.tcl and submit that file, which will act like a diff of sorts...
#
# Log History at end of file
#######################################################################
# Various tags are documented as follows
# Arrow tags:      [See proc connect]
#                  ${srcName} 
#                  ${destName} 
#                  arrow${srcName} 
#                  arrow${destName} 
#                  arrow  
#                  [one of the following tags:
#                  subtype moduleContains include
#                  ]
#           Note: redrawArrow uses the first 2 tags to redraw the arrow.
#           moveArrowsTouching uses the second 2 tags to determine arrows
#           touching a node
# Node Tags:      [see proc makeTextNodeAt]
#                  ${fullNodeText} 
#                  node 
#                  node${fullNodeText} 
#                  ${nodeText} 
#                  [one of the following tags:
#                  abstract 
#                  concrete 
#                  module
#                  ]
# Tags for the text widget: 
# comment          Indicates a commented region (only if highlighting is
#                   on anf file is small enough)
# Tags for the class list widget: [see initClassList]
#######################################################################
# NOTES ON TKKIT:
# The value that comes from a callback is a list of return values.
# Often you will have to extract the 0 the element [lindex ret 0]
# to get the value you want (i.e. the whole tcl list is just the first
# element of the array that is returned)
#######################################################################
#######################################################################
#                             CALLBACKS 
# setCurClassName     Called in proc selectClass - sets the sather side to the
#                     currect class (selected from the list or from the graph)
# getHome             Returns the SATHER_HOME directory. Called in showHelp
# getFullLayout       Returns the layout for all active nodes
#                     (not just ancestors/descendants of the currently 
#                     active node)
#                     Currently used in the undocumented menu option (
#                     dumpFullGraphForDot as an example of what you might do)
# getRestrictedLayout  Called in initGraph. Returns the layout of the graph
#                      consisting of currently active nodes that are also
#                      ancestors/descendants of the current node and are
#                      visible. (i.e. including whichever of modules, concrete
#                      and abstract types are visible) 
#                      { 
#                        { layout.depth layout.width}
#                        { {clasname1 xpos ypos } {classname2 xpos ypos} ... }
#                        { { source1 dest1} { source2 dest2} ....} 
#                     }
# toggleBoolVar       Used to toggle the boolean variables that are held by
#                     both the sather and tcl ends
# activateClasses     Make some classes visible
# deactivateClasses   Make some classes invisible - see toggleNoShow
#                     These two callbacks are used to hide C_ and TEST_
#                     classes... The original version of the browser used
#                     them together with getFullLayout to present all 
#                     interesting parts of the class graph at the same time.
# execName            Unused! Should return the name of the executable...
# allClasses          Used in initClassList. Takes the form of 
#                     {
#                     { classOrModuleName  filename location ContainingModule}
#                     { ... }
#                     }
# getClassInfo       Used in showClassDef (and in updateFeatureList)
#                    and showClassDocs
#                    { className  { {<routname> <file> <loc> <routineType>}
#                                   {<routname> <file> <loc> <routineType>}
#                                 }   }
#  where routineType indicates the type of the routine
# routine. The string must be treated as a seven element array
# Char 0 indicates whether this routine is an iter.
#         if it is, then character 0 is 'i' otherwise it is 'n'
# Likewise, 
# Character  property      if true  if false
# char 0, iter             , 'i'   'n'
# char 1, private          , 'p'   'n'
# char 2, attribute writer,  'w'   'n'
# char 3, attribute reader,  'r'   'n'
# char 4, shared writer,     'w'   'n'
# char 5  shared reader      'r'   'n'
# char 6  constant           'c'   'n'
#  Used in updateFeatureList
#  routineType is documented in browser.sa under get_class_info_cb
#######################################################################
#                             GLOBALS
#######################################################################
#                     Browser Global Data structures
#                     ------------------------------
# List of all current classes
set gAllClasses {}
# Classes that begin with C_
set gCClasses {} 
# Classes that begin with TEST_
set gTestClasses {}

# gCurClassName   Current class selected
set gCurClassName ""

# List of routs for the currently selected class. It is set in showClassDef 
# by a callback to getClassInfo  i.e. get_class_info_cb in sather
set gCurFeatures ""

# Name of current file - used by saveText
set gCurFile ""

# gModToClasses   Mapping from module names to liss of classes. 
#                 set in initClasses


#                     State of display
#                     ---------------
# In general, these variables have counterparts in browser.sa and the initial
# values of both should be changed to get correct behaviour.
# You can do this by either modifying browser.sa directly, or call 
# tkkit_cb toggleBoolVar variableName (see proc toggleFeature)
# to toggle the sather variable to the right state. 
# You should probably do this in proc configure
# A more elaborate procedure should probably be written
# that will set the browser.sa values to the values shown
# here at start-up.
# After startup, we just keep toggling values in browser.sa
# (toggling was simpler than setting the value directly for some silly reason -
# i believe it was because I could not embed the current value of the
# variable in the string passed from the menu command item)
set gShowConcrete 1
set gShowAbstract 1
set gShowModule 1

set gShowPublic 1
set gShowPrivate 1
set gShowAttr 1
set gShowConst 1
set gShowShared 1

set gHome [tkkit_cb getHome]

#                     Various frames
#                      -------------
# gGraphFrame     Frame for canvas for class graphs. Set in configure
#                gGraphFrame.c is actual canvas. See makeCanvas

set gGraphFrame ""
# gTextDisplay    gTextDisplay.t is the text widget. See makeText
set gTextDisplay ""
# gFeatureList    Scrolling listbox. See makeFeatureListBox
#                gFeatureList.l is actual listbox
set gFeatureList ""
# gInformFrame    Bottom of window - information about actions.
set gInformFrame ""
#gMenuFrame      Frame for the menubuttons. See makeMenuBar
set gMenuFrame ""
# gClassList      List of class names. see makeListBox
set gClassList ""
set gNoShow_C_Classes   1
set gNoShow_Test_Classes  1

#                     Browser.tcl specfic variables
#                     -----------------------------
# These variables are unrelated to anything in browser.sa and can
# be changed with impunity
# Indicates whether the text window or emacs should be used
set gUseEmacs 0

# gGraphUpdate  True if graph should not be updated any further
set gGraphUpdate "1"

# Determines the configuration mode MultiWindow or SingleWindow
set gConfigMode "SingleWindow"
# Determines whether the graph is laid out horizontally or vertically
set gHorizontalGraph false

# Indicates whether the initial graph should be of a fixed size
set gFixedSizeInitGraph true

set gSafeFile  "false"

# Files with more lines will not have comment highlighting
set gCommentThreshold 500

# Whether comments should be highlighted or not
set gHighlightComments 0

# Font used for comments when they are highlighted
set gCommentFont "lucidasans-italic-12"

set gTitleFont "lucidasans-14"
#A very basic problem is the ability to determine whether the display
#is monochrome or b/w. initColor attempts to set it correctly. Not
# sure if it works
set colormodel [tk colormodel .]
if { ${colormodel} == "monochrome"} {
    set gIsColor "false"
} else { 
    set gIsColor "true"
}

debugPuts "Color is: ${gIsColor} ${colormodel}"


# Various colors stuff.  Improved by Matt Kennel, ORNL/UTK.
#
# Colors to control display.
#
if { $gIsColor } {
    set gSubtypeColor "firebrick"
    set gSubtypeRev   "white"

    set gIncludeColor "black"
    set gIncludeRev "white"

    set gModuleColor "darkgreen"
    set gModuleRev "white"

    set gCommentColor "brown"

    set gTextTitleColor "white"
    set gTextTitleRev "black"
    #colors for text file window's title.

    # Color for main window border background
    set gDeepBackgroundColor "grey30"

    # Color of inform text
    set gInformColor "white"

    # Color for contents window background
    set gBackgroundColor "white"

    # Color for Module&class list title
    set gClassListTitleColor "steelblue"
    set gClassListTitleRev "white"

    # Colors for Feature list and a highlighted class in class list
    set gFeatureListTitleColor "red2"
    set gFeatureListTitleRev "white"
    
    # colors for selected features
    set gFeatureListSelectionColor "violet"
    set gFeatureListSelectionRev "black"

    set gTextSelectionColor "violet"
    set gTextSelectionRev "black"
} else {
    # b&w choices
    set gSubtypeColor "black" 
    set gSubtypeRev "white" 
    set gIncludeColor "black"
    set gIncludeRev "white"

    # set gGraphMonoModuleContains "@${gHome}/Browser/dashed.bitmap" 
    # set gGraphMonoInclude "@${gHome}/Browser/dashed-hard.bitmap" 
    # set gGraphMonoModuleContains gray25
    # set gGraphMonoInclude gray50
    # set gGraphMonoSubtype ""

    set gModuleColor  "black"
    set gModuleRev    "white"
    set gCommentColor "black"

    # Color of inform text
    set gInformColor "black"

    set gTextTitleColor "white"
    set gTextTitleRev "black"
    set gDeepBackgroundColor "white"
    set gBackgroundColor "white"
    set gClassListTitleColor "black"
    set gClassListTitleRev "white"

    set gFeatureListTitleColor "black"
    set gFeatureListTitleRev "white"
    
    set gFeatureListSelectionColor "black"
    set gFeatureListSelectionRev "white"

    set gTextSelectionColor "white"
    set gTextSelectionRev "black"
}

# Font used to display module names in the list of classes and
# modules pane.
set gClassListModuleFont "-*-lucida-bold-r-normal-*-16-*-*-*-p-*-*-*"
# "lucidasans-bolditalic-14"
set gClassListClassFont  "-*-lucida-medium-r-normal-*-14-*-*-*-p-*-*-*"

# Global table of fonts for use by the graph viewer
set gFontSizes() ""
# Selected font sized was intended to hold a set of larger fonts
set gBoldFontSizes() ""
set gItalicFontSizes() ""
set gBolditalicFontSizes() ""
proc initFonts { } {
    global gFontSizes
    global gBoldFontSizes
    global gItalicFontSizes
    global gBolditalicFontSizes
    # See the file future.tcl for the old settings of these variables...

    set gFontSizes(14) "lucidasans-18"
    set gFontSizes(13) "lucidasans-18"
    set gFontSizes(12) "lucidasans-18"
    set gFontSizes(11) "lucidasans-18"
    set gFontSizes(10) "lucidasans-14"
    set gFontSizes(9)  "lucidasans-14"
    set gFontSizes(8)  "lucidasans-12"
    set gFontSizes(7)  "lucidasans-12"
    set gFontSizes(6)  "lucidasans-10"
    set gFontSizes(5)  "lucidasans-10"
    set gFontSizes(4)  "lucidasans-8"
    set gFontSizes(3)  "lucidasans-8"
    set gFontSizes(2) "nil2"
    set gFontSizes(1) "nil2"

    set gItalicFontSizes(14) "lucidasans-italic-18"
    set gItalicFontSizes(13) "lucidasans-italic-18"
    set gItalicFontSizes(12) "lucidasans-italic-18"
    set gItalicFontSizes(11) "lucidasans-italic-18"
    set gItalicFontSizes(10) "lucidasans-italic-14"
    set gItalicFontSizes(9)  "lucidasans-italic-14"
    set gItalicFontSizes(8)  "lucidasans-italic-12"
    set gItalicFontSizes(7)  "lucidasans-italic-12"
    set gItalicFontSizes(6)  "lucidasans-italic-10"
    set gItalicFontSizes(5)  "lucidasans-italic-10"
    set gItalicFontSizes(4)  "lucidasans-italic-8"
    set gItalicFontSizes(3)  "lucidasans-italic-8"
    set gItalicFontSizes(2) "nil2"
    set gItalicFontSizes(1) "nil2"

    set gBoldFontSizes(14) "lucidasans-bold-18"
    set gBoldFontSizes(13) "lucidasans-bold-18"
    set gBoldFontSizes(12) "lucidasans-bold-18"
    set gBoldFontSizes(11) "lucidasans-bold-18"
    set gBoldFontSizes(10) "lucidasans-bold-14"
    set gBoldFontSizes(9)  "lucidasans-bold-14"
    set gBoldFontSizes(8)  "lucidasans-bold-12"
    set gBoldFontSizes(7)  "lucidasans-bold-12"
    set gBoldFontSizes(6)  "lucidasans-bold-10"
    set gBoldFontSizes(5)  "lucidasans-bold-10"
    set gBoldFontSizes(4)  "lucidasans-bold-8"
    set gBoldFontSizes(3)  "lucidasans-bold-8"
    set gBoldFontSizes(2) "nil2"
    set gBoldFontSizes(1) "nil2"

    set gBolditalicFontSizes(14) "lucidasans-bolditalic-18"
    set gBolditalicFontSizes(13) "lucidasans-bolditalic-18"
    set gBolditalicFontSizes(12) "lucidasans-bolditalic-18"
    set gBolditalicFontSizes(11) "lucidasans-bolditalic-18"
    set gBolditalicFontSizes(10) "lucidasans-bolditalic-14"
    set gBolditalicFontSizes(9)  "lucidasans-bolditalic-14"
    set gBolditalicFontSizes(8)  "lucidasans-bolditalic-12"
    set gBolditalicFontSizes(7)  "lucidasans-bolditalic-12"
    set gBolditalicFontSizes(6)  "lucidasans-bolditalic-10"
    set gBolditalicFontSizes(5)  "lucidasans-bolditalic-10"
    set gBolditalicFontSizes(4)  "lucidasans-bolditalic-8"
    set gBolditalicFontSizes(3)  "lucidasans-bolditalic-8"
    set gBolditalicFontSizes(2) "nil2"
    set gBolditalicFontSizes(1) "nil2"
}


#
# These are variables used by the browser inter-procedurally 
# You probably don't want to mess with them
# gOldScale 1.0   Current zooming scale
set gOldScale 1.0
# gFontSize 14 	 gFontSize at scale 1.0 and current font size
set gFontSize 14



debugPuts "After global declarations"
########################################################################
#                          Generic Sather routines
########################################################################
proc isAbstract { nodeName } {
    set abs [string first "\$" ${nodeName}]
    # possible to return abs != -1 itself?
    if { $abs != -1 } {
	return true
    } else {
	return false
    }
    
}

# Return true if the name ends with a .module
proc isModule { nodeName } {
    set match [regexp {[A-Za-z_0-9]*.(module|com)} $nodeName ]
    if { $match == 1 } {
	return true
    } else {
	return false
    }
}

proc getModuleName { fullName } {
    debugPuts "In getModuleName $fullName \n" 
    set match [regexp {[A-Za-a_/0-9]*/([A-Za-z_0-9]*).(module|com)} $fullName full modname ]
    if { $match == 1 } {
	return $modname
    } else {
	set match [regexp {([A-Za-z_0-9]*).(module|com)} $fullName full modname ]
	if { $match == 1} {
	    return $modname
	} else {
	    return ""
	    #mbk
	}
    }
}
########################################################################
#                          INFORM WINDOW 
########################################################################
proc makeInform { frameName } {
    global gDeepBackgroundColor
    global gInformColor

    label ${frameName}.msg -foreground ${gInformColor} -background $gDeepBackgroundColor
    pack ${frameName}.msg -fill x -expand yes
}

proc inform { text } {
    #    debugPuts $text
    global gInformFrame
    ${gInformFrame}.msg configure -text ${text}
}
########################################################################

########################################################################
#                          MENU BAR 
########################################################################
#######################################################################
# Make the menu bar for the graph display
# Arguments: frameName - name of the frame in which to create the widget
# Creates a menu bar with the following buttons
#######################################################################
proc makeMenuBar { frameName } {
    global gGraphUpdate
    global gHighlightComments
    global gShowConcrete
    global gHorizontalGraph
    global gNoShow_Test_Classes
    global gNoShow_C_Classes
    global gBackgroundColor
    global gFixedSizeInitGraph
    global gUseEmacs

    debugPuts "Making menu bar...."
    set cm ${frameName}.graph.menu
    menubutton ${frameName}.graph -text "Graph" -menu ${cm} \
	    -bg $gBackgroundColor
    menu ${cm}
    $cm add check -label "Concrete Classes" \
	    -command "toggleGraphView gShowConcrete" \
	    -variable gShowConcrete -onvalue 1 -offvalue 0
    $cm add check -label "  Hide TEST_* Classes" \
 	    -command "toggleNoShowTestClasses" 
    $cm add check -label "  Hide C_* Classes" \
 	    -command "toggleNoShowCClasses"
    $cm add check -label "Abstract Classes" \
	    -command "toggleGraphView gShowAbstract" \
	    -variable gShowAbstract -onvalue 1 -offvalue 0
    $cm add check -label "Modules" \
	    -command "toggleGraphView gShowModule" \
	    -variable gShowModule -onvalue 1 -offvalue 0

    $cm add check -label "Graph Update"  \
	    -variable gGraphUpdate -onvalue "1" -offvalue "0"
    $cm add check -label "Horizontal Graph"  -command initGraph \
	    -variable gHorizontalGraph -onvalue "true" -offvalue "false"
    $cm add check -label "Fixed Size Graph"  -command initGraph \
	    -variable gFixedSizeInitGraph -onvalue "true" -offvalue "false"
    $cm add command -label "Dump Graph (Postscript)"  -command dumpPS
    $cm add command -label "Dump Graph (Dot)" -command "dumpCurGraphForDot"
    $cm add command -label "Dump FullGraph (Dot)" -command "dumpFullGraphForDot"
    $cm add command -label "Dump All State" -command dumpState
    debugPuts "Made graph menu"

    #    set cm ${frameName}.custom.menu
    # menubutton ${frameName}.custom -text "Custom" -menu ${cm} \
    #	    -bg $gBackgroundColor
    # menu ${cm}

    debugPuts "Made graph  menu ...."

    set cm ${frameName}.text.menu
    menubutton ${frameName}.text -text "Text" -menu ${cm} \
	    -bg $gBackgroundColor
    menu ${cm} 
    $cm add check -label "Highlight comments" \
	    -variable gHighlightComments -onvalue 1 -offvalue 0 
    $cm add check -label "Use Emacs" -command "toggleUseEmacs"\
		-variable gUseEmacs -onvalue 1 -offvalue 0 
    debugPuts "Made text menu ...."

    set cm ${frameName}.help.menu
    menubutton ${frameName}.help -text "Help" -menu ${cm} \
	    -bg $gBackgroundColor -bd 2
    menu ${cm}
    $cm add command -label "What's New" -command "showHelp NewStuff"
    $cm add command -label "Menus " -command "showHelp menu"
    $cm add command -label "Graph Pane" -command "showHelp graph"
    $cm add command -label "Class Pane" -command "showHelp class-list"
    $cm add command -label "Routine Pane " -command "showHelp routine"
    $cm add command -label "Text Pane" -command "showHelp text"
    $cm add command -label "Using Emacs" -command "showHelp emacs"
    $cm add command -label "Browser Speed" -command "showHelp speed"
    $cm add command -label "Customization" -command "showHelp customization"
    $cm add command -label ".bs.tcl Example" -command "showHelp ExampleCustomization"
    $cm add command -label "Bugs " -command "showHelp bugs"
    $cm add command -label "Acknowledgements" -command "showHelp acknowledgements"
    $cm add command -label "Implementation" -command "showHelp implementation"
    $cm add command -label "Generating HTML" -command "showHelp HTML"
    $cm add command -label "Ideas" -command "showHelp ideas"

    debugPuts "Made help menu ...."

    set cm ${frameName}.feature.menu
    menubutton ${frameName}.feature -text "Features" -menu ${cm} \
	    -bg $gBackgroundColor -bd 2 
    menu ${cm}
    $cm add separator 
    $cm add check -label "Private Features" \
	    -command "toggleFeature gShowPrivate "  \
	    -variable gShowPrivate -onvalue 1 -offvalue 0
    $cm add check -label "Public Features" \
	    -command "toggleFeature gShowPublic" \
	    -variable gShowPublic -onvalue 1 -offvalue 0
    $cm add check -label "Attr Features" \
	    -command "toggleFeature gShowAttr" \
	    -variable gShowAttr -onvalue 1 -offvalue 0
    $cm add check -label "Shared Features" \
	    -command "toggleFeature gShowShared" \
	    -variable gShowShared -onvalue 1 -offvalue 0
    $cm add check -label "Const Features" \
	    -command "toggleFeature gShowConst" \
	    -variable gShowConst -onvalue 1 -offvalue 0

    debugPuts "Made feature menu ...."

    pack ${frameName}.graph -side left -padx 5    
    pack ${frameName}.feature -side left -padx 5
    pack ${frameName}.text -side left -padx 5
    pack ${frameName}.help -side left -padx 10
    #    pack ${frameName}.library -side left -padx 5
    debugPuts "Packed menu items ...."

    tk_menuBar ${frameName} ${frameName}.layout ${frameName}.library ${frameName}.graph ${frameName}.help

    debugPuts "Created menu bar ...."
}

############### Menu items ######################################
proc toggleUseEmacs { } {
    inform "Run M-x server-start in an existing emacs"
}

#  Toggling views of routines #
proc toggleFeature { typeOfFeature } {
    # puts "$typeOfFeature"
    tkkit_cb toggleBoolVar $typeOfFeature 
    showClassDef 
}
proc toggleGraphView { typeOfFeature } {
    tkkit_cb toggleBoolVar $typeOfFeature 
    initGraph
}

proc toggleNoShowCClasses { } {
    global gNoShow_C_Classes
    global gCClasses
    set gNoShow_C_Classes [toggleNoShow $gNoShow_C_Classes ${gCClasses}]
    initGraph
}
 

proc toggleNoShowTestClasses { } {
    global gNoShow_Test_Classes
    global gTestClasses
    set gNoShow_Test_Classes [toggleNoShow $gNoShow_Test_Classes ${gTestClasses}]
    initGraph
}

proc flushCache { } {
    # Temporary hack - don't feel like adding a callback right now!
    tkkit_cb toggleBoolVar "gShowConcrete"
    tkkit_cb toggleBoolVar "gShowConcrete"    
}

proc toggleNoShow { controlFlag classList } {
    flushCache
    if {$controlFlag} {
	tkkit_cb deactivateClasses ${classList}
	return 0
    } else {
	tkkit_cb activateClasses ${classList}
	return 1
    }

}


proc showHelp { fileName } {
    global gHome
    gotoFile "Help" $gHome/Browser/help.${fileName} 1 false
}

proc dumpPS { } {
    global gGraphFrame
    global gFontSizes
    
    set printerFonts($gFontSizes(14)) [list helvetica 14]
    set printerFonts($gFontSizes(13)) [list helvetica 13]
    set printerFonts($gFontSizes(12)) [list helvetica 13]
    set printerFonts($gFontSizes(11)) [list helvetica 13]
    set printerFonts($gFontSizes(10)) [list helvetica 10]
    set printerFonts($gFontSizes(9)) [list helvetica 9]
    set printerFonts($gFontSizes(8)) [list helvetica 8]
    set printerFonts($gFontSizes(7)) [list helvetica 7]
    set printerFonts($gFontSizes(6)) [list helvetica 6]
    set printerFonts($gFontSizes(5)) [list helvetica 5]
    set printerFonts($gFontSizes(4)) [list helvetica 4]
    set printerFonts($gFontSizes(3)) [list helvetica 4]
    set printerFonts($gFontSizes(2)) [list helvetica 4]
    set printerFonts($gFontSizes(1)) [list helvetica 4]

    set dialog [tk_dialog .dialog "Dumping postscript" \
	    "Writing postscript to file browse-layout.eps"\
	    "" 0 "Confirm" "Cancel"]
    if {$dialog == 0} { 
	inform "Dumping to postscript file: browser-layout.eps"
	${gGraphFrame}.c postscript -fontmap printerFonts -file "browse-layout.eps" 
	inform "Dumped postscript in browse-layout.eps"
    }


}

proc dumpState { } {
    global gAllClasses
    global gModToClasses
    set dialog [tk_dialog .dialog "Dumping documentation" \
		    "Writing docs to doc-all-classes.gen.tcl"\
		    "" 0 "Confirm" "Cancel"]

    if {$dialog != 0} {
	return
    }
    set sorted [lsort -increasing -ascii $gAllClasses]
    # First dump an outer file of class names
    set mainfl [open "doc-all-classes.gen.tcl" w]
    set unsortedModules [array names gModToClasses]
    set sortedModules [lsort -increasing -ascii $unsortedModules ]
    set res ""
    foreach module $sortedModules {
	set res  "${res} set gModToClasses($module) { $gModToClasses($module) }\n"
	set classes [lsort -increasing -ascii $gModToClasses($module)]
	foreach cl ${classes} {
	    puts "Processing ${cl}..."
	    set classNm ${cl}
	    set cdef [tkkit_cb getClassInfo ${classNm}]
	    set res "${res} set gClassDef($classNm) { $cdef }\n"
	}
	# End of a module
	regsub -all {[\$]} $res "\\\$" newRes
	puts $mainfl "$newRes"
    }
    close $mainfl
}

# Generates a version of the current graph in a form that is usable
# by "dot"
proc dumpCurGraphForDot { } {
    global gCurClassName
    set layoutInfo [tkkit_cb getRestrictedLayout " "]
    dumpDotGraph "Class${gCurClassName}.gen.dot" ${layoutInfo}
}

# Generates a version of the complete graph of classes in a form
# suitable for use by the graph layout program from AT&T called "dot"
# You can modify this to generate any format you please ...
proc dumpFullGraphForDot { } {
    set fulg [tkkit_cb getFullLayout ]
    dumpDotGraph "AllClasses.gen.dot" fulg
}

proc dumpDotGraph { fileNm g } {
    set dialog [tk_dialog .dialog "Dumping full graph in dot form" \
	    "Writing graph to ${fileNm}"\
	    "" 0 "Confirm" "Cancel"]
    if {$dialog != 0} {
	return
    }
    set abs "gold"
    set conc "seagreen"
    set mod "skyblue"
    set fl [open ${fileNm} w]
    puts $fl "digraph classes {"
    puts $fl "page= \"8.5,11\""
    puts $fl "ranksep=0.5"
    puts $fl "nodesep=0.15"
    puts $fl  "concentrate=TRUE"
    puts $fl  "center=TRUE"
    puts $fl "orientation=landscape"
    puts $fl "node \[height=0.1,fontsize=4\];"

    set edges [lindex ${g} 2]
    foreach edge ${edges} {
	set src [lindex ${edge} 0]
	set dest [lindex ${edge} 1]
	set modedge false
	if { [isAbstract ${src}]} {
	    puts $fl "edge \[style=bold,color=${abs},weight=3\];"
	    puts $fl "\"${src}\" \[color=${abs}\]"
	} elseif { [isModule ${src} ]} {
	    puts $fl "edge \[style=filled,color=${mod},weight=2\];"
	    set src [getModuleName $src]
	    puts $fl "\"${src}\" \[color=${mod}\]"
	    set modedge true
	} else {
	    puts $fl "\"${src}\" \[color=${conc}\]"
	    puts $fl "edge \[style=filled,color=${conc}\];"
	}

	if { [isAbstract ${dest}]} {
	    set destText "\"${dest}\" \[color=${abs}\]"
	} elseif { [isModule ${dest} ]} {
	    set dest [getModuleName $dest]
	    puts $fl "\"${dest}\" \[color=${mod}\]"
	} else {
	    puts $fl "\"${dest}\" \[color=${conc}\]"
	    puts $fl "edge \[style=dotted,color=${conc}\];"
	}
	if {$modedge} {
	} else {
	    puts $fl "\"${src}\" -> \"${dest}\";"
	}
 

    }
    puts $fl "}"
    close $fl
    inform "Trying to execute dot on the generated file to generate ps...."
    update
    exec dot -Tps ${fileNm} -o ${fileNm}.ps
    inform "Executed dot on the generated file to generate ps."
    update
}
########################################################################
#                          End of MENU BAR 
########################################################################
########################################################################
#                          CANVAS/Graph Related
########################################################################
proc makeCanvas { topName xsz ysz viewxsz viewysz } {
    global gDeepBackgroundColor
    global gBackgroundColor
    global gIsColor

    # Create a scrollable canvas under the frame topName
    # The items will be topName.c, topName.vscroll and topName.hscroll

    canvas ${topName}.c -relief raised -height $viewysz -width $viewxsz -bd 2\
	    -scrollregion [list 0 0 $xsz $ysz] -yscroll "${topName}.vscroll set"\
	    -xscroll "${topName}.hscroll set" -bg $gBackgroundColor 
    if {$gIsColor} {
	scrollbar ${topName}.vscroll -relief flat \
		-command  "${topName}.c yview" -bg $gDeepBackgroundColor
	scrollbar ${topName}.hscroll -relief flat -orient horiz \
		-command  "${topName}.c xview" -bg $gDeepBackgroundColor
	scale ${topName}.scale -orient vertical -command "canvasZoom" \
		-from 100 -to 10 -bg $gDeepBackgroundColor -relief raised
    }  else {
	scrollbar ${topName}.vscroll -relief flat \
		-command  "${topName}.c yview" -bg $gDeepBackgroundColor
	scrollbar ${topName}.hscroll -relief flat -orient horiz \
		-command  "${topName}.c xview" -bg $gDeepBackgroundColor
	scale ${topName}.scale -orient vertical -command "canvasZoom" \
		-from 100 -to 10 -bg $gDeepBackgroundColor -relief raised
    }

    ${topName}.scale configure -showvalue 0
    ${topName}.scale set 100

    pack ${topName}.scale -side left -fill y -anchor w -padx 5 
    frame ${topName}.separator -width 10 -relief flat -bg $gDeepBackgroundColor
    pack ${topName}.separator -side left -fill y

    pack ${topName}.vscroll -side right -fill y
    pack ${topName}.hscroll -side bottom -fill x
    pack ${topName}.c -side top -fill both 


    #  place ${topName}.c  -relwidth 1.0 -relheight 1.0
}


######################### Graph Related Procedures ###################
#proc canvasZoom { ignoredScale } {
#    canvasZoomAroundOrigin
#}

# Zoom the canvas around the selected item, or around the origin if
# no item is selected
proc canvasZoom { ignoredScale } {
    global gGraphFrame
    global gCurClassName
    global gOldScale

    # Find coods of last selected item
    set sel [${gGraphFrame}.c find withtag ${gCurClassName}]
    if { $sel == "" } {
	# If not selected item
	canvasZoomAroundOrigin
    } else {
	set scaleFac [${gGraphFrame}.scale get]
	set desiredScale [expr $scaleFac/100.0]
	set scaleFactor [expr $desiredScale/$gOldScale]
	set gOldScale $desiredScale
	setCanvasFonts
	set coods [${gGraphFrame}.c coords ${gCurClassName}]
	debugPuts "Scaling around ${gCurClassName} $coods"
	set cx [lindex $coods 0]
	set cy [lindex $coods 1]
	${gGraphFrame}.c scale node $cx $cy $scaleFactor $scaleFactor
	${gGraphFrame}.c scale arrow $cx $cy $scaleFactor $scaleFactor
	${gGraphFrame}.c scale selectRect $cx $cy $scaleFactor $scaleFactor
    }
    return
}

# Zoom the canvas around the origin - (ignore current view)
# Used by initGraph for initial zooming
proc canvasZoomAroundOrigin {  } {
    global gGraphFrame
    global gCurClassName
    global gOldScale
    
    debugPuts "Scaling around origin"
    set scaleFac [${gGraphFrame}.scale get]
    set desiredScale [expr $scaleFac/100.0]
    set scaleFactor [expr $desiredScale/$gOldScale]
    set gOldScale $desiredScale
    setCanvasFonts
    # Find coods of last selected item
    debugPuts "Scaling around origin"
    ${gGraphFrame}.c scale node 0.5 0.5 $scaleFactor $scaleFactor
    ${gGraphFrame}.c scale arrow 0.5 0.5 $scaleFactor $scaleFactor
    ${gGraphFrame}.c scale selectRect 0.5 0.5 $scaleFactor $scaleFactor
}

# Set the fonts of nodes on the canvas appropriately, according to
# the current canvas scale and making use of the table 
# gFontSizes and gBoldFontSizes.
proc setCanvasFonts { } {
    global gGraphFrame
    global gFontSizes
    global gBoldFontSizes
    global gItalicFontSizes
    global gIsColor 

    set scaleFac [${gGraphFrame}.scale get]
    set desiredScale [expr $scaleFac/100.0]
    set fontSz   [expr $desiredScale * 14]
    set fs  [format "%.0f" $fontSz]

    set font $gFontSizes($fs)
    set boldFont $gBoldFontSizes($fs)
    set italicFont $gItalicFontSizes($fs)
    ${gGraphFrame}.c itemconfigure node -font ${font}
    ${gGraphFrame}.c itemconfigure abstract -font ${boldFont}
    ${gGraphFrame}.c itemconfigure module -font ${italicFont}
}

proc setScale { scaleValue } {
    global gGraphFrame
    ${gGraphFrame}.scale set ${scaleValue}
    canvasZoom "ignored"
}
########################### Nodes ############################
#  Full node text refers to the full file name in the case of a module
#  and is the same as nodeText in other cases
# Node Tags:      [see proc makeTextNodeAt]
#                  ${fullNodeText} 
#                  node 
#                  node${fullNodeText} 
#                  ${nodeText} 
#                  [one of the following tags:
#                  abstract 
#                  concrete 
#                  module
#                  ]

proc makeTextNodeAt { classX classY nodeText fullNodeText } {
    # Add the text object centered at x, y with text nodeText
    # Add the nodeText as the first tag of the node
    global gGraphFrame
    global gIsColor
    global gSubtypeColor

    set c [string index ${nodeText} 0]
    if { $c == "\$" } {
	if { $gIsColor } {
	    set thecolor $gSubtypeColor
	} else {
	    set thecolor "black"
	}
	set tag "${fullNodeText} node node${fullNodeText} ${nodeText} abstract"
	${gGraphFrame}.c  create text ${classX} ${classY}\
		-text ${nodeText} -fill ${thecolor} -anchor center \
		-tags ${tag}
    } elseif { [regexp {[A-Za-z0-9./_]*.module} $fullNodeText ] } {
	${gGraphFrame}.c  create text ${classX} ${classY}\
		-text ${nodeText}  -anchor center \
		-tags "${fullNodeText} node node${fullNodeText} ${nodeText} module"

    } else {
	${gGraphFrame}.c  create text ${classX} ${classY}\
		-text ${nodeText} -anchor center \
		-tags "${fullNodeText} node node${fullNodeText} ${nodeText} concrete"
    }

}


#######################################################################
# Arrow tags:      [See proc connect]
#                  ${srcName} 
#                  ${destName} 
#                  arrow${srcName} 
#                  arrow${destName} 
#                  arrow  
#                  [one of the following tags:
#                  subtype moduleContains include
#                  ]
# Note: redrawArrow uses the first 2 tags to redraw the arrow.
#       moveArrowsTouching uses the second 2 tags to determine arrows
#       touching a node
#######################################################################
proc connect { frameName srcName destName } {
    global gHorizontalGraph

    set srcCood [${frameName}.c coords ${srcName}]	
    set destCood [${frameName}.c coords ${destName} ]

    set sl  [llength ${srcCood}]
    set dl  [llength ${destCood}]
    # If either source or destination do not exist, return.
    if { ${sl}  == 0 } {
	return
    }
    if { $dl == 0 } {
	return
    }
    set  srcx [lindex $srcCood 0]
    set  srcy [lindex $srcCood 1]	
    set  destx [lindex $destCood 0]
    set desty [lindex $destCood 1]
    set yoff 10
    set xoff 50
    if { $gHorizontalGraph } {
	set srcBB [${frameName}.c bbox node${srcName}]
	set destBB [${frameName}.c bbox node${destName}]
	set sty [expr $srcy]
	set endy [expr $desty]
	if { $srcx > $destx } {
	    # A left to right arrow
	    set stx [lindex $srcBB 0]
	    set endx [lindex $destBB 2]
	    # set stx [expr $srcx - $xoff]
	    # set endx [expr $destx + $xoff]
	} else {
	    set stx [lindex $srcBB 2]
	    set endx [lindex $destBB 0]
	    #set stx [expr $srcx + $xoff]
	    #set endx [expr $destx - $xoff]

	}
    } else {
	# Vertical graph - does not need bounding box to determine where
	# arrows should start and end
	set stx [expr $srcx]
	set endx [expr $destx]
	if { $srcy > $desty } {
	    # An upward arrow
	    set sty [expr $srcy - $yoff]
	    set endy [expr $desty + $yoff]
	} else {
	    set sty [expr $srcy + $yoff]
	    set endy [expr $desty - $yoff]
	}
    }
    set srcTags [${frameName}.c gettags ${srcName}]
    set srcType [lindex $srcTags 4]
    #    puts "Source name: ${srcName} ${srcType}"
    if { $srcType == "abstract" } {
	${frameName}.c create line ${stx} ${sty} ${endx} ${endy} \
		-arrow last -width 1.0 \
		-tags "${srcName} ${destName} arrow${srcName} \
		arrow${destName} arrow subtype"
    } elseif { $srcType == "module" } {
	${frameName}.c create line ${stx} ${sty} ${endx} ${endy} \
		-arrow last -width 1.0 \
		-tags "${srcName} ${destName} arrow${srcName} \
		arrow${destName} arrow moduleContains "

    } else {
	${frameName}.c create line ${stx} ${sty} ${endx} ${endy} \
		-arrow last -width 1.0 \
		-tags "${srcName} ${destName} arrow${srcName} \
		arrow${destName} arrow include"

    }

}

proc redrawArrow {canvasFrame arrow } {
    # Arrow is the tagId of the arrow
    set tags [${canvasFrame}.c gettags ${arrow}]
    connect ${canvasFrame} [lindex $tags 0] [lindex $tags 1] 
    ${canvasFrame}.c delete ${arrow}
}
debugPuts "Done with main graph procs"

# ############## Code to move around nodes #####################
#  Moving a node begins when the correct binding fires
#  nodeBeginMove ${canvasFrame} %x %y
###############################################################
# Global variables that hold the coods of the object currently being moved
set gMovingNodePos(lastX) 0
set gMovingNodePos(lastY) 0

proc nodeBeginMove { canvasFrame xpos ypos } {
    global gMovingNodePos
    # Delete the previously marked selection
    ${canvasFrame}.c dtag selected
    # Name the current object the selected object
    ${canvasFrame}.c addtag selected withtag current
    # Get the node name and the corresponding box
    set tags [${canvasFrame}.c gettags selected]
    set movingNodeName [lindex $tags 0]
    # Move over other objects
    ${canvasFrame}.c raise selected
    # Set the moving coods to those of the current (selected) object
    set gMovingNodePos(lastX) $xpos
    set gMovingNodePos(lastY) $ypos

}

proc nodeContinueMove { canvasFrame xpos ypos } {
    # As movement continues, update the position and redraw object
    # Don't redraw arrows
    global gMovingNodePos
    ${canvasFrame}.c move selected \
	    [expr $xpos-$gMovingNodePos(lastX)] \
	    [expr $ypos-$gMovingNodePos(lastY)]
    set gMovingNodePos(lastX) $xpos
    set gMovingNodePos(lastY) $ypos
}


proc nodeEndMove { canvasFrame xpos ypos } {
    # At the end of the movement, lower the object, and move arrows
    # that were connected to it
    global gMovingNodePos
    ${canvasFrame}.c lower selected
    # Get the first tag of the node, which is its name
    set itemTaggedCurrent [${canvasFrame}.c find withtag current ]
    set tags [${canvasFrame}.c gettags $itemTaggedCurrent]
    set nodeName [lindex $tags 0]
    moveArrowsTouching ${canvasFrame} ${nodeName}
    ${canvasFrame}.c dtag selected
    canvasShowSelected
}

proc  moveNodeTo { canvasFrame itemName  xpos ypos } {
    # Move node to xpos ypos, and all the arrows touching it
    set curId [${canvasFrame}.c find withtag node${itemName}]
    if { [llength $curId] == 0 } {
	return
    }
    ${canvasFrame}.c raise $curId
    set oldCood [${canvasFrame}.c coords $curId]
    set oldX [lindex $oldCood 0]
    set oldY [lindex $oldCood 1]
    ${canvasFrame}.c move $curId \
	    [expr ${xpos} - ${oldX}] [expr ${ypos} - ${oldY}]
    ${canvasFrame}.c lower $curId
    moveArrowsTouching ${canvasFrame} ${itemName}
}

debugPuts "Done with node movement"
#################### Code to move arrows to follow a node ################
# Find all arrows that have the moved node as the source or destination
# Find the actual line and redraw it, using the node names 
##########################################################################
proc moveArrowsTouching { canvasFrame nodeName } {
    # Get the new coods of the node with 
    # puts "Moving arrows touching: ${nodeName}"
    set touchingArrows [${canvasFrame}.c find withtag "arrow${nodeName}"]
    # puts "Touching ${touchingArrows}"
    foreach arrow ${touchingArrows} {
	redrawArrow ${canvasFrame} ${arrow}
    }
    canvasHighlight
}

proc canvasHighlight { } {
    global gIsColor
    if {$gIsColor} {
	canvasHighlightColor
    } else {
	canvasHighlightMono
    }
}

proc canvasHighlightMono { } {
    global gGraphFrame
    global gHome

    ${gGraphFrame}.c itemconfigure subtype -width 2 
    ${gGraphFrame}.c itemconfigure moduleContains -width 2 -stipple gray50
    ${gGraphFrame}.c itemconfigure include -width 1

    ${gGraphFrame}.c itemconfigure node -fill black

    # ${gGraphMonoSubtype}
    # "@${gHome}/Browser/dashed-hard.bitmap" 
    # -stipple ${gGraphMonoModuleContains} 
    # -stipple ${gGraphMonoInclude}  -width 2


}
proc canvasHighlightColor { } {
    global gGraphFrame
    global gSubtypeColor
    global gModuleColor
    global gIncludeColor
    
    ${gGraphFrame}.c itemconfigure subtype -fill $gSubtypeColor
    ${gGraphFrame}.c itemconfigure moduleContains -fill $gModuleColor
    ${gGraphFrame}.c itemconfigure include -fill $gIncludeColor
    ${gGraphFrame}.c itemconfigure module -fill $gModuleColor
    # Raises moduleContains edges above subtype edges
    #${gGraphFrame}.c raise moduleContains subtype
    #${gGraphFrame}.c raise moduleContains include

    ${gGraphFrame}.c itemconfigure node -fill black
}


# Handle clicking on a node in the graph
# Find and highlight the appropriate entry in the list of classes,
# thereby going to the correct file/class location.
proc selectGraphNode { xpos ypos } {
    global gCurClassName
    global gGraphFrame
    global gClassList
    global gAllClasses

    set sel [${gGraphFrame}.c find withtag current]
    # Unset the previously selected class
    set ctags [${gGraphFrame}.c gettags $sel]
    set nm [lindex $ctags 0]
    set ci  [lsearch $gAllClasses [list ${nm}]]
    selectClass $nm [expr $ci + 1] false
    ${gClassList}.l yview  $ci
}

proc clearGraph { } {
    global gGraphFrame
    ${gGraphFrame}.c delete node
    ${gGraphFrame}.c delete arrow
}

# Reinitialize the display graph, based on the currently selected
# class/module and the current set of visible features
proc initGraph { } {
    global gFixedSizeInitGraph

    if {$gFixedSizeInitGraph} {
	initGraphVariableScale
    } else {
	initGraphFixedScale
    }
    
}

# Initialize the graph to fit on one screen
proc initGraphVariableScale { } {
    global gGraphFrame
    global gClassList
    global gGraphUpdate
    global gCurClassName
    global gHorizontalGraph

    if {!${gGraphUpdate}} {
	return
    }
    debugPuts "Redrawing Graph"
    clearGraph
    set layoutInfo [tkkit_cb getRestrictedLayout " "]
    setScale 100
    debugPuts "Layout ${layoutInfo}"
    set graphSize [lindex $layoutInfo 0]
    # Class locations  
    set classes [lindex $layoutInfo 1]  
    # Edge information
    set edges [lindex $layoutInfo 2]    
    set singletons [lindex $layoutInfo 3]    
    set width [lindex ${graphSize} 1]
    set height [lindex ${graphSize} 0]
    set newScale 70
    
    if { $width > 3 } { 
	set newScale  70
    }
    if { $width > 7 } {
	set newScale 50
    }
    if { $width > 15 } {
	set newScale 30
    }
    if { $width > 17 } {
	set newScale 20
    }
    if { $width > 20 } {
	set newScale 10
    }
    set xFactor [expr 800.0 / $newScale  ]
    set yFactor [expr 250.0 / $newScale ]
    foreach class $classes {
	set className [lindex ${class} 0]
	if {$gHorizontalGraph} {
	    set rawX  [lindex ${class} 2]
	    set rawY  [lindex ${class} 1]
	    set classX [expr ${rawX} * $xFactor +50.0 ]
	    set classY [expr ${rawY} * $yFactor + 20.0 ]
	} else {
	    set rawX  [lindex ${class} 1]
	    set rawY  [lindex ${class} 2]
	    set classX [expr ${rawX} * $xFactor ]
	    set classY [expr ${rawY} * $yFactor + 60.0 ]
	}

	#puts "rawX: $rawX rawY $rawY ClassX: $classX classY $classY"
	set fullNodeText $className
	if { [isModule $className] } {
	    set nodeText [getModuleName $className]
	} else {
	    set nodeText $className
	}
	makeTextNodeAt ${classX} ${classY} ${nodeText} ${fullNodeText}
    }
    foreach edge $edges {
	connect  ${gGraphFrame} [lindex $edge 0] [lindex $edge 1]
    }

    ${gGraphFrame}.c addtag selected withtag ${gCurClassName}
    ${gGraphFrame}.scale set ${newScale}
    canvasZoomAroundOrigin 
    canvasHighlight
    ${gGraphFrame}.c bind node <Enter> {
	set tags [${gGraphFrame}.c gettags current]
	set nodeName [lindex $tags 0]
	inform "Over node ${nodeName}"
    }
    ${gGraphFrame}.c bind node <Leave> {
	inform ""
    }
    #    ${gGraphFrame}.c itemconfigure abstract -fill $gSubtypeColor
    setCanvasFonts
    canvasShowSelected

}

debugPuts "Done with init graph"

# Reinitialize the display graph with a fixed scale..
proc initGraphFixedScale { } {
    global gGraphFrame
    global gClassList
    global gGraphUpdate
    global gCurClassName
    global gHorizontalGraph

    if {!${gGraphUpdate}} {
	return
    }
    debugPuts "Redrawing Graph"
    clearGraph
    set layoutInfo [tkkit_cb getRestrictedLayout " "]
    setScale 100
    debugPuts "Layout ${layoutInfo}"
    set graphSize [lindex $layoutInfo 0]
    # Class locations  
    set classes [lindex $layoutInfo 1]  
    # Edge information
    set edges [lindex $layoutInfo 2]    
    set singletons [lindex $layoutInfo 3]    
    set width [lindex ${graphSize} 1]
    set height [lindex ${graphSize} 0]
    set newScale 70
    set xFactor [expr ${width}]
    set yFactor [expr ${height}]
    foreach class $classes {
	set className [lindex ${class} 0]
	if {$gHorizontalGraph} {
	    set rawX  [lindex ${class} 2]
	    set rawY  [lindex ${class} 1]
	    set classX [expr ${rawX} * $xFactor +50.0 ]
	    set classY [expr ${rawY} * $yFactor + 20.0 ]
	} else {
	    set rawX  [lindex ${class} 1]
	    set rawY  [lindex ${class} 2]
	    set classX [expr ${rawX} * $xFactor ]
	    set classY [expr ${rawY} * $yFactor + 60.0 ]
	}

	#puts "rawX: $rawX rawY $rawY ClassX: $classX classY $classY"
	set fullNodeText $className
	if { [isModule $className] } {
	    set nodeText [getModuleName $className]
	} else {
	    set nodeText $className
	}
	makeTextNodeAt ${classX} ${classY} ${nodeText} ${fullNodeText}
    }
    foreach edge $edges {
	connect  ${gGraphFrame} [lindex $edge 0] [lindex $edge 1]
    }

    ${gGraphFrame}.c addtag selected withtag ${gCurClassName}
    ${gGraphFrame}.scale set ${newScale}
    canvasZoomAroundOrigin 
    canvasHighlight
    ${gGraphFrame}.c bind node <Enter> {
	set tags [${gGraphFrame}.c gettags current]
	set nodeName [lindex $tags 0]
	inform "Over node ${nodeName}"
    }
    ${gGraphFrame}.c bind node <Leave> {
	inform ""
    }
    #    ${gGraphFrame}.c itemconfigure abstract -fill $gSubtypeColor
    setCanvasFonts
    canvasShowSelected

}

debugPuts "Done with init graph"


# Highlight the selected node with a rectangle. 
# Problem arises since the size is not easily known (fonts
# scaling is not exact). Hence, moving the node around
# when the size is small will result in a bigger box shown!
# CURRENTLY UNUSED
proc canvasShowSelectedBox { } {
    global gGraphFrame
    global gCurClassName
    set sel [${gGraphFrame}.c find withtag node${gCurClassName}]
    set bb [${gGraphFrame}.c bbox node${gCurClassName}]
    set ts [${gGraphFrame}.c gettags node${gCurClassName}]
    if {$bb == "" } {
	
    } else {
	${gGraphFrame}.c delete withtag selectRect
	${gGraphFrame}.c create rectangle \
		[expr [lindex $bb 0] - 10] \
		[expr [lindex $bb 1] - 10 ]\
		[expr [lindex $bb 2] + 10] \
		[expr [lindex $bb 3] + 10] \
		-tags selectRect
    }
    
}

proc canvasShowSelected { } {
    global gGraphFrame
    global gCurClassName
    global gFeatureListTitleColor
    global gSubtypeColor
    global gModuleColor
    global gIsColor

    set sel [${gGraphFrame}.c find withtag node${gCurClassName}]
    ${gGraphFrame}.c itemconfigure node -fill black 
    if { $gIsColor } {
	${gGraphFrame}.c itemconfigure abstract -fill $gSubtypeColor
	${gGraphFrame}.c itemconfigure module -fill $gModuleColor
	
    }
    ${gGraphFrame}.c itemconfigure ${sel} -fill $gFeatureListTitleColor 
}

debugPuts "After canvasShowSelected"

########################################################################
#                          End of Canvas/Graph
########################################################################
########################################################################
#                          Text Display widget
########################################################################
# Make the text display widget
# Arguments: frameName - name of the frame in which to create the widget
#            title  - title displayed on top of the widget
# Creates a text widget with the following subparts
#  .t        text      : actual text window 
#  .yscroll  scrollbar : Vertical scroll
#  .title    lable     : Holds the frameName
#######################################################################
proc makeText { frameName } {
    global gBackgroundColor 
    global gDeepBackgroundColor
    global gTextTitleColor
    global gTextTitleRev
    global gTitleFont

    set title "Class Text"
    text ${frameName}.t -bd 2 \
	    -yscrollcommand "${frameName}.yscroll set" -wrap none \
	    -font "7x13" -relief raised -bg $gBackgroundColor
    scrollbar ${frameName}.yscroll  -command "${frameName}.t yview" \
	    -bg $gDeepBackgroundColor
    ${frameName}.t insert 0.0 {\
	    Begin browsing by clicking on a class. }
    ${frameName}.t mark set insert 0.0 
    pack ${frameName}.yscroll -side right -fill y
    frame ${frameName}.top
    button ${frameName}.top.saveButton -text "Save" -command "saveText" \
	    -bg $gBackgroundColor -fg "black" 
    pack ${frameName}.top.saveButton -side right
    if { ${title} != ""} {
	label ${frameName}.top.title -text ${title} -borderwidth 2 \
	-bg $gTextTitleColor -fg $gTextTitleRev  -relief raised -font ${gTitleFont}
	pack ${frameName}.top.title -side left -expand yes -fill x
    }
    pack ${frameName}.t -expand yes -fill both -side bottom
    pack ${frameName}.top -side top -fill x 
    bind ${frameName}.t <1> {focus %W; %W mark set insert @%x,%y ; %W mark set anchor insert}
    bind ${frameName}.t <2> { 
	set tk_priv(selectMode) char
	%W mark set insert @%x,%y
	%W mark set anchor insert
	if {[lindex [%W config -state] 4] == "normal"} {focus %W}
	selection clear %W
    }
    bind ${frameName}.t <Left>  {%W mark set insert "insert -1 chars"}
    bind ${frameName}.t <Control-b>  {%W mark set insert "insert -1 chars"}
    bind ${frameName}.t <Control-k>  {%W delete insert "insert + 1 lines linestart"}
    bind ${frameName}.t <Alt-b>  {%W mark set insert "insert -1 chars"; \
	    %W mark set insert "insert wordstart"; }
    bind ${frameName}.t <Control-a>  {%W mark set insert "insert linestart"}
    bind ${frameName}.t <Control-e>  {%W mark set insert "insert lineend"}
    bind ${frameName}.t <Right> {%W mark set insert "insert +1 chars"}
    bind ${frameName}.t <Control-f> {%W mark set insert "insert +1 chars"}
    bind ${frameName}.t <Up> {%W mark set insert "insert -1 lines"; %W yview -pickplace insert}
    bind ${frameName}.t <Down> {%W mark set insert "insert +1 lines"; %W yview -pickplace insert}
    bind ${frameName}.t <Control-p> {%W mark set insert "insert -1 lines"; %W yview -pickplace insert }
    bind ${frameName}.t <Control-n> {%W mark set insert "insert +1 lines"; %W yview -pickplace insert}
    bind ${frameName}.t <Control-l> {%W yview insert}
    bind ${frameName}.t <Control-d> { 
	set s [%W tag ranges "sel"]
	if { $s == ""} {
	} else {
	    %W delete sel.first sel.last
	}
    }
}

debugPuts "before add Comments"

proc addCommentTags { } {
    global gTextDisplay
    global gCommentThreshold
    global gCommentFont 
    global gCommentColor
    global gHighlightComments
    set txt [$gTextDisplay.t get 1.0 end]
    set txtl [ split $txt "\n" ]
    set txtlsize [ llength $txtl ] 
    if {!$gHighlightComments} {
	return
    }
    if { $txtlsize < $gCommentThreshold } {
	inform "Highlighting comments (may be slow for large files)..."
	update
	set lineno 0
	while { $lineno < $txtlsize } {
	    set thisline [ lindex $txtl $lineno ]
	    incr lineno
	    set from [ string first "--" $thisline ]
	    if { $from >= 0 } {
		#puts "Line: $lineno  From: $from"
		${gTextDisplay}.t tag add comment "$lineno.$from" \
			"$lineno.$from lineend"
	    }
	}
	${gTextDisplay}.t tag configure comment \
		-font $gCommentFont \
		-foreground $gCommentColor
	inform "Done highlighting ..."
	update
    } else {
	inform "Large file - highlighting comments would take too long"
	update
    }
}

debugPuts "Before saveTextIfNecc..."
proc saveTextIfNecc { } {
    global gCurFile
    global gTextDisplay
    if { $gCurFile != "" } {
	set oldText  [ string trimright [getAllFile ${gCurFile}]]
	set curText [string trimright [${gTextDisplay}.t get "1.0" "end" ]]
	if {  ${oldText} != ${curText}  } {
	    set dialog [tk_dialog .dialog "Unsaved changes" \
		    "Current text is different from file ${gCurFile}"\
		    "" 0 "Save" "Discard"]
	    if {$dialog == 0} { 
		saveText
	    }
	}
    } 
}

# Goto the fileName, with the highlight on line "lineNumber"
# "isModifiable" indicates whether save should be set or not
# gCurFile is set to the new file name, and the previous
# file is saved if it has been changed (user is queried)
proc gotoFile { title fileName lineNumber isModifiable } {
    global gCurFile
    global gTextDisplay
    global gModuleColor
    global gModuleRev
    global gSubtypeColor
    global gSubtypeRev
    global gIncludeColor
    global gIncludeRev
    global gUseEmacs

    if {${gUseEmacs}} {
	exec emacsclient +${lineNumber} ${fileName} >& /dev/null &
    } elseif { $gCurFile != $fileName } {
	# puts "Old file: ${gCurFile} NewFile: ${fileName}"
	saveTextIfNecc
	set gCurFile $fileName
	set txt [getAllFile ${fileName}]
	if { [file writable ${fileName} ] } {
	    loadText "$title in ${fileName}" $txt  $isModifiable
	} else {
	    loadText "$title in ${fileName}" $txt  false
	}
	gotoLine $lineNumber
    } else {
	inform "Staying in same file...."
	${gTextDisplay}.top.title configure -text "$title"
	gotoLine $lineNumber
    }

    # mbk logic
    if { [isModule $title] == "true" } {
	set col $gModuleColor
	set colr $gModuleRev
    } else {
	if { [string index $title 0] == "\$" } {
	    set col $gSubtypeColor
	    set colr $gSubtypeRev
	} else {
	    set col $gIncludeColor
	    set colr $gIncludeRev 
	}
    }
    ${gTextDisplay}.top.title configure -bg $col -fg $colr


}

proc loadText { title txt isModifiable} {
    global gTextDisplay

    ${gTextDisplay}.t configure -state normal
    ${gTextDisplay}.top.title configure -text "$title"

    ${gTextDisplay}.t delete "1.0" "end "
    ${gTextDisplay}.t insert 0.0 $txt
    
    if { $isModifiable } {
	${gTextDisplay}.top.saveButton configure -state normal
	${gTextDisplay}.t configure -state normal
	inform "Text is writable"
    } else {
	${gTextDisplay}.top.saveButton configure -state disabled
	${gTextDisplay}.t configure -state disabled
	inform "Text is not writable"
    }
    addCommentTags
}

proc gotoLine { lineNumber } {
    global gTextDisplay
    global gTextSelectionColor
    global gTextSelectionRev


    ${gTextDisplay}.t  yview -pickplace "0.0 + ${lineNumber} lines"
    ${gTextDisplay}.t tag delete activeLine

    ${gTextDisplay}.t tag configure activeLine \
	    -foreground $gTextSelectionColor \
	    -background $gTextSelectionRev

    ${gTextDisplay}.t mark set insert "0.0 + ${lineNumber} lines"
    ${gTextDisplay}.t tag add activeLine "insert linestart" "insert lineend"
    set tags [${gTextDisplay}.t tag ranges activeLine]

}

debugPuts "After gotoFile...."


debugPuts "Before saveText...."
proc saveText { } {
    global gCurFile
    global gSafeFile
    global gTextDisplay
    if { $gCurFile == "" } {
	inform "No current file"
	return
    }
    set txt [${gTextDisplay}.t get "1.0"  "end"]
    if { $gSafeFile } {
	inform "Backing up. Change the value of $gSafeFile in browser.tcl to avoid backup"
	set orig [open ${gCurFile} r]
	set ret [read $orig]
	close $orig
	set backup [open ${gCurFile}.safetybackup w]
	puts $backup $ret
	close $backup
    }
    set orig [open ${gCurFile} w]
    puts $orig $txt
    close ${orig}

}

debugPuts "before getAllFile"

proc getAllFile { fname } {
    set ret ""
    if { [file exists ${fname}] } {
	set f [open ${fname} r]
	set ret [read $f]
	close $f
    } else {
	inform "No Description file: $fname"
	set ret "No Description file."
    }
    return $ret
}


debugPuts "before routNameFromSig"

proc routNameFromSig { routSig } {
    set endPt [string first ( ${routSig})]
    if {$endPt == -1} {
	set endPt [string first : ${routSig} ]
    }
    return [string range ${routSig} 0 [expr ${endPt} - 1]]
}



########################################################################
#                          End of Text Display widget
########################################################################
########################################################################
#                          List of Classes
########################################################################
########################################################################
# Create a list box out of a text pane with the following subpanes
# .l listbox itself
# .vscroll vertical scroller
# .hscroll horizontal scroller
#######################################################################
proc makeClassListBox { topName } {
    global gFeatureListTitleColor
    global gClassListTitleColor
    global gClassListTitleRev
    global gDeepBackgroundColor
    global gBackgroundColor
    global gTitleFont

    set title "Modules and Classes"
    text ${topName}.l \
	    -yscrollcommand "${topName}.vscroll set" -width 25 \
	    -wrap none 
    ${topName}.l configure -bg $gBackgroundColor -relief raised -bd 2 
    ${topName}.l configure -selectbackground $gFeatureListTitleColor

    scrollbar ${topName}.vscroll -bg $gDeepBackgroundColor -command "${topName}.l yview"
    pack ${topName}.vscroll -side right -fill y 
    if {${title} != ""} {
	label ${topName}.title -text ${title} -bg $gClassListTitleColor \
		-fg $gClassListTitleRev -relief raised -bd 2 \
		-font ${gTitleFont}
	pack ${topName}.title -fill x
    }
    pack ${topName}.l -expand yes -fill both

    bind ${topName}.l <B1-Motion> " "

}

##########################################################################    
# Initialize the list of all classes. Gather together classes under
# their modules, tag the text appropriately
debugPuts "before initClassList ..."
# Tagging of the class list
# Modules:        
#                 module  
#                 ${textNameForModule} 
#     module                 is used for selecting colors/fonts]    
#     ${textNameForModule}    (i.e. last part of path) is bound to the command
#                              selectClass ${module}
#                              where ${module} is the full path name
# Classes:              
#                   class
#                   ${cl}  
#     class    is used for setting fonts, colors etc.
#     ${cl}   (i.e. the class name)is bound with key 1 to selecting the class
#              and with key 2 to the documentation
##########################################################################    
proc initClassList { } {
    global gAllClasses
    global gFontSizes
    global gClassList
    global gSubtypeColor
    global gIncludeColor
    global gModuleColor
    global gModToClasses
    global gClassListModuleFont
    global gClassListClassFont

    set classesRet [tkkit_cb allClasses]
    set unsorted [lindex ${classesRet} 0]
    # puts "Unsorted:$unsorted"
    # First go through and just fill in empty entries into gModToClasses, or it
    # won't work later on
    # gModToClasses is a mapping from modulename -> list of classes
    set commandLineModuleName "CommandLine.module"
    foreach class $unsorted {
	set mod [lindex $class 3]
	if { ${mod} == "" } {
	    set gModToClasses($commandLineModuleName) {}
	} else {
	    set gModToClasses($mod) {}
	}
	
    }
    # Put each class under its appropriate module in gModToClasses 
    # (should really be module table)
    foreach class $unsorted {
	set cnm [lindex $class 0]
	set fileloc [lindex $class 1]
	set mod [lindex $class 3]
	if { $mod == "" } {
	    set mod $commandLineModuleName
	    debugPuts "Command line module ${cnm}"
	}
	set clist $gModToClasses($mod)
	# puts "$cnm $fileloc $mod Classes:$clist"
	if { $clist == {} } {
	    set gModToClasses($mod) [list $cnm ]
	} else {
	    set clist2 [concat $clist [list $cnm]]
	    set gModToClasses($mod) $clist2
	}
    }
    set gAllClasses { }
    set unsortedModules [array names gModToClasses]
    debugPuts "Going through unsorted ... $unsortedModules"
    # Note that this sorts the _full_ path names of the modules, 
    # not just the module names
    set sortedModules [lsort -increasing -ascii $unsortedModules ]
    debugPuts "Going through modules ... $sortedModules"
    set line 1
    # Go through each of the modules
    foreach module $sortedModules  {
	# Tag each module line with "module" and with the name of the module
	debugPuts "Next module $module"
	set textNameForModule [getModuleName $module]
	set gAllClasses [concat $gAllClasses [list ${module}]]
	# Adding the module text
	${gClassList}.l insert end "${textNameForModule}                      "

	# Taggging the text with "module" and the text name of the module
	${gClassList}.l tag add module \
		"$line.0" "$line.0+20 chars"
	${gClassList}.l tag add ${textNameForModule} \
		"$line.0" "$line.0+20 chars"

	# Configuring the module text with properties
	${gClassList}.l tag configure ${textNameForModule} \
		-foreground ${gModuleColor} \
		-font ${gClassListModuleFont}

	# Set up button binding
	${gClassList}.l tag bind ${textNameForModule} <1> \
		"selectClass {${module}} $line false"

	${gClassList}.l insert end "\n"
	incr line
	set classes [lsort -increasing -ascii $gModToClasses($module)]
	# For all the classes within the current module, add them to
	# the class list.
	# Bind the mouse buttons to select the appropriate file location
	# Set the colors appropriately
	foreach cl $classes { 
	    # Insert the text
	    ${gClassList}.l insert end "   ${cl}                      " 
	    # Add the class tags
	    ${gClassList}.l tag add ${cl} "$line.0" "$line.0 lineend"
	    ${gClassList}.l tag add class "$line.0" "$line.0 lineend"
	    ${gClassList}.l insert end "\n"

	    # Set the button bindings
	    ${gClassList}.l tag bind ${cl} <1> "selectClass {${cl}} $line false"
	    ${gClassList}.l tag bind ${cl} <2> "selectClass {${cl}} $line true"
	    set c [string index ${cl} 0]
	    if { $c == "\$" } {
		${gClassList}.l tag configure ${cl} \
			-foreground $gSubtypeColor -font ${gClassListClassFont}
	    } else {
		${gClassList}.l tag configure ${cl} \
			-foreground $gIncludeColor -font ${gClassListClassFont}
	    }
	    incr line
	}
	set gAllClasses [concat $gAllClasses $classes]
    }
    debugPuts "Done with modules ..."
    
    ${gClassList}.l configure -state disabled
    initOtherClassLists 
}

proc initOtherClassLists { } {
    global gAllClasses
    global gCClasses
    global gTestClasses

    foreach c ${gAllClasses} {
	set isC [string first "C_" ${c}]
	if { ${isC} == 0 } {
	    set gCClasses [concat ${gCClasses} [list ${c}]]
	}
	set isTest [string first "TEST_" ${c}]
	if { ${isTest} == 0 } {
	    set gTestClasses [concat ${gTestClasses} [list ${c}]]
	}
    }
    debugPuts "C_ ${gCClasses}"
    debugPuts "TEST ${gTestClasses}"
}
# Respond to a click on the list of classes.  Show the class  
# defintion, and the list of features.  If in narrowed view,
# have the graph show currently reachable classes
# if commentDef is true then show the comment definition of the class
debugPuts "before selectClass...."
proc selectClass { nm linenum commentDef } {
    global gClassList
    global gGraphFrame
    global gCurClassName
    global gFeatureListTitleColor
    global gFeatureListTitleRev

    inform "Selecting $nm"
    ${gClassList}.l config -cursor watch

    ${gClassList}.l tag delete current
    ${gClassList}.l tag add current "${linenum}.0 linestart" "${linenum}.0 lineend"
    ${gClassList}.l tag configure current -foreground $gFeatureListTitleRev -background $gFeatureListTitleColor
    
    set gCurFeatures {} 
    updateFeatureList
    update 

    set gCurClassName ${nm}
    tkkit_cb setCurClassName $nm
    ${gGraphFrame}.c addtag selected withtag  node${gCurClassName}
    if { [isModule $nm] } {
	gotoFile $nm $nm 1 true
    } else {
	if {$commentDef } {
	    showClassDocs
	} else {
	    showClassDef 
	}
    }
    inform "Initializing graph for $nm..."
    update
    initGraph
    inform "Displaying $nm"
    update
    ${gClassList}.l config -cursor arrow

}

debugPuts "Before ShowClassDef..."

# Show the list of features that constitute the class
# in the routine window and the class definition in the text window
proc showClassDef { } {
    global gCurFeatures
    global gCurClassName

    set cname ${gCurClassName}
    if { ${cname} == "" } {
	return
    }
    set cdef [tkkit_cb getClassInfo ${cname}] 
    set fname [lindex ${cdef} 0]
    set lineoffset [lindex ${cdef} 1] 
    gotoFile ${cname} ${fname} ${lineoffset} true
    set features [lindex $cdef 2]
    # puts "Class def: ${cdef}"
    set gCurFeatures [lsort -increasing -ascii $features]
    updateFeatureList
}

########################################################################
#                          FeatureList Box
########################################################################
########################################################################
# Create a list box with the following subpanes
# .l listbox itself
# .vscroll vertical scroller
# .hscroll horizontal scroller
#######################################################################
debugPuts "before makeFeatureListBox...."
proc makeFeatureListBox { topName } {
    global gDeepBackgroundColor
    global gBackgroundColor
    global gFeatureListTitleColor
    global gFeatureListTitleRev
    global gFeatureListSelectionColor
    global gFeatureListSelectionRev
    global gTitleFont

    set title "Features"
    listbox ${topName}.l -yscroll "${topName}.vscroll set" -xscroll "${topName}.hscroll set" 

    ${topName}.l configure -bg $gBackgroundColor -relief raised
    ${topName}.l configure -selectbackground $gFeatureListSelectionColor
    ${topName}.l configure -selectforeground $gFeatureListSelectionRev

    scrollbar ${topName}.vscroll -bg $gDeepBackgroundColor -command   \
	    "${topName}.l yview"
    scrollbar ${topName}.hscroll -bg $gDeepBackgroundColor -orient horizontal -command  \
	    "${topName}.l xview"
    pack ${topName}.vscroll -side right -fill y
    pack ${topName}.hscroll -side bottom -fill x
    if {${title} != ""} {
	label ${topName}.title -text ${title} -bg $gFeatureListTitleColor -fg $gFeatureListTitleRev -relief raised -font ${gTitleFont}
	pack ${topName}.title -fill x
    }
    pack ${topName}.l -expand yes -fill both
    tk_listboxSingleSelect ${topName}.l
}

# To display routine information #
# When the user clicks on a particular routine, display the associated
# definition in the text window
debugPuts "before showRoutineDef...."
proc showRoutineDef { routListFrame x y } {
    global gCurFeatures
    global gCurClassName

    set ind [${routListFrame}.l nearest $y]
    ${routListFrame}.l select clear
    ${routListFrame}.l select from $ind
    ${routListFrame} config -cursor watch
    set sel [${routListFrame}.l curselection]
    if { [llength ${sel}] > 1 } {
	set sel [lindex ${sel} 0]
    }
    if { $sel == ""} { return "" }
    set rout [lindex ${gCurFeatures} $sel]
    set rsig  [lindex ${rout} 0]	
    set rname [routNameFromSig ${rsig}]
    set fname [lindex ${rout} 1]
    set offset [lindex ${rout} 2]
    inform "Selected routine $rsig in file: $fname line:$offset"
    update
    gotoFile "${gCurClassName}::${rname}"  ${fname} ${offset} true
    
    ${routListFrame} config -cursor arrow
}

debugPuts "before udpateFeatureList..."

# Update the feature list to indicate the current set of features
proc updateFeatureList { } {
    global gCurFeatures
    global gFeatureList
    global gCurClassName
    if { [isModule $gCurClassName]  != "true" } {
	# If the current class is not a module, set the title of
	# the feature list box to the classname
	${gFeatureList}.title configure -text ${gCurClassName}
    }
    ${gFeatureList}.l delete 0 end	
    foreach rout ${gCurFeatures} {
	set routName [lindex $rout 0]
	# Please the head of this file (under callbacks) re: routineType
	# Routtype holds a string indicating the properties of "routName"
	set routType [lindex $rout 3]

	# ipwrwrc  vs. nnnnnnn
	set isPriv [string index ${routType} 1]
	if { ${isPriv} == "p"} {
	    ${gFeatureList}.l insert end "\[pri\][lindex $rout 0]"
	} else {
	    ${gFeatureList}.l insert end [lindex $rout 0]
	}
    }
}

########################################################################
#                          End of Feature List box
########################################################################

########################################################################
#                          Finding class documentation
########################################################################

# Show the comments associated with a given class.
# For the head of the class and for each routine, go through the
# list of routines (gCurFeatures) and find the comments following that
# signature.  Create a piece of text for signature+routines and
# display the text in the text window.
debugPuts "before showRoutineComment..."
proc showClassDocs { } {
    global gCurFeatures
    global gCurClassName
    global gTextDisplay
    global gCurFile
    # gCurFile is set to "" to indicate no file is involved
    
    set cdef [tkkit_cb getClassInfo ${gCurClassName}] 
    inform "Obtained class info. Processing..."
    update
    
    set features [lindex $cdef 2]
    set classFileName [lindex ${cdef} 0]
    set classLineoffset [lindex ${cdef} 1] 
    # puts "Class def: ${cdef}"
    set gCurFeatures [lsort -increasing -ascii $features]
    updateFeatureList
    set classComment \
	    [getCommentFollowing $classFileName [expr $classLineoffset - 1]]
    if {[isAbstract ${gCurClassName}] } {
	set txt "type ${gCurClassName} \n$classComment \n"
    } else {
	set txt "class ${gCurClassName} \n$classComment \n"
    }
    foreach rout ${gCurFeatures} {
	set featureSig  [lindex ${rout} 0]	
	if { [string first "pri" $featureSig] >= 0 } {
	} else {
	    set featureFileName [lindex ${rout} 1]
	    set featureFileOffset [lindex ${rout} 2]
	    set com [getCommentFollowing $featureFileName \
		    [expr $featureFileOffset - 1]]
	    set txt "${txt}${featureSig} \n${com}"
	}
    }
    inform "Done processing. Loading file..."
    update
    set gCurFile ""
    loadText "Doc for ${gCurClassName}" $txt f
    gotoLine 1
    inform "Displaying documented form of class ${gCurClassName}"
}


# Return the comment that follows a feature or a class definition
# i.e. go through lines until you find one without a -- or an "include" in it.
debugPuts "before getCommentFollowing..."
proc getCommentFollowing { fileName routineLoc } {
    set fileText [getAllFile $fileName]
    set lines [split $fileText "\n"]
    set firstLine [lindex $lines $routineLoc]
    set isCom [string first "--" $firstLine]
    if { $isCom >= 0 } {
	set l [string length ${firstLine}]
	set nl [string range ${firstLine} $isCom $l]
	set retLines "${nl} \n"
    } else {
	set retLines ""
    }
    incr routineLoc
    set nextLine [string trimleft [lindex $lines $routineLoc]]
    set isCom [string first "--" $nextLine]
    set isInc [string first "include" $nextLine]
    #    puts "Next line: $nextLine $isCom $isInc"
    while { [expr (($isCom == 0) || ($isInc >= 0))] } {
	incr routineLoc
	if { $isCom == 0 } {
	    set retLines "$retLines$nextLine \n"
	}
	set nextLine [string trimleft [lindex $lines $routineLoc]]
	set isCom [string first "--" $nextLine]
	set isInc [string first "include" $nextLine]
    }
    return $retLines

}


proc getRoutineComment { fileName routineLoc } {
    set fileText [getAllFile $fileName]
    set lines [split $fileText "\n"]
    set firstLine [lindex $lines $routineLoc]
    set isCom [string first "--" $firstLine]
    # Handle lines of the form attr a: INT --- 
    if { $isCom >= 0 } {
	set l [string length ${firstLine}]
	set nl [string range ${firstLine} $isCom $l]
	set retLines "${nl} \n"
    } else {
	set retLines ""
    }
    # Otherwise keep looking till you find an "is"?
    # This is dift for attributes!... not yet working...
    incr routineLoc
    set nextLine [string trimleft [lindex $lines $routineLoc]]
    set isCom [string first "--" $nextLine]
    #    puts "Next line: $nextLine $isCom $isInc"
    while { [expr (($isCom == 0) || ($isInc >= 0))] } {
	incr routineLoc
	if { $isCom == 0 } {
	    set retLines "$retLines$nextLine \n"
	}
	# First trim left space
	set nextLine [string trimleft [lindex $lines $routineLoc]]
	# A comment line only if -- begins the line
	set isCom [string first "--" $nextLine]
    }
    return $retLines
}

##################### Set up bindings ####################
# For the moving to work, 
# When the cursor enters a node, it should be highlighted
# When the first button is clicked, it should be selected and readied for
# the move nodeBeginMove
# As the node moves it should be redrawn (nodeContinueMove)
# When the first button is released, the node should be repositioned and
# all arrows redrawn. (nodeEndMove)
# ######################################################################
debugPuts "before setBindings..."
proc setBindings { } {
    global gFeatureList
    global gGraphFrame
    global gTextDisplay
    global gClassList
    global gMenuFrame

    debugPuts "Making bindings..."
    ${gGraphFrame}.c bind node <1> "selectGraphNode  %x %y"
    ${gGraphFrame}.c bind node <2> "nodeBeginMove  ${gGraphFrame} %x %y"
    ${gGraphFrame}.c bind node <B2-Motion> \
	    "nodeContinueMove ${gGraphFrame} %x %y "
    ${gGraphFrame}.c bind node <B2-ButtonRelease> \
	    "nodeEndMove ${gGraphFrame} %x %y "

    # Selecting routines
    bind ${gFeatureList}.l  <B1-ButtonRelease> "showRoutineDef ${gFeatureList} %x %y "
    bind ${gClassList}  <Any-Enter> {inform "LEFT button: Show class definition MIDDLE button: Show class documentation"}
    bind ${gFeatureList}  <Any-Enter> {inform "LEFT button: Show feature definition"}
    bind ${gGraphFrame}  <Any-Enter> {inform "LEFT button: Select class   MIDDLE button drag: Move class"}
    bind ${gTextDisplay}  <Any-Enter> {inform "LEFT button: Start editing"}
    bind ${gMenuFrame}  <Any-Enter> {inform "LEFT button: Select Menu MIDDLE button: tear off menu (click on menu bar to undo)"}
    debugPuts "Created bindings."

}

debugPuts "before checkForColor..."
proc checkForColor { } {
    global gIsColor
    set cinfo [winfo screenvisual .]
    puts $cinfo
    
}

################################## Running #########################
# Multiple window configuration by MBK

debugPuts "before configure..."
proc configure { } {
    global gConfigMode
    if {${gConfigMode} == "MultiWindow"} {
	configureMultiWindow
    } else {
	configureSingleWindow
    }
}
# Multiple window configuration by MBK
proc configureMultiWindow { } {
    global gBackgroundColor
    global gGraphFrame
    global gTextDisplay
    global gFeatureList
    global gClassList
    global gInformFrame
    global gMenuFrame
    global gDeepBackgroundColor

    
    wm minsize . 1 1

    debugPuts "Making graph ...."
    . configure -background $gDeepBackgroundColor

    frame .top 
    frame .bottom 
    .bottom configure -background $gDeepBackgroundColor
    frame .bottom.upperbottom 
    frame .bottom.lowerbottom -relief flat 

    toplevel .toplevelgraph
    set gGraphFrame .toplevelgraph.gGraphFrame
    frame ${gGraphFrame} -relief flat -borderwidth 5 -bg $gDeepBackgroundColor
    makeCanvas ${gGraphFrame} 200c 50c 200c 50c  

    debugPuts "Making gFeatureList ...."
    set gFeatureList .bottom.upperbottom.gFeatureList
    frame ${gFeatureList} -relief flat -borderwidth 5 -bg $gDeepBackgroundColor
    makeFeatureListBox ${gFeatureList} 


    debugPuts "Making gClassList2 ...."
    set gClassList .bottom.upperbottom.gClassList
    frame ${gClassList} -relief flat -borderwidth 5 -bg $gDeepBackgroundColor
    makeClassListBox ${gClassList} 

    debugPuts "Making text display ...."
    set gTextDisplay .top.gTextDisplay
    frame ${gTextDisplay} -borderwidth 5 -relief flat -bg $gDeepBackgroundColor
    makeText ${gTextDisplay}  

    set gTopFrame .top.gTopFrame
    frame $gTopFrame -bg $gDeepBackgroundColor

    debugPuts "Making menubar ...."
    set  gMenuFrame .top.gTopFrame.gMenuFrame
    frame ${gMenuFrame} -bg $gDeepBackgroundColor -relief flat -bd 2
    makeMenuBar ${gMenuFrame}
    
    set gButtonFrame .top.gTopFrame.gButtonFrame
    frame  ${gButtonFrame} -bg $gDeepBackgroundColor
    button ${gButtonFrame}.quit -text "Quit" \
	    -command "saveTextIfNecc; destroy ." -relief raised \
	    -bg ${gBackgroundColor}

    button ${gButtonFrame}.help -text "Help" -command "showHelp help"
    # pack ${gButtonFrame}.help ${gButtonFrame}.quit -side left -padx 5
    pack ${gButtonFrame}.quit -side left -padx 5

    pack ${gButtonFrame} -side right
    pack ${gMenuFrame} -side left

    debugPuts "Making inform ..."
    set gInformFrame .bottom.lowerbottom.gInformFrame
    frame ${gInformFrame} 
    makeInform ${gInformFrame}
    
    set top 0.5
    set bottom 0.9
    debugPuts "Packing ...."
    place .bottom -relx 0.0 -rely 0.5 -relh 0.5 -relw 1.0
    place .top -relx 0.0 -rely 0.0 -relh 0.5 -relw 1.0
    #    pack .top -side top -expand true
    #    pack .bottom -side bottom -expand true 
    #    pack .top .bottom -fill y

    pack ${gTopFrame} -side top -fill x 

    pack ${gGraphFrame} -side top -fill y -expand yes

    pack ${gInformFrame} -side bottom -expand yes -fill x
    pack ${gClassList} -side left -expand yes -fill both 
    pack ${gFeatureList} -side right -expand yes -fill both 
    pack ${gTextDisplay} -side right -expand yes -fill both 

    pack .bottom.upperbottom -side top -expand yes -fill both
    pack .bottom.lowerbottom -side bottom -expand no -fill x 
    # took off -expand yes 


    debugPuts "Done packing"

    wm geometry . 900x800
    set browsername "Sather Browser"
    wm geometry .toplevelgraph 900x400
    wm title .toplevelgraph "Graph for ${browsername}"
    wm title . ${browsername}    
    debugPuts "Done with wm"

}



# Multiple window configuration by MBK
proc configureSingleWindow { } {
    global gBackgroundColor
    global gGraphFrame
    global gTextDisplay
    global gFeatureList
    global gClassList
    global gInformFrame
    global gMenuFrame
    global gDeepBackgroundColor

    
    wm minsize . 1 1

    debugPuts "Making graph ...."
    . configure -background $gDeepBackgroundColor

    frame .top 
    frame .bottom 
    .bottom configure -background $gDeepBackgroundColor
    frame .bottom.upperbottom 
    frame .bottom.lowerbottom -relief flat 

    set gGraphFrame .top.gGraphFrame
    frame ${gGraphFrame} -relief flat -borderwidth 5 -bg $gDeepBackgroundColor
    makeCanvas ${gGraphFrame} 200c 50c 200c 50c  

    debugPuts "Making gFeatureList ...."
    set gFeatureList .bottom.upperbottom.gFeatureList
    frame ${gFeatureList} -relief flat -borderwidth 5 -bg $gDeepBackgroundColor
    makeFeatureListBox ${gFeatureList} 

    debugPuts "Making gClassList2 ...."
    set gClassList .bottom.upperbottom.gClassList
    frame ${gClassList} -relief flat -borderwidth 5 -bg $gDeepBackgroundColor
    makeClassListBox ${gClassList} 

    debugPuts "Making text display ...."
    set gTextDisplay .bottom.upperbottom.gTextDisplay
    frame ${gTextDisplay} -borderwidth 5 -relief flat -bg $gDeepBackgroundColor
    makeText ${gTextDisplay} 

    set gTopFrame .top.gTopFrame
    frame $gTopFrame -bg $gDeepBackgroundColor

    debugPuts "Making menubar ...."
    set  gMenuFrame .top.gTopFrame.gMenuFrame
    frame ${gMenuFrame} -bg $gDeepBackgroundColor -relief flat -bd 2
    makeMenuBar ${gMenuFrame}
    
    set gButtonFrame .top.gTopFrame.gButtonFrame
    frame  ${gButtonFrame} -bg $gDeepBackgroundColor
    button ${gButtonFrame}.quit -text "Quit" \
	    -command "saveTextIfNecc; destroy ." -relief raised \
	    -bg ${gBackgroundColor}

    button ${gButtonFrame}.help -text "Help" -command "showHelp help"
    # pack ${gButtonFrame}.help ${gButtonFrame}.quit -side left -padx 5
    pack ${gButtonFrame}.quit -side left -padx 5

    pack ${gButtonFrame} -side right
    pack ${gMenuFrame} -side left

    debugPuts "Making inform ..."
    set gInformFrame .bottom.lowerbottom.gInformFrame
    frame ${gInformFrame} 
    makeInform ${gInformFrame}
    
    set top 0.5
    set bottom 0.9
    debugPuts "Packing ...."
    place .bottom -relx 0.0 -rely 0.5 -relh 0.5 -relw 1.0
    place .top -relx 0.0 -rely 0.0 -relh 0.5 -relw 1.0
    #    pack .top -side top -expand true
    #    pack .bottom -side bottom -expand true 
    #    pack .top .bottom -fill y

    pack ${gTopFrame} -side top -fill x 

    pack ${gGraphFrame} -side top -fill y -expand yes

    pack ${gInformFrame} -side bottom -expand yes -fill x
    pack ${gClassList} -side left -expand yes -fill both 
    pack ${gFeatureList} -side left -expand yes -fill both 
    pack ${gTextDisplay} -side right -expand yes -fill both 

    pack .bottom.upperbottom -side top -expand yes -fill both
    pack .bottom.lowerbottom -side bottom -expand no -fill x 
    # took off -expand yes 


    debugPuts "Done packing"

    wm geometry . 900x800
    set browsername "Sather Browser"
    wm title . ${browsername}    
    debugPuts "Done with wm"

}

proc run { } {
    debugPuts "Running..."
    global gGraphFrame
    global gFeatureList
    global gTextDisplay

    initFonts
    debugPuts "Initialized fonts"

    configure

    debugPuts "Initializing class list...."
    initClassList
 
    debugPuts "Setting bindings...."
    setBindings 

    afterInitHook

    inform "Initializing graph...."
}

######################## USER CUSTOMIZATION HOOKS ###################
proc afterInitHook { } {
    # User can redefine this in .bs.tcl to do something useful
    # At this point, the browser tables like gModToClasses have been setup
    # Example usage:
    # Defining new customized menu items
    # Printing out some error messages, perhaps
}

if [file exists "~/.bs.tcl" ] {
    # Sourcing of .bs.tcl - user's customization file
    # This file can:
    # - define new procedures
    # - redefine existing procedures which will over-ride the current procs
    # - In particular, redefine afterInitHook to add new menu items etc.
    # The recommended way to make changes is by redefining the procedures
    # you wish to change in a minimal .bs.tcl and submitting that to me
    puts "Sourcing ~/.bs.tcl"
    source "~/.bs.tcl"
}

#####################################################################

run 

# puts "Error info:$errorInfo"



# $Log: browser.tcl,v $
# Revision 1.28  1995/07/19 23:43:58  gomes
#  * Added new initGraph which permits a variable size graph display
#  * Added menu option for variable size graph
#  * Removed globals controlling look/thickness of edges
#
# Revision 1.27  1995/07/19 21:51:09  gomes
#  * Changed to use italics and lucida fonts...
#
# Revision 1.26  1995/07/19 21:39:59  gomes
#  Implemented suggestion by Claudio, with monochrome highlighting.
#   * Eliminated "selectedFont"
#   * Installed boldFontSizes to represent abstract class namess
#   * Created new bitmaps to draw stippled lines for subtype and include edges
#   * Changed gIsColor to be set automatically.
#   * Collapsed procedure curFont into setCanvasFont, which checks for color
#
# Revision 1.25  1995/07/10 20:43:05  gomes
#  * Better Documentation of routine type.
#  * Added HTML documentation to help menu
#
# Revision 1.24  1995/07/08 06:13:25  gomes
# * Added dump state.
#
# Revision 1.23  1995/07/08 06:03:49  gomes
#  * Made changes to handle printing of (pri) at tcl level
#
# Revision 1.22  1995/06/29 05:58:56  gomes
# *** empty log message ***
#
# Revision 1.21  1995/06/29 05:56:55  gomes
#  * Incoporated Yon Uriarte's changes for showing/hiding C_ and TEST_ classes
#
# Revision 1.20  1995/06/29 05:48:21  gomes
#  * Fixed problem with updating color of text window when file is
#    the same
#
# Revision 1.19  1995/06/29 05:12:16  gomes
# Working version after incorporating ukl - fiddling with colors now
#
# Revision 1.18  1995/06/23 16:18:48  gomes
#  * Improved documentation of tags, removed HTML stuff into .bs.tcl
#
# Revision 1.17  1995/06/23 15:48:03  gomes
#  * Put the dump html stuff in my .bs.tcl
#
# Revision 1.16  1995/06/23 15:45:49  gomes
#  * Added a couple of user customizations hooks that will allow the
#    user to define new menu items (added the top level menu item "Custom")
#
# Revision 1.15  1995/06/23 14:45:04  gomes
#  * Made some changes for the b/w version to make it look less grim
#    Changed inform foreground color (added global gInformColor)
#    Changed relief for menu buttons (was a mixture - now flat)
#    Changed borderwidth on canvas
#
# Revision 1.14  1995/06/23 05:07:47  gomes
#  * Changed the point at which .bs.tcl is sourced so that procedures
#    may be redefined as well. Recommended way to make changes...
#
# Revision 1.13  1995/06/23 04:51:49  gomes
#  * Added sourcing of a configuration file, .bs.tcl so that the user
#    can set up a default configuration
#  * Added the old single window format in addition to MBK's multi-window
#    format
#  * Fixed bugs with b/w colors (reverse colors were missing)
#  * Found that I could not make things look less grim in b/w - highlighting
#    colors are tied to the graph viewing colors
#
# Revision 1.12  1995/06/23 04:00:46  gomes
#  * Changed gHighlightComments default value to 0 - highlighting comments
#    is very slow and gives people a bad first impression of browser speed in
#    general.
#
# Revision 1.11  1995/06/23 03:28:59  gomes
#  * Minor formatting
#
# Revision 1.10  1995/06/23 03:27:38  gomes
#  ***Incorporated MBK's changes:
#  * introduces many new globals for colors.
#  * consistent colors for the selected class, highlighted class, graph
# 	selected class etc.
#  * Event flushing so that inform works correctly
#  * Changed the layout so that the graph window is separate
#  *** Incorporated Emacs usage:
#  * Added call to emacsclient
#  * Inform message to tell user what to do (M-x server-start)
#  * Changed comments and organization of global variables
#  ***HTML stuff not in the current version (temporary)
#
# Revision 1.4  1995/05/06  01:06:43  gomes
#  * Changed documentation mode to print out documentation with signature
#    provided by sather. Won't work with attributes which are documented
#    on the same line.
#  * Split gotoText into gotoText (goto a line in a string) and
#    gotoFile
#  * Changed getCommentsFollowing to not always get the first line visited
#
# Revision 1.3  1995/05/03  01:44:22  gomes
#  * Almost ready for release. Still don't know how to tell the difference
#    between color and monochrome
#  * Added the ability to view the graph horizontally or vertically.
#  * Fixed bug involving labelling of modules
#  * Changed module edge labels to moduleContains so as to avoid
#    confusion between module nodes and module edges
#  * Added some stuff (including variable gIsColor) to determine
#    whether to change arrow styles or colors
#  * Added function checkForColor which does not work to try and
#    set gIsColor automatically
#
# Revision 1.2  1995/05/03  00:12:51  gomes
#  * Fixing colors.... almost working.
#  * May try next to change around x and y orientation of graph
# Revision 1.1  1995/05/01  22:38:06  gomes
#  * Initial revision (new copy of browser.tcl)
#
# Revision 1.2  1995/04/15  04:15:36  gomes
#  * Changing to use a text widget for the class list box
#
# Revision 1.6  1995/04/14  00:54:05  gomes
#  * Ability to mark the selected node with a square
#  * Better resizing/centering of classes
#  * Some new fonts
#  * Marking of private routines
#  * Added quit/help buttons (not quite working)
#  * Temporarily removed postscript stuff
#
# Revision 1.5  1995/04/07  20:32:27  gomes
#  * Eliminating singleton box. Focussing on restricted mode
#
# Revision 1.4  1995/04/07  00:29:51  gomes
#  * Changed all globals to have the g prefix (window names cannot begin with
#    a capital)
#
# Revision 1.3  1995/04/06  19:04:09  gomes
#  * Getting ready for release

# Initial work:
# Apr 31 95: Incoporated graph layout code of davidS
# Apr 31 95: Integrated with the new compiler and tkkit
# Jun 1: Read in whole file and set pointer to correct position
#        Must document globals and remove the unnecessary ones
#         Separate list of singleton nodes at the left side of the graph
#
