;;; sather.el -- Emacs 19 mode for editing Sather 1.0 programs.

;;; Copyright (C) International Computer Science Institute, 1990,
;;; 1994, 1995

;;; Authors:         1994-1995 Kevin K. Lewis <lewikk@aud.alcatel.com>
;;;                  1990-1994 Stephen M. Omohundro <om@icsi.berkeley.edu>
;;; Maintainer:      <lewikk@aud.alcatel.com>
;;; Version:         1.36
;;; Last Modified:   1995/07/17 00:49:23
;;; Bugs, Comments:  Please read the documentation below on sending
;;;                  bug reports and making comments.

;;; COPYRIGHT NOTICE: This code is provided WITHOUT ANY WARRANTY and
;;; is subject to the terms of the SATHER LIBRARY GENERAL PUBLIC
;;; LICENSE contained in the file: Sather/Doc/License of the Sather
;;; distribution.  The license is also available from ICSI, 1947
;;; Center St., Suite 600, Berkeley CA 94704, USA.

;;; Portions of this code (please see the notes in the code) are
;;; subject to the GNU General Public License.

;;; Commentary:

;; Major mode for editing Sather programs.  Sather-mode is based on:
;; an earlier Eiffel mode including modifications made by Bob Weiner
;; of Motorola; pieces eiffel3.el, distributed by Tower Technology
;; Corporation; pieces of cc-mode, by Barry A. Warsaw; and various and
;; sundry ideas found in the dark crevices of the Elisp world.
;;
;; The following statements, placed in a `.emacs' file or
;; `site-start.el', will cause this file to be autoloaded, and
;; sather-mode invoked when visiting `*.sa' files:
;;
;;	(autoload 'sather-mode "sather" "Sather Mode" t)
;;	(setq auto-mode-alist
;;	      (append '(("\\.sa$" . sather-mode)
;;			) auto-mode-alist))
;;
;; Sather-mode is developed and tested using GNU Emacs 19.xx on Linux
;; and SunOS systems.  Thanks to Peter Arius for porting it to XEmacs.
;; It has not been tested on any non-*ix system.  If someone would
;; like to volunteer to do this, I'd be happy to include any necessary
;; modifications.  It would be really cool if the same someone could
;; provide those modifications.
;;
;; NOTE FOR XEMACS USERS: `sather.el' requires the easymenu package,
;; which is not part of the XEmacs distribution (and probably won't be
;; in the future). As not all versions of `easymenu.el' are suitable
;; for XEmacs, you should get `auc-menu.el' from the AUC-TeX package
;; and install it as `easymenu.el' (resp. elc).  auc-menu is available
;; at any elisp-archive mirror:
;;
;; ;; LCD Archive Entry:
;; ;; auc-menu|Per Abrahamsen|abraham@iesd.auc.dk|
;; ;; Easy menu support for GNU Emacs and XEmacs|
;; ;; 1995/01/24 22:52:29|5.8|
;;
;; BUGS AND COMMENTS: Submit bug reports and comments for sather-mode
;; using C-c C-b m (or M-x sather-submit-bug-report).  Please describe
;; what you did to cause the problem, and what the problem is.  A
;; small amount of code that duplicates the problem is helpful.
;;
;; HIDING: Sather-mode supports selective display of Sather classes
;; and routines.  Selective display allows classes and routines to be
;; "hidden" so that code can be read with less clutter, and browsed
;; more easily.  Caveat: This all works most of the time, but there
;; might still be some cases in which the mode gets lost and is unable
;; to hide things properly.  To unhide everything without context, use
;; `sather-show-all'.
;;
;; INDENTING: There are several variables that control indenting in
;; sather-mode.  The default values of these variables are
;; recommended, but just by me (the folks working on the compiler use
;; different, inferior, settings ;-).  A brief description of these
;; variables follows.
;;
;; sather-indent: The basic indentation offset.  Routines will be
;; indented by this value from the beginning of the line.  The bodies
;; of conditionals, loops, etc, are indented from the surrounding
;; block by this amount.
;;
;; sather-continued-line-indent: This is the extra indentation for
;; lines that continue a statement (but not a clause).
;;
;; sather-auto-indent: If this variable is non-nil (t), the new, empty
;; line is indented following entry of a <newline> character.
;;
;; sather-auto-newline: If this variable is non-nil, a newline is
;; automatically inserted with the entry of a semi-colon (`;').
;;
;; sather-newline-prefix-indent: If this variable is non-nil, the
;; _current_ line is indented when a <newline> character is inserted,
;; prior to point being advanced to the next line.  This is convenient
;; for `else' and `when' clauses, that have no syntax to aid
;; indentation.  It also allows you to write code without ever hitting
;; the tab key (like auto-indent, but no nasty trailing white-space).
;;
;; sather-comment-col: The column to which comments are indented with
;; they're to the right of code.
;;
;; FEATURES: Sather-mode provides some additional "features" (for lack
;; of a less confusing term), some of which involve the indentation
;; variables above, the rest are simply cool functionality that some
;; might find useful.  Active features are displayed in the "feature
;; list", which is a list of flags shown after a "/", which follows
;; the mode name "Sather" (eg, it might look like "Sather/hp", to
;; indicate that hungry-delete and prefix-indent are active).
;;
;; Features may be turned on using the variable associated with that
;; feature, or (better) by calling the associated toggle function.
;; The function names are listed after the feature "title".
;;
;; Hungry delete (sather-toggle-hungry-state): If active, this feature
;; causes deleting to consume all whitespace that precedes point.
;; This feature was stolen from cc-mode.
;;
;; Auto-indent (sather-toggle-indent-state): Described above under
;; `sather-auto-indent'.
;;
;; Auto-newline (sather-toggle-newline-state): Described above under
;; 'sather-auto-newline'.
;;
;; Prefix Indenting (sather-toggle-prefix-state): Described above
;; under `sather-newline-prefix-indent'.
;;
;; HIGHLIGHTING: If hl319 (recommended), or hilit19 is loaded,
;; sather-mode will define patters for highlighting Sather source
;; code.  Some additional faces are used for Sather code.  They are:
;; `builtin', `iterator', and `sugar'.  They can be defined using the
;; `hilit-user-face-table' (hl319), or with `hilit-translate'
;; (hilit19).  A basic setup in your startup file might look like this
;; for hl319:
;;
;;	(cond (window-system
;;	       (add-hook 'find-file-hooks    'hilit-install-line-hooks)
;;	       (setq hilit-mode-enable-list  '(not text-mode)
;;	             hilit-background-mode   'dark)
;;	       (setq hilit-user-face-table
;;	        ;; face         light           dark                 mono
;;		'((builtin      white-bold      white-bold           nil)
;;		  (exception    red             red                  nil)
;;		  (iterator     green-bold      green-bold           nil)
;;	          (sugar        Plum1           Plum1                nil)
;;		  ))
;;	        (require 'hl319)
;;	        ))
;;
;; Or like this for hilit19:
;;
;;	(cond (window-system
;;	       (setq hilit-mode-enable-list  '(not text-mode)
;;	             hilit-background-mode   'dark
;;	             hilit-inhibit-hooks     nil
;;	             hilit-inhibit-rebinding nil)
;;	       (require 'hilit19)
;;	       (hilit-translate sugar    'Plum1
;;				builtin  'white-bold
;;				iterator 'green-bold)
;;	       ))
;;
;; You can speed up highlighting by setting some colors to `nil',
;; which tells the highlight package not to highlight that pattern.
;; For example, I use:
;;
;;	       (builtin      white-bold      white-bold           nil)
;;	       (exception    nil             nil                  nil)
;;	       (iterator     green-bold      green-bold           nil)
;;	       (sugar        nil             nil                  nil)
;;
;; You'll find adding the following lines to your `sather-mode-hook'
;; will cause highlighting patterns to work better:
;;
;;	(modify-syntax-entry ?_ "w")
;;	(modify-syntax-entry ?! "w")
;;	(modify-syntax-entry ?# "w")
;;	(modify-syntax-entry ?$ "w")
;;
;; There is also support for font-lock mode, made better by
;; contributions from Peter Stoehr.  A basic setup for using
;; font-lock-mode with sather-mode might look like this:
;;
;;	(cond (window-system
;;	       (setq font-lock-function-name-face 'underline)
;;	       (require 'font-lock)
;;	       ))
;;
;; Then add this line to your sather-mode-hook:
;;
;;	(font-lock-mode 1)
;;
;; You should read the documenation for the package you use so that
;; you can optimally configure it for your environment.  Further
;; highlighting contributions for any of the supported packages are
;; welcome.

;;; Code:

;;; Portability:

(defconst sather-xemacs-p (string-match "XEmacs" emacs-version)
  "Non-nil, if running under XEmacs.")

;;; Dependencies:
(require 'info)
(require 'compile)
(require 'easymenu)

;;; Keymap:
(defvar sather-mode-map nil
  "Keymap for Sather mode.")

(if sather-mode-map
    ()
  (setq sather-mode-map (make-sparse-keymap))
  (define-key sather-mode-map "\C-c\C-bc" 'sather-compiler-submit-bug-report)
  (define-key sather-mode-map "\C-c\C-bm" 'sather-mode-submit-bug-report)
  (define-key sather-mode-map "\C-c\C-c" 'compile)
  (define-key sather-mode-map "\C-c\C-k" 'kill-compilation)
  (define-key sather-mode-map "\C-c\C-nc" 'sather-new-class)
  (define-key sather-mode-map "\C-c\C-nt" 'sather-new-type)
  (define-key sather-mode-map "\C-c\C-w" 'sather-which-class)
  (define-key sather-mode-map "\C-c\C-l" 'sather-topic-lookup)
  (define-key sather-mode-map "\eq" 'sather-fill-paragraph)
  (define-key sather-mode-map "\e\C-q" 'sather-indent-sexp)

  (define-key sather-mode-map "\C-c\C-hc" 'sather-hide-current-class)
  (define-key sather-mode-map "\C-c\C-hl" 'sather-hide-classes)
  (define-key sather-mode-map "\C-c\C-hr" 'sather-hide-current-routine)
  (define-key sather-mode-map "\C-c\C-hs"
    'sather-hide-current-class-routines)
  (define-key sather-mode-map "\C-c\C-ha" 'sather-hide-routines)
  (define-key sather-mode-map "\C-c\C-sc" 'sather-show-current-class)
  (define-key sather-mode-map "\C-c\C-sl" 'sather-show-classes)
  (define-key sather-mode-map "\C-c\C-sr" 'sather-show-current-routine)
  (define-key sather-mode-map "\C-c\C-ss"
    'sather-show-current-class-routines)
  (define-key sather-mode-map "\C-c\C-sa" 'sather-show-routines)
  (define-key sather-mode-map "\C-c\C-sA" 'sather-show-all)

  (if sather-xemacs-p
      (progn
	(define-key sather-mode-map 'tab 'sather-indent-command)
	(define-key sather-mode-map '(control tab) 'sather-indent-command)
	(define-key sather-mode-map 'return 'sather-return)
	(define-key sather-mode-map 'delete 'sather-electric-delete))
    (define-key sather-mode-map "\t" 'sather-indent-command)
    (define-key sather-mode-map [C-tab] 'sather-indent-command)
    (define-key sather-mode-map "\r" 'sather-return)
    (define-key sather-mode-map "\177" 'sather-electric-delete))

  (define-key sather-mode-map "\M-;" 'sather-comment)
  (define-key sather-mode-map ";" 'sather-electric-semi)

  ;; Mouse bindings.
  (define-key sather-mode-map [S-mouse-2] 'sather-mouse-hide-something)

  ;; Add menus:

  ;; Menu to show hidden things.
  (easy-menu-define 
   s-show-menu sather-mode-map "\"Show\" menu for `sather-mode'."
   '("S/Show"
     ["Show Current Class" sather-show-current-class t]
     ["Show All Classes" sather-show-classes t]
     ["Show Current Routine" sather-show-current-routine t]
     ["Show Current Class Routines" sather-show-current-class-routines t]
     ["Show All Routines" sather-show-routines t]
     ["Show All" sather-show-all t]
     ))

  ;; Menu to hide things.
  (easy-menu-define 
   s-hide-menu sather-mode-map "\"Hide\" menu for `sather-mode'."
   '("S/Hide"
     ["Hide Current Class" sather-hide-current-class t]
     ["Hide All Classes" sather-hide-classes t]
     ["Hide Current Routine" sather-hide-current-routine t]
     ["Hide Current Class Routines" sather-hide-current-class-routines t]
     ["Hide All Routines" sather-hide-routines t]
     ))

  ;; Menu for general Sather things.
  (easy-menu-define 
   sather-menu sather-mode-map "General menu for `sather-mode'."
   '("Sather"
     ["Compile Now!" sather-compile-now t]
     ["Compile" compile t]
     ["Fill Comment Paragraph" sather-fill-paragraph t]
     ["New Type" sather-new-type t]
     ["New Class" sather-new-class t]
     ["Which Class?" sather-which-class t]
     ["Lookup Topic" sather-topic-lookup t]
     "--"
     ("Documentation"
      ["Browse Sather Manual" sather-browse-manual t]
      ["Browse Eclectic Tutorial" sather-browse-tutorial t])
     "--"
     ("Options"
      ["Auto-Indent" sather-toggle-indent-state t]
      ["Auto-Newline" sather-toggle-newline-state t]
      ["Hungry Delete Key" sather-toggle-hungry-state t]
      ["Newline-Indent" sather-toggle-prefix-state t])
     "--"
     ("Send Bug Report"
      ["Send Sather Compiler Bug Report" sather-compiler-submit-bug-report t]
      ["Send Sather Mode Bug Report" sather-mode-submit-bug-report t])
     ))
  )

(defvar sather-mode-syntax-table nil
  "Syntax table in use in Sather-mode buffers.")

(if sather-mode-syntax-table
    ()
  (let ((table (make-syntax-table)))
    (modify-syntax-entry ?\\ "\\" table)
    (modify-syntax-entry ?/ "." table)
    (modify-syntax-entry ?* "." table)
    (modify-syntax-entry ?+ "." table)

    ;; Here's your Sather comment starter.
    (modify-syntax-entry ?- ". 12" table)
    ;; All these end comments.
    (modify-syntax-entry ?\n ">" table)
    (modify-syntax-entry ?\r ">" table)
    (modify-syntax-entry ?\f ">" table)

    ;; These are parts of symbols.
    (modify-syntax-entry ?_ "_" table)
    (modify-syntax-entry ?! "_" table)
    (modify-syntax-entry ?$ "_" table)
    (modify-syntax-entry ?# "_" table)

    (modify-syntax-entry ?= "." table)
    (modify-syntax-entry ?% "." table)
    (modify-syntax-entry ?< "." table)
    (modify-syntax-entry ?> "." table)
    (modify-syntax-entry ?& "." table)
    (modify-syntax-entry ?| "." table)
    (modify-syntax-entry ?\' "\"" table)
    (setq sather-mode-syntax-table table)
    ))

(defvar sather-hilit-type nil
  "The highlighting support being used.")

(defvar sather-font-lock-keywords nil
  "Keywords for highlighting with font-lock.")

;;; Highlighting support for use with hl319, hilit19, or font-lock:
(defun sather-add-hiliting ()
  (cond

   ;; Has hl319 (hilit R3.19) been loaded?
   ((featurep 'hl319)
    (setq sather-hilit-type 'hl319)
    (hilit-set-mode-patterns
     'sather-mode
     '(("--" "$" comment)
       (:buildme:
	(string "\"[^\n\"]*\"")
	(:and:
	 (defun "^\\s-+\\(private\\s-+\\)?[a-zA-Z_][a-zA-Z0-9_]*!?")
	 "\\(\\s-*([^;)]+)\\)?[^;)]*\\s-is\\(\\s-*\\($\\|--\\|
\\)\\|\\s-\\)")
	(builtin sather-builtin-hilit)
	(constant sather-constant-numeric-hilit-a)
	(constant sather-constant-numeric-hilit-b)
	(constant sather-constant-value-hilit)
	(type sather-type-hilit)
	(iterator sather-iterator-hilit)
	(exception sather-exception-hilit)
	(sugar sather-sugar-hilit)
	(keyword sather-keyword-hilit))
       ))
    )

   ;; Else, has hilit19 (hilit R2.19) been loaded?
   ((featurep 'hilit19)
    (setq sather-hilit-type 'hilit19)
    (hilit-set-mode-patterns
     'sather-mode
     '(("--" "$" comment)
       ("\"[^\n\"]*\"" nil string)
       ("\\<\\(\\$OB\\|ARRAY\\|AREF\\|AVAL\\|BOOL\\|CHAR\\|EXT_OB\\|FLT\\|FLTD\\|FLTX\\|FLTDX\\|FLTI\\|INT\\|INTI\\|\\$REHASH\\|SAME\\|STR\\|SYS\\)\\>" nil builtin)
       ("\\<$?[A-Z][A-Z0-9_]*\\>" nil type)
       ("[-+]?\\<\\(0[box]\\)?[0-9_]+\\(\\.[0-9_]+\\([Ee][+-]?[0-9_]+\\)?\\)?[dix]?\\>" nil constant)
       ("\\('\\\\?.'\\|#\\|\\<nil\\>\\|\\<self\\>\\)" nil constant)
       ("[A-Za-z_][A-Za-z0-9_]*!" nil iterator)
       ("\\<\\(aget\\|aset\\|div\\|is_eq\\|is_geq\\|is_gt\\|is_leq\\|is_lt\\|is_neq\\|minus\\|mod\\|negate\\|not\\|plus\\|pow\\|times\\)\\>" nil sugar)
       ;; key words
       ("\\<\\(ITER\\|ROUT\\|SAME\\|a\\(nd\\|ssert\\|ttr\\)\\|break!\\|c\\(ase\\|lass\\|onst\\|reate\\)\\|do@?\\|e\\(ls\\(e\\|if\\)\\|nd\\|x\\(ception\\|ternal\\)\\)\\|false\\|fork\\|i\\([fs]\\|n\\(clude\\|itial\\|variant\\)\\)\\|lo\\(ck\\|op\\)\\|main\\|new\\|or\\|p\\(ar\\(loop\\)?\\|ost\\|r\\(e\\|ivate\\|otect\\)\\)\\|quit\\|r\\(aise\\|e\\(adonly\\|sult\\|turn\\)\\)\\|s\\(elf\\|hared\\)\\|t\\(hen\\|rue\\|ype\\(\\|case\\)\\)\\|until!\\|v\\(alue\\|oid\\)\\|wh\\(en\\|ile!\\)\\|yield\\)\\>" 1 keyword))))

   ((featurep 'font-lock)
    (setq sather-hilit-type 'font-lock)
    (if (not sather-font-lock-keywords)
	(setq font-lock-keywords
	      (list
	       '("--.*$" 0 font-lock-comment-face)
	       '("\\<\\($\\(OB\\|REHASH\\)\\|A\\(R\\(EF\\|RAY\\)\\|VAL\\)\\|BOOL\\|CHAR\\|EXT_OB\\|FLT\\(\\|[DIX]\\|DX\\)\\|INTI?\\|S\\(AME\\|TR\\|YS\\)\\)\\>"
		 1 font-lock-type-face)
	       (list
		(concat "^\\s-+\\(\\(private\\s-+\\)?"
			"\\([a-zA-Z_][a-zA-Z0-9_]*!?\\)\\)"
			"\\(\\s-*([^;)]+)\\)?[^;)]*\\s-is"
			"\\(\\s-*\\($\\|--\|
\\)\\|\\s-\\)"
			) 1 'font-lock-function-name-face t)
	       '("\\<\\(ITER\\|ROUT\\|SAME\\|a\\(nd\\|ssert\\|ttr\\)\\|break!\\|c\\(ase\\|lass\\|onst\\|reate\\)\\|do@?\\|e\\(ls\\(e\\|if\\)\\|nd\\|x\\(ception\\|ternal\\)\\)\\|false\\|fork\\|i\\([fs]\\|n\\(clude\\|itial\\|variant\\)\\)\\|lo\\(ck\\|op\\)\\|main\\|new\\|or\\|p\\(ar\\(loop\\)?\\|ost\\|r\\(e\\|ivate\\|otect\\)\\)\\|quit\\|r\\(aise\\|e\\(adonly\\|sult\\|turn\\)\\)\\|s\\(elf\\|hared\\)\\|t\\(hen\\|rue\\|ype\\(\\|case\\)\\)\\|until!\\|v\\(alue\\|oid\\)\\|wh\\(en\\|ile!\\)\\|yield\\)\\>"
		 1 font-lock-keyword-face)
	       ))
      )
    )
  ))

;;; User variables:
(defvar sather-indent 3
  "*This variable gives the indentation in sather-mode")

(defvar sather-continued-line-indent (* 3 sather-indent)
  "*Extra indentation for continued lines.
Set it to just `sather-indent' if you don't want any extra indentation.")

(defvar sather-auto-indent nil
  "*If non-nil, <return> indents for the next line.")

(defvar sather-auto-newline nil
  "*If non-nil, `;' inserts a newline.")

(defvar sather-hungry-delete-key nil
  "*If non-nil, delete-key deletes all preceding white space.")

(defvar sather-newline-prefix-indent t
  "*If non-nil, <return> indents the line before adding a newline.")

(defvar sather-comment-col 32
  "*This variable gives the desired comment column for comments to the
right of text.")

(defvar sather-compile-command "cs .module"
  "*The command and args to use to compile Sather programs.")

(defvar sather-delete-function 'backward-delete-char-untabify
  "Function called by `sather-electric-delete' when deleting a single char.")

;; Some regexps for highlighting Sather code.  User's may want to
;; redefine these if they have something that works better, or don't
;; want everything highlighted.
(defvar sather-builtin-hilit
  "\\<\\(\\$OB\\|ARRAY\\|AREF\\|AVAL\\|BOOL\\|CHAR\\|EXT_OB\\|FLT\\|FLTD\\|FLTX\\|FLTDX\\|FLTI\\|INT\\|INTI\\|\\$REHASH\\|SAME\\|STR\\|SYS\\)\\>"
  "*Highlight pattern for builtin classes.")

(defvar sather-type-hilit "\\<[#$]?[A-Z][A-Z0-9_]*\\>"
  "*Highlight pattern for Sather types and classes.")

(defvar sather-constant-numeric-hilit-a
  "[-+]?\\<[0-9_]+\\(\\.[0-9_]+\\([Ee][+-]?[0-9_]+\\)?\\)?[dix]?\\>"
  "*Highlight pattern for Sather numeric constants.")

(defvar sather-constant-numeric-hilit-b
  "\\<0[box][0-9A-Fa-f_]+\\>"
  "*Highlight pattern for Sather numeric constants.")

(defvar sather-constant-value-hilit
  "\\('\\\\?.'\\|#\\|\\<nil\\>\\|\\<self\\>\\)"
  "*Highlight pattern for Sather constant values.")

(defvar sather-iterator-hilit "[A-Za-z_][A-Za-z0-9_]*!"
  "*Highlight pattern for Sather iterators.")

(defvar sather-sugar-hilit
  "\\<\\(aget\\|aset\\|div\\|is_eq\\|is_geq\\|is_gt\\|is_leq\\|is_lt\\|is_neq\\|minus\\|mod\\|negate\\|not\\|plus\\|pow\\|times\\)\\>"
  "*Highlight pattern for Sather sugar routines.")

(defvar sather-exception-hilit "\\<\\(assert\\|exception\\|initial\\|post\\|pre\\|protect\\|raise\\|result\\)\\>"
  "*Highlight pattern for Sather keywords.")

(defvar sather-keyword-hilit "\\<\\(ITER\\|ROUT\\|SAME\\|a\\(nd\\|ssert\\|ttr\\)\\|break!\\|c\\(ase\\|lass\\|onst\\|reate\\)\\|do@?\\|e\\(ls\\(e\\|if\\)\\|nd\\|x\\(ception\\|ternal\\)\\)\\|false\\|fork\\|i\\([fs]\\|n\\(clude\\|itial\\|variant\\)\\)\\|lo\\(ck\\|op\\)\\|main\\|new\\|or\\|p\\(ar\\(loop\\)?\\|ost\\|r\\(e\\|ivate\\|otect\\)\\)\\|quit\\|r\\(aise\\|e\\(adonly\\|sult\\|turn\\)\\)\\|s\\(elf\\|hared\\)\\|t\\(hen\\|rue\\|ype\\(\\|case\\)\\)\\|until!\\|v\\(alue\\|oid\\)\\|wh\\(en\\|ile!\\)\\|yield\\)\\>"
  "*Highlight pattern for Sather keywords.")

;; Things for documenting code ...

(defvar sather-site
  (concat "@"
	  (cond ((and (boundp 'user-mail-address)
		      (string-match "@\\([a-z0-9A-Z.-]+\\)" user-mail-address))
		 (substring user-mail-address 
			    (match-beginning 1)
			    (match-end 1)))
		((boundp 'gnus-local-domain) 
		 gnus-local-domain)
		(t (system-name))))
  "*Mailing address of site where mode is being used.
Should include initial \@ sign. Use nil for none.")

(defvar sather-short-copyright
"-- Copyright (C) 1995, International Computer Science Institute\n"
"*Short copyright notice to be inserted in the header.
Should be commented and include trailing newline.  Use nil for none.")

(defvar sather-long-copyright
"-- COPYRIGHT NOTICE: This code is provided WITHOUT ANY WARRANTY
-- and is subject to the terms of the SATHER LIBRARY GENERAL PUBLIC
-- LICENSE contained in the file: Sather/Doc/License of the
-- Sather distribution. The license is also available from ICSI,
-- 1947 Center St., Suite 600, Berkeley CA 94704, USA.\n"
"*Long copyright notice to be inserted in the header.
Should be commented and have trailing newlines.  Use nil for none.")

(defvar sather-manual-name "sather"
  "*The base-name of the Info version of the Sather Manual.")

(defvar sather-tutorial-name "sather-tutorial"
  "*The base-name of the Info version of the Sather Eclectic Tutorial.")

;;; No user definable variables beyond this point:

(defvar sather-class-regexp
  "^\\s-*\\(\\(type\\)\\|\\(\\(value\\|external\\)\\s-+\\)?class\\)[ \t\n]+\\($?[A-Z][A-Z0-9_]*\\)\\(\\s-*<\\)?"
  "Matches the first line of a class definition.")

(defvar sather-class-end-regexp "^end;"
  "Matches the last line of a class definition.")

;; For some reason, using `^M' works here, but `?\r' doesn't.  If
;; anyone knows why, please tell me.  It seems to have something to do
;; with the \\(...\\) grouping.
(defconst sather-routine-regexp
  "^\\(\\s-+private\\)?\\s-+\\([A-Za-z0-9_]+!?\\)[^;]*\\s-\\(is\\)\\s-*\\($\\|--\\|
\\)"
  "Matches the first line of a class routine.")

(defconst sather-block-regexp
  "\\s-+\\(fork\\|is\\|if\\|lock\\|loop\\|par\\(loop\\)?\\|protect\\|case\\|typecase\\|end;?\\)\\($\\|\\s-\\)"
  "Matches an expression that opens or closes a block.")

(defconst sather-error-regexp
  '("^\\([^:]+\\):\\([0-9]+\\):\\([0-9]+\\):" 1 2 3)
  "The regexp to process Sather Compiler errors.")

(defvar sather-feature-string nil
  "Feature string to display as minor mode.")

(make-variable-buffer-local 'sather-auto-newline)
(make-variable-buffer-local 'sather-auto-indent)
(make-variable-buffer-local 'sather-hungry-delete-key)
(make-variable-buffer-local 'sather-newline-prefix-indent)
(make-variable-buffer-local 'sather-feature-string)

;; This adds variables where we can place the `/hinp' string to
;; display the current active feature list.
(defun sather-adjust-modeline ()
  (or (assq 'sather-feature-string minor-mode-alist)
      (setq minor-mode-alist
	    (cons '(sather-feature-string sather-feature-string)
		  minor-mode-alist)))) 

;;;###autoload
(defun sather-mode ()
  "A major mode for editing Sather code.
\\<sather-mode-map>
Comments are begun with `--'.
Paragraphs are separated by blank lines
Delete converts tabs to spaces as it moves back.
Tab anywhere on a line indents it according to Sather conventions.
\\[sather-comment] inserts and indents a comment on the line, or indents an
existing comment if there is one.

A new class/type skeleton is inserted (along with a file header if
neccessary) with:

 \\[sather-new-class] <class>
 \\[sather-new-type] <type>

Variables controlling style:
   sather-indent                Indentation of Sather statements.
   sather-continued-line-indent Indentation of continued statements.
   sather-comment-col           Goal column for inline comments.
   sather-auto-indent           Automatically indent the next line on <return>.
   sather-auto-newline          Automatically insert a newline on `;'.
   sather-hungry-delete-key     Delete key consumes all previous whitespace.
   sather-newline-prefix-indent Automatically indent line before newline.
   sather-site                  Mailing address of site for header.
   sather-short-copyright       Short copyright message for header.
   sather-long-copyright        Long copyright message for header.

Turning on Sather mode calls the value of the variable sather-mode-hook
with no args, if that value is non-nil."
  (interactive)
  (kill-all-local-variables)
  (use-local-map sather-mode-map)
  (setq major-mode 'sather-mode)
  (setq mode-name "Sather")

  ;; For class/routine hiding.
  (setq selective-display t)
  (setq selective-display-ellipses t)

  (set-syntax-table sather-mode-syntax-table)
  (make-local-variable 'indent-line-function)
  (setq indent-line-function 'sather-indent-command)
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "--+\\s-*")
  (make-local-variable 'comment-start)
  (setq comment-start "--")
  (make-local-variable 'comment-end)
  (setq comment-end "")
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "^$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'paragraph-ignore-fill-prefix)
  (setq paragraph-ignore-fill-prefix t)
  (make-local-variable 'require-final-newline)
  (setq require-final-newline t)
  (make-local-variable 'compilation-error-regexp-alist)
  (setq compilation-error-regexp-alist (list sather-error-regexp))
  (setq compile-command sather-compile-command)
  (sather-adjust-modeline)
  (s-update-modeline)
  (sather-add-hiliting)
  (easy-menu-add sather-menu)
  (easy-menu-add s-hide-menu)
  (easy-menu-add s-show-menu)
  (run-hooks 'sather-mode-hook))

;; Insert the file header at point.
(defun sather-header ()
   (let ((header (read-string "File header: "
			     (concat "-- " (buffer-name) ": "))))
    (insert header "\n"
	    "-- Author: " (user-full-name)
	    " <" (user-login-name) sather-site ">\n"
	    sather-short-copyright "-- \$Id\$\n--\n" sather-long-copyright
"-------------------------------------------------------------------\n"
"-- \$Log\$\n"
"-------------------------------------------------------------------\n"
)))

(defun sather-new-class ()
  "Insert a `class' template."
  (interactive)
  (if (not (s-empty-line-p))
      (progn
	(end-of-line)
	(newline)))
  (beginning-of-line)
  (if (s-prev-class-p) nil (sather-header))
  (let ((cname (read-string "Class: ")))
    (insert
     "class " (upcase cname) " is\n\n"
     "end; -- class " (upcase cname) "\n\n"
     "-------------------------------------------------------------------\n")
    )
  (re-search-backward "\nend")
  (sather-indent-line))

(defun sather-new-type ()
  "Insert a `type' template."
  (interactive)
  (if (not (s-empty-line-p))
      (progn (end-of-line)
	     (newline)))
  (beginning-of-line)
  (if (s-prev-class-p) nil (sather-header))
  (let ((cname (upcase (read-string "Type: "))))
    (or (string-match "^\\$" cname)
	(setq cname (concat "$" cname)))
    (insert
     "type " cname " is\n\n"
     "end; -- type " cname "\n\n"
     "-------------------------------------------------------------------\n")
    )
  (re-search-backward "\nend")
  (sather-indent-line))

(defun sather-return (arg)
  "Return and conditionally indent the new line."
  (interactive "p")
  (save-excursion
    (if (and sather-newline-prefix-indent
	     (not (s-empty-line-p)))
	(sather-indent-command)))
  (newline arg)
  (if sather-auto-indent
      (sather-indent-line)))

;; Delete function that may peform hungry delete.  Stolen from
;; cc-mode.
(defun sather-electric-delete (arg)
  "Deletes preceding character or whitespace.
If `sather-hungry-delete-key' is non-nil, then all preceding
whitespace is deleted.  However, if an ARG is supplied, or
`sather-hungry-delete-key' is nil, or point is inside a string, then
the function in `sather-delete-function' is called."
  (interactive "P")
  (if (or (not sather-hungry-delete-key)
	  arg
	  (s-in-quoted-string-p))
      (funcall sather-delete-function (prefix-numeric-value arg))
    (let ((p (point)))
      (skip-chars-backward " \t\n")
      (if (/= (point) p)
	  (delete-region (point) p)
	(funcall sather-delete-function 1))
      )))

(defun sather-electric-semi (arg)
  "Insert a semi-colon, indent, and conditionally add a newline."
  (interactive "P")
  (self-insert-command (prefix-numeric-value arg))
  (sather-indent-line)
  (if sather-auto-newline
      (progn
	(newline)
	(if sather-auto-indent
	    (sather-indent-line))
	))
  )

(defun sather-indent-line ()
  "Indent the current line as Sather code."
  (interactive)
  (save-excursion
    (beginning-of-line)
    (delete-horizontal-space)
    (indent-to (s-calc-indent)))
  (skip-chars-forward " \t"))

(defun sather-indent-command ()
  "Conditionally indent the current line and position the point."
  (interactive)
  (save-excursion
    (beginning-of-line)
    (skip-chars-forward " \t")
    (let ((in-to (s-calc-indent)))
      (if (/= (current-column) in-to)
	  (progn (delete-horizontal-space)
		 (indent-to in-to))
      )))
  (skip-chars-forward " \t"))

;; A line is one of the following:
;;    blank
;;    just a comment
;;    block-cont: starts with end, elsif, else, when, then, against
;;    block-head: ends with is, or starts with if, loop, case, typecase,
;;                assert, or protect.
;;    none of the above

;; Return the appropriate indentation for this line as an integer.
(defun s-calc-indent ()
  (cond
   ((s-empty-line-p)			;an empty line
    (s-standard-indent))
   ((s-comment-line-p)			;a comment line
    (s-comment-indent))
   ((s-starts-with-class-p) 0)
   ((s-starts-with-pre-p) (* 2 sather-indent))
   ((s-ends-with-is-p)
    (if (s-continued-line-p)
	(+ sather-continued-line-indent sather-indent)
      sather-indent))
   ((s-block-cont-p)			;begins with block-cont keyword
    (s-get-block-indent))		;indent same as block
   (t					;block-head or something else
    (s-standard-indent))
   ))

;; Do standard line indentation.
(defun s-standard-indent ()
  (cond ((s-in-sexp-p))
	((s-continued-line-p)
	 (+ sather-continued-line-indent (s-get-block-indent)))
	(t
	 (+ sather-indent (s-get-block-indent))) ;go in one from block
	))

;;; Minor features:

;; Update the modeline to include activated features.
(defun s-update-modeline ()
  (setq sather-feature-string
	(if sather-hungry-delete-key
	    "/h"
	  nil))
  (if sather-auto-indent
      (setq sather-feature-string
	    (if sather-feature-string
		(concat sather-feature-string "i")
	      "/i")))
  (if sather-auto-newline
      (setq sather-feature-string
	    (if sather-feature-string
		(concat sather-feature-string "n")
	      "/n")))
  (if sather-newline-prefix-indent
      (setq sather-feature-string
	    (if sather-feature-string
		(concat sather-feature-string "p")
	      "/p")))
  (force-mode-line-update))

;;; Stolen code:

;; This was mostly taken from cc-mode.
(defun s-calculate-state (arg prevstate)
  ;; Calculate the new state of PREVSTATE, t or nil, based on arg. If
  ;; arg is nil or zero, toggle the state. If arg is negative, turn
  ;; the state off, and if arg is positive, turn the state on.
  (if (or (not arg)
	  (zerop (setq arg (prefix-numeric-value arg))))
      (not prevstate)
    (> arg 0)))

(defun sather-toggle-indent-state (arg)
  "Toggle auto-indent feature.
Optional numeric ARG turns auto-indent on when positive, off when
negative, and toggles when zero.

When the auto-indent feature is enabled (indicated by an `i' displayed
in the modeline following the `/' feature separator) lines are
automatically indented after a <return> is entered."
  (interactive "P")
  (setq sather-auto-indent (s-calculate-state arg sather-auto-indent))
  (s-update-modeline))

(defun sather-toggle-newline-state (arg)
  "Toggle auto-newline feature.
Optional numeric ARG turns auto-newline on when positive, off when
negative, and toggles when zero.

When the auto-newline feature is enabled (indicated by an `n'
displayed in the modeline following the `/' feature separator) some
special characters, like semicolon, will cause newlines to be
automatically inserted."
  (interactive "P")
  (setq sather-auto-newline (s-calculate-state arg sather-auto-newline))
  (s-update-modeline))

(defun sather-toggle-hungry-state (arg)
  "Toggle hungry-delete-key feature.
Optional numeric ARG turns hungry-delete on when positive, off when
negative, and toggles when zero.

When the hungry-delete-key feature is enabled (indicated by an `h'
displayed in the modeline following the `/' feature separator) the
delete key deletes any and all preceding whitespace."
  (interactive "P")
  (setq sather-hungry-delete-key
	(s-calculate-state arg sather-hungry-delete-key))
  (s-update-modeline))

(defun sather-toggle-prefix-state (arg)
  "Toggle newline-prefix-indent feature.
Optional numeric ARG turns newline-prefix-indent on when positive, off
when negative, and toggles when zero.

When the newline-prefix-indent feature is enabled (indicated by a `p'
displayed in the modeline following the `/' feature separator) the
<return> key indents the current line before adding the newline."
  (interactive "P")
  (setq sather-newline-prefix-indent
	(s-calculate-state arg sather-newline-prefix-indent))
  (s-update-modeline))

;;; Handling Comments:

;; The following two functions (with slight modifications) were taken
;; from eiffel3.el 1.73, distributed by Tower Technology Corporation,
;; freely available under the GNU General Public License.  It seems to
;; work very well.

;; Prefix that starts a comment that begins a line.  Comments that are
;; not the only thing on a line return nil as their prefix.
(defun s-comment-prefix ()
  (save-excursion
    (end-of-line)
    (let ((limit (point)))
      (beginning-of-line)
      (cond ((re-search-forward "^\\s-*--|?\\s-*" limit t)
	     (buffer-substring (match-beginning 0) (match-end 0)))
	    (t nil))
      )
    ))

(defun sather-fill-paragraph ()
  "Textually fills Sather comments ala fill-paragraph."
  (interactive)
  (save-excursion
    (let ((current-point (point))
	  (last-point nil)
	  (para-begin nil)
	  (para-end   nil)
	  (fill-prefix (s-comment-prefix)))
      (if fill-prefix
	  (progn
	    (setq last-point (point))
	    (forward-line -1)
	    (end-of-line)
	    (while (and (not (= (point) last-point))
			(s-comment-prefix))
	      (setq last-point (point))
	      (forward-line -1)
	      (end-of-line))
	    (if (= (point) last-point)
		(setq para-begin (save-excursion (beginning-of-line) (point)))
	      (setq para-begin (1+ (point))))
	    (goto-char current-point)
	    (setq last-point (point))
	    (forward-line 1)
	    (end-of-line)
	    (while (and (not (= (point) last-point))
			(s-comment-prefix))
	      (setq last-point (point))
	      (forward-line 1)
	      (end-of-line))
	    (if (= (point) last-point)
		(setq para-end (point))
	      (beginning-of-line)
	      (setq para-end (point)))
	    (fill-region para-begin para-end))
        ))
    ))

;; Point to beginning of comment on line.  Assumes line contains a
;; comment.
(defun s-goto-comment-beg ()
  (beginning-of-line)
  (search-forward "--" nil t)
  (backward-char 2))

;; Return indentation for a comment line.
(defun s-comment-indent ()
  (save-excursion
    (beginning-of-line)
    (if (looking-at "^--")
	0
      (let ((in (s-get-block-indent))
	    (prev-is-blank
	     (save-excursion (and (= (forward-line -1) 0)
				  (s-empty-line-p))))
	    (prev-is-class-end
	     (save-excursion (forward-line -1)
			     (looking-at "^end")))
	    )
	(if (or (and prev-is-blank (= in 0))
					;move to prev line if there is one
		prev-is-class-end
		(not (s-move-to-prev-non-blank)))
	    0				;early comments start to the left
	  (cond ((s-comment-line-p)	;is a comment, same indentation
		 (s-current-indentation))
		(t			;otherwise indent once
		 (+ sather-indent in)))
	  ))				; let + if (or...)
      ))				; (save-...) + (if ...)
  )

(defun sather-comment ()
  "Edit a comment on the line. If one exists, reindents it and moves
to it, otherwise creates one. Gets rid of trailing blanks, puts one
space between comment header comment text, leaves point at front of
comment.  If comment is alone on a line it reindents relative to
surrounding text.  If it is before any code, it is put at the line
beginning.  Uses the variable sather-comment-col to set goal start on
lines after text."
  (interactive)
  (cond ((s-comment-line-p)             ;just a comment on the line
         (beginning-of-line)
         (delete-horizontal-space)
         (indent-to (s-comment-indent))
         (forward-char 2)(delete-horizontal-space)(insert " "))
        ((s-comment-on-line-p)          ;comment already at end of line
         (cond ((s-ends-with-end-p)     ;end comments come immediately
                (s-goto-comment-beg)(delete-horizontal-space)(insert " ")
                (forward-char 2)(delete-horizontal-space)(insert " "))
               (t
                (s-goto-comment-beg)(delete-horizontal-space)
                (if (< (current-column) sather-comment-col)
                    (indent-to sather-comment-col)
                  (insert " "))
                (forward-char 2)
		(delete-horizontal-space)
		(insert " "))))
        ((s-empty-line-p)               ;put just a comment on line
         (beginning-of-line)
         (delete-horizontal-space)
         (indent-to (s-comment-indent))
         (insert "-- "))
        ((s-ends-with-end-p)            ;end comments come immediately
         (end-of-line)(delete-horizontal-space)(insert " -- "))
        (t                              ;put comment at end of line
         (end-of-line)
         (delete-horizontal-space)
         (if (< (current-column) sather-comment-col)
             (indent-to sather-comment-col)
           (insert " "))
         (insert "-- "))
	) ; cond
  )

;;; Lots of Predicates:

;; True if there is a class definition before this one.
(defun s-prev-class-p ()
  (save-excursion
    (re-search-backward sather-class-regexp nil t)))

;; True if line starts with value class, type, external class or
;; class.
(defun s-starts-with-class-p ()
  (save-excursion
    (beginning-of-line)
    (looking-at sather-class-regexp)))

;; True if line starts with keyword end.
(defun s-starts-with-end-p ()
  (save-excursion
    (beginning-of-line)
    (looking-at "^\\s-*end")
    ))

;; True if part of extended class beginning.
(defun s-special-match-class-begin-p ()
  (save-excursion
    (let ((pt0 (point))
	  (pt1 (re-search-backward sather-class-regexp))
	  (pt2 (re-search-forward "\\s-is")))
      (if (= pt0 (- pt2 2)) 1 nil))
    ))

;; True if line starts with either pre or post.
(defun s-starts-with-pre-p ()
  (save-excursion
    (beginning-of-line)
    (looking-at "^\\s-*\\(pre\\|post\\)[ \t\n]")
    ))

;; `nil' if we're not in an sexp; otherwise, return the location of
;; the inner-most open.
(defun s-in-sexp-p ()
  (save-excursion
    (let* ((op (point))
	   (ccb (re-search-backward sather-class-regexp nil t))
	   (state (parse-partial-sexp ccb op nil nil nil nil))
	   (last-open (car (cdr state))))
      (if last-open
	  (progn
	    (goto-char last-open)
	    (1+ (+ (current-indentation)
		   (- (point) (progn (beginning-of-line)
				     (skip-chars-forward " \t")
				     (point)))))
	    )
	nil)
      )
    ))

;; True if previous line does not end a statement.
(defun s-continued-line-p ()
  (save-excursion
    (forward-line -1)
    (beginning-of-line)
    ;; Go to the previous non-blank and non-comment line.
    (while (and (or (looking-at "\\s-*--")
		    (looking-at "\\s-*$"))
		(not (= (point) (point-min))))
      (forward-line -1))
    ;; If the previous expression is not a block opener, or a
    ;; complete statement, return that the expression is
    ;; continuing.
    (let ((la (looking-at ".*\\s-\\(do\\(@[A-Za-z_][A-Za-z0-9_]+!\\)?\\|fork\\|is\\|then\\|lock\\|loop\\|par\\(loop\\)?\\|protect\\|else\\|elsif\\|end\\)\\($\\|\\s-\\)"))
	  (mb (match-beginning 1)))
      (if (or (and la
		   (not (s-in-comment-p mb))
		   (not (s-in-quoted-string-p mb)))
	      (looking-at ".*;"))
	  nil
	t))
    ))

;; True if line ends with 'end' or 'end;' and a comment.
(defun s-ends-with-end-p ()
  (save-excursion
    (beginning-of-line)
    (and (looking-at "^\\(.*\\s-+\\)?\\(end\\);?\\($\\|\\s-\\)")
	 (not (s-in-comment-p (match-beginning 2))))
    ))

;; True if current line is empty.
(defun s-empty-line-p ()
  (save-excursion
    (beginning-of-line)
    (looking-at "^\\s-*$")
    ))

;; True if current line is just a comment.
(defun s-comment-line-p ()
  (save-excursion
    (beginning-of-line)
    (looking-at "\\s-*--")
    ))

;; True if current line contains a comment.
(defun s-comment-on-line-p ()
  (save-excursion
    (beginning-of-line)
    (looking-at "[^\n]*--")
    ))

;; True if point is in a comment.
(defun s-in-comment-p (&optional arg)
  (save-excursion
    (if arg (goto-char arg))
    (and (/= (point) (point-max)) (or (looking-at "$") (forward-char 1)))
    (search-backward "--" (save-excursion (beginning-of-line)
					  (point)) t)
    ))

;; True if line continues the indentation of enclosing block.  Begins
;; with end, else, elsif, when, then, or against.
(defun s-block-cont-p ()
  (save-excursion
    (beginning-of-line)
    (looking-at
     "^\\s-*\\(do\\(@[A-Za-z_][A-Za-z0-9_]+!\\)?\\|end\\|else\\|elsif\\|when\\|then\\|against\\)[ ;\t\n]")
    ))

;; True if current line ends with the keyword 'is' and an optional
;; comment.
(defun s-ends-with-is-p ()
  (save-excursion
    (end-of-line)
    (let ((end (point)))
      (beginning-of-line)
      (and (re-search-forward "\\(^\\|\\s-\\)is\\s-*\\($\\|--\\)" end t)
	   (not (s-in-comment-p))))
    ))

;; True if point is in a quoted string.
(defun s-in-quoted-string-p (&optional arg)
  (save-excursion
    (if arg (goto-char arg))
    (let ((p (point)))
      (beginning-of-line)
      ;; If there are an odd number of quotes between point and the
      ;; beginning of the line, we're in a quoted string.
      (let ((odd nil))
	(while (re-search-forward "\"" p t)
	  (setq odd (null odd)))
	odd)
      )))

;; Moves point to previous line excluding blank lines.  Returns t if
;; successful, nil if not.
(defun s-move-to-prev-non-blank ()
  (beginning-of-line)
  (re-search-backward "^\\s-*[^ \t\n]" nil t))

;; Returns current line indentation.
(defun s-current-indentation ()
  (save-excursion
    (beginning-of-line)
    (skip-chars-forward " \t")
    (current-indentation)
    ))

;; Return the outer indentation of the current block. Returns 0 or
;; less if it can't find one. Looks for first unpaired is, if, loop,
;; case, typecase, or protect.
(defun s-get-block-indent ()
  (save-excursion
    (let ((depth 1))
      (while (and (> depth 0)
		  ;; Search for start of keyword
		  (re-search-backward sather-block-regexp nil t))
	(goto-char (match-beginning 1))
	(cond ((or (s-in-comment-p)
		   (s-in-quoted-string-p))
	       nil)			;ignore it
	      ((looking-at "end")	;end of block
	       (setq depth (1+ depth)))
	      ((looking-at "is")
	       (cond ((s-special-match-class-begin-p) (setq depth -2))
		     ((= depth 1) (setq depth -1))
		     (t (setq depth -2))))
	      (t			;head of block
	       (setq depth (1- depth)))))
      (cond ((> depth 0)		;check whether we hit top of file
	     0)
	    ((= depth -1)		;Hit an "is" in a routine def
	     sather-indent)
	    ((= depth -2)		;Hit class def or outside rout
	     0)
	    (t (current-indentation))))
    ))

(defun sather-which-class ()
  "Return the name of the class in which point resides."
  (interactive)
  (save-excursion
    (let ((current-point (point)))
      (if (re-search-backward sather-class-regexp nil t)
	  (let ((class-point (point))
		;; Sub 1 is the string containing the `class' keyword.
		;; Sub 5 is the string containing the name of the
		;; class.
		(mb1 (match-beginning 1))
		(me1 (match-end 1))
		(mb5 (match-beginning 5))
		(me5 (match-end 5)))
	    (if (and (re-search-forward "^end" nil t)
		     (<= current-point (point))
		     (>= current-point class-point))
		(message "Point currently within %s `%s'."
			 (buffer-substring mb1 me1)
			 (buffer-substring mb5 me5))
	      (message "Previous %s is `%s'."
		       (buffer-substring mb1 me1)
		       (buffer-substring mb5 me5))
	      ))
	(message "Point not within class boundries.")
	))
    ))

;;; Support for class/routine hiding (outlining with selective
;;; display):

;; Class hiding works on classes that begin and end in the first
;; column.  Routine hiding should work for routines of `any' format.
;; I've tried it on the Sather library code, and it seems to work okay
;; with most of it, if somewhat slowly.

;; Hide all lines in the region.
(defun s-hide-region (start end)
  (subst-char-in-region start end ?\n ?\r t))

;; Show (unhide) all lines in the region.
(defun s-show-region (start end)
  (subst-char-in-region start end ?\r ?\n t))

;; Search for the next class-terminating end.
(defun s-next-class-end ()
  (re-search-forward sather-class-end-regexp nil t))

;; v2: This should now work (all the time?).  It should also be a bit
;; faster.

;; Find the "end" statement that matches the current level open
;; expression.  The `end' parameter gives the buffer position ending
;; the open expression.
(defun s-find-matching-end (end)
  (save-excursion
    (goto-char end)
    (let ((depth 1))
      (while (> depth 0)
        (let ((mb) (me))
          (if (re-search-forward sather-block-regexp nil t)
              (progn
                (setq mb (match-beginning 1)) ; s-in-comment-p kills
                (setq me (match-end 1))       ; these, so save them.
                (if (and (not (s-in-comment-p mb))
			 (not (s-in-quoted-string-p mb)))
                    (progn
                      (goto-char mb)
                      (if (looking-at "end")
                          (setq depth (1- depth))
			(setq depth (1+ depth)))
                      )) ; if (and...) + progn
                )) ; if (re-...) + progn
          (goto-char me)
          )) ; while + let
      )
    (point)
    ))

(defun sather-hide-classes ()
  "Hide all the class definitions in the current buffer."
  (interactive)
  (save-excursion
    (beginning-of-buffer)
    (while (re-search-forward sather-class-regexp nil t)
	(message "Hiding class... `%s'"
		 (buffer-substring (match-beginning 5) (match-end 5)))
      (s-hide-region (match-beginning 0)
		     (progn (s-next-class-end)
			    (match-end 0))) ; From `s-next-class-end'.
      (end-of-line)
      (if (not (eobp))
	  (forward-char 1))
      )
    (message "Hiding class... done")
    ))

(defun sather-hide-current-class ()
  "Hide the current class definition."
  (interactive)
  (save-excursion
    (if (not (looking-at sather-class-regexp))
	(re-search-backward sather-class-regexp nil t))
    (s-hide-region (point) (progn (s-next-class-end) (point)))
    )
  (beginning-of-line))

(defun sather-show-classes ()
  "Show (unhide) all the class definitions in the current buffer."
  (interactive)
  (save-excursion
    (beginning-of-buffer)
    (while (and (not (eobp))
		(re-search-forward sather-class-regexp nil t))
      (message "Showing class... `%s'"
	       (buffer-substring (match-beginning 5) (match-end 5)))
      (s-show-region (point) (progn (end-of-line) (point)))
      (if (not (eobp))
	  (forward-line 1))
      )
    (message "Showing class... done")
    ))

(defun sather-show-current-class ()
  "Show (unhide) the current class definition."
  (interactive)
  (save-excursion
    (beginning-of-line)
    (s-show-region (point) (progn (end-of-line) (point)))
    ))

(defun sather-hide-routines ()
  "Hide the routine definitions in the current buffer."
  (interactive)
  (save-excursion
    (beginning-of-buffer)
    (while (and (not (eobp))
		(re-search-forward sather-routine-regexp nil t))
      (message "Hiding routine... `%s'"
	       (buffer-substring (match-beginning 2) (match-end 2)))
      (goto-char (match-beginning 3))
      (s-hide-region (point) (s-find-matching-end (match-end 3)))
      (end-of-line)
      (if (not (eobp))
	  (forward-char 1)))
    (message "Hiding routine... done"))
  (beginning-of-line))

(defun sather-hide-current-routine ()
  "Hide the current routine definition."
  (interactive)
  (let ((b-o-l))
    (save-excursion
      (let ((start-point (point)))
	(beginning-of-line)
	(if (looking-at sather-routine-regexp)
	    (progn
	      (s-hide-region (match-beginning 3)
			     (s-find-matching-end (match-end 3)))
	      (setq b-o-l t))
	  (if (re-search-backward sather-routine-regexp nil t)
	      (let* ((mb (match-beginning 0))
		     (me (match-end 3))
		     (end (s-find-matching-end me)))
		(if (and
		     (>= start-point mb)
		     (<= start-point end))
		    (progn
		      (s-hide-region mb end)
		      (setq b-o-l t))
		  ))
	    ))
	)) ; save + let
    (if b-o-l
	(beginning-of-line))
    ))

(defun sather-hide-current-class-routines ()
  "Hide the routines of the current class."
  (interactive)
  (let ((valid))
    (save-excursion
      (let* ((save-point (point))
	     (cs (re-search-backward sather-class-regexp nil t))
	     (ce (progn
		   (s-next-class-end)
		   (end-of-line)
		   (if (not (eobp))
		       (forward-char 1))
		   (point))))
	(if (and (>= save-point cs) (< save-point ce))
	    (progn
	      (narrow-to-region cs ce)
	      (setq valid t))
	  (progn
	    (message "Not within class boundries.")
	    (setq valid nil)))
	)) ; save- + let
    (if valid
	(progn
	  (sather-hide-routines)
	  (widen)))
    ))

(defun sather-show-routines ()
  "Show (unhide) all the routine definitions in the current buffer."
  (interactive)
  (save-excursion
    (beginning-of-buffer)
    (while (not (eobp))
      (re-search-forward sather-routine-regexp nil t)
      (message "Showing routine... `%s'"
	       (buffer-substring (match-beginning 2) (match-end 2)))
      (beginning-of-line)
      (s-show-region (match-beginning 3) (progn (end-of-line) (point)))
      (if (not (eobp))
	  (forward-char 1)))
    (message "Showing routine... done")
    ))

(defun sather-show-current-routine ()
  "Show (unhide) the current routine definition."
  (interactive)
  (save-excursion
    (beginning-of-line)
    (if (looking-at sather-routine-regexp)
	(s-show-region (match-beginning 3)
		       (progn
			 (goto-char (match-end 3))
			 (end-of-line)
			 (point)))
      )
    ))

(defun sather-show-current-class-routines ()
  "Show (unhide) the routines of the current class."
  (interactive)
  (save-excursion
    (re-search-backward sather-class-regexp nil t)
    (narrow-to-region (point) (progn (s-next-class-end) (point)))
    (sather-show-routines)
    (widen)
    ))

(defun sather-show-all ()
  "Show (unhide) everything.
Use this in case some of the specialized functions don't work.  Also,
it's a lot faster than some of the specialized functions."
  (interactive)
  (save-excursion
    (widen)
    (s-show-region (point-min) (point-max))
    ))

(defun sather-mouse-hide-something (event)
  "Look at the context of the mouse click, and hide whatever is \"there\"."
  (interactive "e")
  (save-excursion

    (goto-char (posn-point (event-start event)))
    (beginning-of-line)

    (cond ((looking-at ".*\r")
	   (s-show-region (point) (progn (end-of-line) (point))))
	  ((looking-at sather-class-regexp)
	   (sather-hide-current-class))
	  (t
	   (sather-hide-current-routine)))
    ))

;;; Sexps for Sather:

;; This is easy because we already have `s-find-matching-end'.  It
;; would require more work to go the other way.
(defun sather-forward-sexp ()
  "Go to the matching end of the current block opener."
  (interactive)
  (cond ((looking-at sather-class-regexp)
	 (re-search-forward sather-class-end-regexp nil t))
	((looking-at sather-routine-regexp)
	 (goto-char (s-find-matching-end (match-end 3))))
	((looking-at sather-block-regexp)
	 (if (looking-at "\\s-+end")
	     (message "No opening expression found.")
	   (goto-char (s-find-matching-end (match-end 0)))))
	(t
	 (message "No opening expression found."))
	)
  (point)
  )

(defun sather-indent-sexp ()
  "Indent the sexp that begins the line containing point."
  (interactive)
  (save-excursion
    (indent-region (progn (beginning-of-line) (point))
		   (sather-forward-sexp)
		   nil)
    ))

;;; Compilation:

;; Since (at least) Emacs 19.28, there is a function `recompile' in
;; "compile.el". XEmacs 19.11 still does not have it.

(if (fboundp 'recompile)
    (defalias 'sather-compile-now 'recompile)
  (defun sather-compile-now ()
    "Rerun compilation with current `compile-command'."
    (interactive)
    (compile compile-command)))

;;; Interacting with the Sather Manual.

(defun sather-topic-lookup ()
  "Search for a topic using the Sather Specification info manual.

This will take over any info session you have going.  You may want to
rename any current info buffers using \\[rename-buffer]."
  (interactive)
  (save-excursion
    (let ((topic))
      (if (not (looking-at "\\<\\sw+"))
	  (backward-sexp))
      (if (looking-at "\\<\\sw+")
	  (setq topic (buffer-substring
		       (point)
		       (progn (forward-sexp) (point)))))
      (setq topic (read-string "Topic: " topic))
      (info sather-manual-name)
      (Info-index topic))
    ))

(defun sather-browse-manual ()
  "Browse the Sather Manual."
  (interactive)
  (info sather-manual-name))

(defun sather-browse-tutorial ()
  "Browse the Sather Eclectic Tutorial."
  (interactive)
  (info sather-tutorial-name))

;;; Submitting bug reports (mostly stolen from cc-mode):

(defconst sather-version "1.36"
  "The sather-mode version number.")

(defconst sather-compiler-help-address "sather-bugs@icsi.berkeley.edu"
  "Address accepting submission of Sather Compiler bug reports.")

(defconst sather-mode-help-address "lewikk@grasshopper.aud.alcatel.com"
  "Address accepting submission of sather-mode bug reports.")

(defun sather-version ()
  "Echo the current version of sather-mode in the minibuffer."
  (interactive)
  (message "Using sather-mode v%s" sather-version))

(defun sather-compiler-submit-bug-report ()
  "Submit via mail a bug report on the Sather compiler."
  (interactive)
  ;; Load in reporter.
  (let ((reporter-prompt-for-summary-p t))
    (and
     (y-or-n-p "Do you want to submit a report on the Sather compiler? ")
     (require 'reporter)
     (reporter-submit-bug-report
      sather-compiler-help-address
      "Compiler Report"
      nil
      nil
      nil
      "Dear David:\n\n\
This report is for Sather Compiler version: \n\n\
This is a description of the problem:\n\n\
The following code segment demonstrates the problem:\n\n"
      )
     ) ; and
    ))

(defun sather-mode-submit-bug-report ()
  "Submit via mail a bug report on sather-mode."
  (interactive)
  ;; Load in reporter.
  (let ((reporter-prompt-for-summary-p t))
    (and
     (y-or-n-p "Do you want to submit a report on sather-mode? ")
     (require 'reporter)
     (reporter-submit-bug-report
      sather-mode-help-address
      (concat "Report on sather-mode v" sather-version)
      ;; Variables that affect indentation (mostly).
      (list 'sather-indent
	    'sather-continued-line-indent
	    'sather-auto-indent
	    'sather-auto-newline
	    'sather-newline-prefix-indent
	    'sather-comment-col
	    'sather-compile-command
	    'tab-width
	    'sather-hilit-type)
      nil
      nil
      "Dear Kevin, nice try, but: "
      )
     ) ; and
    ))

;;; Give it up:
(provide 'sather)
