/* 
 * Copyright (C) International Computer Science Institute, 1994,1995. COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the SATHER LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/License" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email comments to "sather-bugs@icsi.berkeley.edu". 
 */

#define IN_PSATHER_LIBRARY
#ifndef SOLARIS
#define SOLARIS
#endif
#ifndef SOLARIS_THREADS
#define SOLARIS_THREADS
#endif
#ifndef _REENTRANT
#define _REENTRANT
#endif
#include  "sather.h"
#ifndef _CONFIG_LH_
#define _CONFIG_LH_
/*
 * This file allows you to configure different aspects of the pSather
 * runtime module for solaris and the lwp-thread package/
 *
 * If you want to disable a feature just comment the corresponding line out.
 *
 */

/*
 * To get a complete Trace file usable for the pSather Lock Trace,
 * use PSATHER_RT_TRACE. If you are just interested about the 
 * statistics only, define PSATHER_RT_STATISTICS
 */
#ifdef PSATHER_STATS
#  define PSATHER_RT_TRACE
#  ifndef PSATHER_TRACE
#    define PSATHER_RT_STATISTICS_ONLY
#  endif
#else
#  ifdef PSATHER_TRACE
#    define PSATHER_RT_TRACE
#  endif
#endif

/*
 * By not defining NO_CHECKS the runtime runs faster, but fails to see
 * some errors you made in your program (like the following code)
 * lock a then
 *    if cond_1 then unlock a; end;
 *    if cond_2 then unlock a; end;
 * end;
 * If both conditions hold, you made an error. If the runtime shall not
 * catch such errors, its behaviour is undefined if this happens
 */
#ifndef PSATHER_CHK
#define NO_ERROR_CHECKS
#endif

/*
 * The deadlock detection used finds only deadlocks when all threads
 * are blocked on either a lock or a sync. If any thread is still running
 * it will not complain.
 * Overhead: minimal, but there are some counters used, and the access to
 *           this counters is protected by a lock.
 * The deadlock scheme works as follows: each time a plate is waked up,
 * it checks to see if all threads are either waiting in a sync, or are waiting
 * for locks. If this is the case, and the plate is the lowest on
 * all sticks and cannot obtain them, it increments the failed_plate counter.
 * If this counters equals the number of plates speared on the sticks, a
 * deadlock occured. The counter is set to zero each time a threads acquires
 * the locks it needs.
 */
#define DEADLOCK_DETECTION

/*
 * Check status whenever a plate is waked up by a timer. Theoretically 
 * a plate cannot acquire all the locks if it is woken up by a 
 * timer, as otherwise it would have been woken up by the thread
 * which opened the locks. But between the time the timer actually rang
 * and the plate restarts work the locks may become free.
 * If you use DEADLOCK_DETECTION, you must use this optimization too!
 */
#define WAKEUP_AND_CHECK_LOCKS

/*
 * it is actually not necessary to start a timer if a thread owns the top plate
 * (that is, no other plate lies above it). This implies however, that each
 * time a new late is speared on top, it wakes up all plates below which
 * have no timer running.
 */
#define NO_TIMER_IF_TOP_PLATE

/*
 * A plate which does not lie at the bottom of at least one stick,
 * does not need a timer, as one of the plates below will vanish
 * and THEN it will have to start its timer.
 */
#define NO_TIMER_IF_NOT_AT_THE_BOTTOM

/*
 * If a plate is speared upon another plate, this other plate
 * is normally woken up, if NO_TIMER_IF_TOP_PLATE is on. It is however
 * not necessary, if this other plate does not see the bottom of at least 
 * one stick (this is only true if NO_TIMER_IF_NOT_AT_THE_BOTTOM is on).
 * This optimization checks if the other plate has a size of one, and
 * in this case it is only woken up if there is no plate below.
 * Works only if NO_TIMER_IF_NOT_AT_THE_BOTTOM and NO_TIMER_IF_TOP_PLATE
 * is on.
 */
#define NO_WAKEUP_FOR_1_PLATES_IF_NOT_AT_BOTTOM

/*
 * If a lock opens, the bottom plate is normally woken up with a CHANGE
 * flag. If this plate cannot get the lock because of its condition,
 * it will starts its timer. As no other plate may get that lock BEFORE
 * the timer rings, this plate could instead make an immediately jump.
 * The following define triggers this optimization. It will make
 * all plates jump up, if their condition does not hold.
 * If no plate exists with a correct condition, no plate will have to jump.
 * The jumping plate should NOT wakeup any other plate.
 */
#define MAKE_PLATE_WITH_WRONG_CONDITION_JUMP
 
/*
 * If several Plates try to use some locks, they will start
 * to rotate, as the bottom one will set a timer, jump up,
 * and so on. As long as the locks do not change, there is
 * no point in doing this jumps. 
 *
 * This situation is detected as follows:
 * - Each lock has a counter. Each time the lock is opened or closed,
 *   the counter is incremented. The counter starts with 1.
 *
 * - Each plate has a mark for each lock, where it stores the counter
 *   value of the locks when it checked them the last time
 *
 * The algorithm described below includes all optimization done so far.
 *
 **GENERAL REMARKS**
 * * Double Locks
 *   A thread may relock a lock it already has. If it tries to do so,
 *   two possibilities exists:
 *   1. The condition for this lock does not hold, in this case
 *      a TRY will fail, and a LOCK will abort because of a deadlock
 *      situation.
 *   2. The conditions are satisfied. In this case the thread needs
 *      no plate on this particular stick (although it must lock it
 *      before it returns because of possible unlocks).
 *
 * * Timer
 * Each time a plate would need to restart its timer (according to
 * the simple algorithm), it makes a note of the current time. If it later
 * needs to start the timer, it starts it as if it had been started then.
 *
 * * Signals
 * A plate may receive different signals before it wakes up (due to
 * scheduling). The signals are checked in this order:
 * LOCK_CHANGE, JUMP_TO_TOP, TIMER_SIGNAL START_TIMER
 *
 *
 **NEW PLATE**
 * If a new plate is speared on the sticks, it checks if it could
 * acquire all locks it needs regardless of plates below.
 * (the locks must be open and the conditions
 * must hold). If this is so, it checks the bottom plate on each stick.
 * If the plate there could acquire the lock and its lock marker is
 * outdated, it receives a START_TIMER
 * signal (of course only if the timer is not already running). Otherwise
 * the plate receives a "JUMP_TO_TOP" signal.
 * If some of the conditions do not hold or a some locks are closed,
 * the new plate notes the current value of all lock marks and the current time
 * In this case it does not wake up any plate.
 *
 **LOCK_CHANGE**
 * - the plate checks to see if it could acquire all locks, if so
 *   it falls through the sticks
 * otherwise:
 * - if it could acquire all locks if there were no plates below, it sends 
 *   these plate a START_TIMER, if their conditions hold and their mark
 *   is outdated, else they get a JUMP_TO_TOP
 *   it sets all its marker to zero.
 * otherwise
 * - It notes the counter of all locks
 *   If it cannot acquire one of the locks because of a wrong condition,
 *   then 
 *        if one of the plates above could acquire the lock
 *           and its lock marker is outdated it behaves as if it had 
 *           received a JUMP_TO_TOP signal
 *   otherwise
 *   if one of the plates above has an outdated mark for an open lock
 *   and the condition for that lock holds, it starts its own timer.
 *
 **TIMER_SIGNAL**
 * - same as JUMP_TO_TOP
 *
 **START_TIMER**
 * - If a plates wakes up because of a START_TIMER, it checks if it could
 *   acquire all locks if no plates were below.
 *      if this is the case and there are no plates below, it gets the locks,
 *      otherwise, it sets all lockmarks to zero and wakes up the bottom
 *      plate on each stick with START_TIMER
 *   otherwise
 *   If one of the locks is either locked ot the conditions do not hold,
 *   it updates its lock marks.
 *   If some of the locks it needs are actually free, but the conditions
 *   do not hold and the conditions for all plates below on this sticks do not
 *   hold either, it will jump to the top immediately,
 *   Otherwise it will start the timer as if it had been
 *   started when the plate checked the current time 
 *
 **JUMP_TO_TOP**
 * - If a plate receives this signal, it jumps to the top of all plates.
 *   Then it checks each stick corresponding to an open lock to see
 *   if some plate could acquire that lock (its condition must hold and
 *   lock marker must be outdated). If this plate happens
 *   to be at the bottom, it will be woken up with "START_TIMER"
 *   otherwise the bottom plate will be woken up with "JUMP_TO_TOP"
 *   If one of the locks it needs is either locked ot the conditions 
 *   do not hold, it updates its lock marks.
 * - it makes a note of the current time.
 *   
 *
 **ACQUIRING A LOCK**
 * - If a plate acquires a lock, it will check all plates on the stick
 *   of this lock, and it will
 *   change their lock mark to reflect the new counter of the lock.
 *   If some of the plates are TRY-Plates, they are woken up with 
 *   CHANGE_LOCK
 *   The bottom plate (which must not be a TRY plate) is woken up with 
 *   START_TIMER
 * 
 **RELEASING A LOCK**
 * - If a plate releases a lock, it increments the lock  counter and checks all
 *   plates on this stick if they could acquire that lock. If one
 *   can, and it is the bottom plate, it will be woken up with "LOCK_CHANGE".
 *   If it is not the bottom plate, the bottom plate will be woken up with
 *   "JUMP_TO_TOP"
 */

/*
 * Whenever a lock checks if it could acquire a lock, it can ignore all
 * plates below on its stick, IF their conditions do not hold AND the
 * lock is currently open.
 */
#define IGNORE_WRONG_CONDITION_PLATES


/*
 * Do not changes any of the lines below, they enforce some 
 * defines which MUST be there, as the runtime will not work otherwise
 */
#ifdef DEADLOCK_DETECTION
#ifndef WAKEUP_AND_CHECK_LOCKS
#define WAKEUP_AND_CHECK_LOCKS
#endif

#ifdef NO_WAKEUP_FOR_1_PLATES_IF_NOT_AT_BOTTOM
#define NO_TIMER_IF_TOP_PLATE
#define NO_TIMER_IF_NOT_AT_THE_BOTTOM
#endif  

#ifdef PSATHER_RT_STATISTICS_ONLY
#define PSATHER_RT_TRACE
#endif 
#endif

#endif 
/* 
 * Copyright (C) International Computer Science Institute, 1994.  COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the SATHER LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/License" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email comments to "sather-bugs@icsi.berkeley.edu". 
 */
#ifndef _PSATHER_H_
#define _PSATHER_H_

#include <stddef.h>
#include <stdlib.h>
#include <setjmp.h>

/*
 * PSATHER is only defined when the cs compiler compiles this
 * file. 
 */
#ifndef PSATHER 
#define PSATHER    
typedef void *OB;
typedef struct { int tag; } OB_HEADER;
#endif

/*
 * To get good error messages, you need to define DEBUG and PSATHER_CHK
 * when compiling your program and the psather library.
 * FORBID FLO
 * FORBID FL
 * FORBID FLA
 * FORBID FLAO
 * FORBID FLP
 * FORBID FLPO
 */
#ifdef DEBUG
#define FLO	__FILE__,__LINE__
#define FL	FLO,
#define FLAO	char*file,long line
#define FLA	FLAO,
#define FLPO	file,line
#define FLP	FLPO,
#define FLDO	"?",0
#define FLD	FLDO,
#define RFATAL(s) do { fprintf(stderr,"%s:%ld:%s\n",__FILE__,(long)__LINE__,s);abort(); } while(0)
#define RFATALexp(s) (fprintf(stderr,"%s:%ld:%s\n",__FILE__,(long)__LINE__,s),abort(),0)
#define RFATALp(s) do { fprintf(stderr,"%s:%ld:%s\n",file,(long)line,s);abort(); } while(0)
#else
#define FLO
#define FL
#define FLAO	void
#define FLA
#define FLPO	
#define FLP	
#define FLDO
#define FLD
#define RFATAL(s) do { fprintf(stderr,"%s\n",s);abort(); } while(0)
#define RFATALexp(s) (fprintf(stderr,"%s\n",s),abort(),0)
#define RFATALp(s) do { fprintf(stderr,"%s\n",s);abort(); } while(0)
#endif

/*
 * HERE should return the cluster id (an integer).
 * For one cluster machines like solaris threads it defaults
 * to 0
 */
#define HERE		0
#define ANY		0

/*
 * CLUSTERS defines the number of clusters for the current
 * run. This must be a constant during run time, but for
 * each run it may have a different value.
 */
#define CLUSTERS	1

/*
 * PSATHER_START: must be called before any of the functions defined
 *               in this header files are used.
 * PSATHER_STOP:  should be called before ending the program.
 */
#define PSATHER_START(progname)	psather_start(FL progname)
#define PSATHER_STOP		psather_stop(FLO)

/*
 * the SYS class
 */
#define SYS_CLEARED_SET(value)		p_cleared_set(FL value)
#define SYS_CLEARED_GET			p_cleared_get(FLO)

#define SYS_TRAP_CLEAR_SET(value)	p_trap_clear_set(FL value)
#define SYS_TRAP_CLEAR_GET		p_trap_clear_get(FLO)

#define SYS_DEFER			p_defer(FLO)

#define SYS_IMPORT
#define SYS_EXPORT

/*
 * VASSIGN guarantees atomicity for value type assignment
 */
#define VASSIGN(a,b) do { p_lock_v(&a);a=b;p_unlock_v(&a); } while(0)

/*
 * ATTACH, uses four arguments
 * function: the function to execute
 * gate:     each thread is attached to a gate, and here you pass the
 *           gate for the new process.
 * argument: a structure passed as argument to function.
 * position: the cluster on which this function should be executed.
 */
#define ATTACH(func,self,arg,gate,pos) p_attach(FL (FORK_FUNC)func,self,(void *)arg,sizeof(*arg),((GATE)gate)->lck,pos)

/*
 * LOCKS, MUTEX and GATES
 * First, we define the type of the different locks. 
 * Note that each lock type must use one bit. These bits are | together,
 *
 * Note: the pSather Compiler should not generate identifiers with the
 *       following names. Forbid is used to extract them automagically.
 *	 FORBID LOCK_TYPES
 *	 FORBID L_IGNORE
 *	 FORBID L_MUTEX
 *	 FORBID L_NO_THREADS
 *	 FORBID L_THREADS
 *	 FORBID L_EMPTY
 *	 FORBID L_NOT_EMPTY
 *	 FORBID DLOCK
 *	 FORBID LOCK
 *	 FORBID dLOCK
 *	 FORBID DLOCK_struct
 */
enum LOCK_TYPES { 
	L_IGNORE=0x8000, 
	L_MUTEX=1,
	L_NO_THREADS=2,
	L_THREADS=4,
	L_EMPTY=8,
	L_NOT_EMPTY=16
};
typedef struct DLOCK_struct { 
	OB_HEADER header;
	struct LOCK_T_struct *lck; 
	enum LOCK_TYPES type; 
	char count; 
} DLOCK;

typedef DLOCK *LOCK; /* anachronism */
typedef DLOCK *dLOCK;
typedef DLOCK *GATE;
typedef DLOCK *GATEV;
typedef DLOCK *GATET;
typedef DLOCK *MUTEX;

/* 
 * MUTEX
 */
#define MUTEX_CREATE			p_get_lock(FL (struct LOCK_T_struct *)p_mutex_create(FLO),L_MUTEX)

/*
 * Gates
 */
#define GATE_CREATE			(GATE)p_get_lock(FL p_gate_create(FL 0),L_MUTEX)
#define GATET_CREATE			(GATET)p_get_lock(FL p_gate_create(FL 4),L_MUTEX)
#define GATEV_CREATE(V)			(GATEV)p_get_lock(FL p_gate_create(FL sizeof(V)),L_MUTEX)
 
#define GATE_SIZE(gate)			CHKGATEi(gate,p_gate_size(FL (struct GATE_T_struct *)((GATE)gate)->lck))
#define GATET_SIZE(gate)		CHKGATEi(gate,p_gate_size(FL (struct GATE_T_struct *)((GATET)gate)->lck))
#define GATEV_SIZE(gate)		CHKGATEi(gate,p_gate_size(FL (struct GATE_T_struct *)((GATEV)gate)->lck))
 
#define GATE_HAS_THREADS(gate)		CHKGATEi(gate,p_gate_has_threads(FL (struct GATE_T_struct *)((GATE)gate)->lck))
#define GATET_HAS_THREADS(gate)		CHKGATEi(gate,p_gate_has_threads(FL (struct GATE_T_struct *)((GATET)gate)->lck))
#define GATEV_HAS_THREADS(gate)		CHKGATEi(gate,p_gate_has_threads(FL (struct GATE_T_struct *)((GATEV)gate)->lck))
 
#define GATE_SET(gate)			CHKGATEv(gate,p_gate_set(FL (struct GATE_T_struct *)((GATE)gate)->lck,0))
#define GATET_SET(gate,pointer)		CHKGATEv(gate,p_gate_set(FL (struct GATE_T_struct *)((GATET)gate)->lck,pointer,0))
#define GATEV_SET(gate,str)		CHKGATEv(gate,p_gate_set(FL (struct GATE_T_struct *)((GATEV)gate)->lck,&str,sizeof(str)))
 
#define GATE_GET(gate)			CHKGATEp(gate,p_gate_get(FL (struct GATE_T_struct *)((GATE)gate)->lck,NULL))
#define GATET_GET(gate,pointer)		CHKGATEp(gate,pointer=p_gate_get(FL (struct GATE_T_struct *)((GATET)gate)->lck,NULL))
#define GATEV_GET(gate,str)		CHKGATEp(gate,p_gate_get(FL (struct GATE_T_struct *)((GATEV)gate)->lck,&str))

#define GATE_ENQUEUE(gate)		CHKGATEv(gate,p_gate_enqueue(FL (struct GATE_T_struct *)((GATE)gate)->lck))
#define GATET_ENQUEUE(gate,pointer)	CHKGATEv(gate,p_gate_enqueue(FL (struct GATE_T_struct *)((GATET)gate)->lck,pointer,0))
#define GATEV_ENQUEUE(gate,str)		CHKGATEv(gate,p_gate_enqueue(FL (struct GATE_T_struct *)((GATEV)gate)->lck,&str,sizeof(str)))

#define GATE_DEQUEUE(gate)		CHKGATEp(gate,p_gate_dequeue(FL (struct GATE_T_struct *)((GATE)gate)->lck,NULL))
#define GATET_DEQUEUE(gate,pointer)	CHKGATEp(gate,pointer=(void*)p_gate_dequeue(FL (struct GATE_T_struct *)((GATET)gate)->lck,NULL))
#define GATEV_DEQUEUE(gate,str) 	CHKGATEp(gate,p_gate_dequeue(FL (struct GATE_T_struct *)((GATEV)gate)->lck,&str))

#define GATE_CLEAR(gate)		CHKGATEv(gate,p_gate_clear(FL (struct GATE_T_struct *)((GATE)gate)->lck))
#define GATET_CLEAR(gate)		CHKGATEv(gate,p_gate_clear(FL (struct GATE_T_struct *)((GATET)gate)->lck))
#define GATEV_CLEAR(gate)		CHKGATEv(gate,p_gate_clear(FL (struct GATE_T_struct *)((GATEV)gate)->lck))

#define GATE_SYNC(gate)			CHKGATEv(gate,p_gate_sync(FL (struct GATE_T_struct *)((GATE)gate)->lck))
#define GATET_SYNC(gate)		CHKGATEv(gate,p_gate_sync(FL (struct GATE_T_struct *)((GATET)gate)->lck))
#define GATEV_SYNC(gate)		CHKGATEv(gate,p_gate_sync(FL (struct GATE_T_struct *)((GATEV)gate)->lck))

#define GATE_MUTEX(n)			CHKGATEp(n,p_get_lock(FL ((GATE)n)->lck,L_MUTEX))
#define GATET_MUTEX(n)			CHKGATEp(n,p_get_lock(FL ((GATET)n)->lck,L_MUTEX))
#define GATEV_MUTEX(n)			CHKGATEp(n,p_get_lock(FL ((GATEV)n)->lck,L_MUTEX))

#define GATE_EMPTY(n)			CHKGATEp(n,p_get_lock(FL ((GATE)n)->lck,L_EMPTY))
#define GATET_EMPTY(n)			CHKGATEp(n,p_get_lock(FL ((GATET)n)->lck,L_EMPTY))
#define GATEV_EMPTY(n)			CHKGATEp(n,p_get_lock(FL ((GATEV)n)->lck,L_EMPTY))

#define GATE_NOT_EMPTY(n)		CHKGATEp(n,p_get_lock(FL ((GATE)n)->lck,L_NOT_EMPTY))
#define GATET_NOT_EMPTY(n)		CHKGATEp(n,p_get_lock(FL ((GATET)n)->lck,L_NOT_EMPTY))
#define GATEV_NOT_EMPTY(n)		CHKGATEp(n,p_get_lock(FL ((GATEV)n)->lck,L_NOT_EMPTY))

#define GATE_THREADS(n)			CHKGATEp(n,p_get_lock(FL ((GATE)n)->lck,L_THREADS))
#define GATET_THREADS(n)		CHKGATEp(n,p_get_lock(FL ((GATET)n)->lck,L_THREADS))
#define GATEV_THREADS(n)		CHKGATEp(n,p_get_lock(FL ((GATEV)n)->lck,L_THREADS))

#define GATE_NO_THREADS(n)		CHKGATEp(n,p_get_lock(FL ((GATE)n)->lck,L_NO_THREADS))
#define GATET_NO_THREADS(n)		CHKGATEp(n,p_get_lock(FL ((GATET)n)->lck,L_NO_THREADS))
#define GATEV_NO_THREADS(n)		CHKGATEp(n,p_get_lock(FL ((GATEV)n)->lck,L_NO_THREADS))

/*
 * WARNING: you can get the gate to which a thread is attached, BUT
 *          you must know what kind of gate this is.
 * There is currently no way in pSather to get this gate, as the
 * cohort is implemented by directly using the specific gate.
 */
#define GATE_COHORT			(p_get_lock(FL p_cohort(FLO),L_MUTEX))
#define GATET_COHORT			(p_get_lock(FL p_cohort(FLO),L_MUTEX))
#define GATEV_COHORT			(p_get_lock(FL p_cohort(FLO),L_MUTEX))

/*
 * And now the protect, lock, try and unlock
 * a maximum of 5 arguments are currently supported for lock and try
 */

/*
 * the structure EXCEPT_ELEMENT should not be used anywhere in the code
 * FORBID EXCEPT_ELEMENT
 * FORBID EXCEPT_ELEMENT0
 * FORBID EXCEPT_ELEMENT1
 * FORBID EXCEPT_ELEMENT2
 * FORBID EXCEPT_ELEMENT3
 * FORBID EXCEPT_ELEMENT4
 * FORBID EXCEPT_ELEMENT5
 */
struct EXCEPT_ELEMENT {
        struct EXCEPT_ELEMENT *next;
        jmp_buf jmp;
        OB exception;
#ifdef DEBUG
	long id;
#endif
	short flags;
	short mark;
        short nlocks;
        DLOCK locks[1];
};
#ifdef DEBUG
#define EXCEPT_ELEMENTn(n) struct EXCEPT_ELEMENT ## n { \
        struct EXCEPT_ELEMENT *next;			\
        jmp_buf jmp;					\
        OB exception;					\
	long id;					\
	short flags;					\
	short mark;					\
        short nlocks;					\
        DLOCK locks[n];					\
}
#else
#define EXCEPT_ELEMENTn(n) struct EXCEPT_ELEMENT ## n { \
        struct EXCEPT_ELEMENT *next;			\
        jmp_buf jmp;					\
        OB exception;					\
	short flags;					\
	short mark;					\
        short nlocks;					\
        DLOCK locks[n];					\
}
#endif
EXCEPT_ELEMENTn(1);
EXCEPT_ELEMENTn(2);
EXCEPT_ELEMENTn(3);
EXCEPT_ELEMENTn(4);
EXCEPT_ELEMENTn(5);
#define LOCK_RESERVE 1

/*
 * While executing a lock, try or protect statement, an exception stack
 * is build. You can mark specific stack frames and unwind the stack to the
 * first mark at any time. You have 16 different mark (0-15), but MARK(0)
 * and MARK(1) are predefined. The first is used to mark the stack at the
 * beginning of a function, while the second can be used for loops.
 * Be warned however, that while ESCAPE unwinds the stack end ends the
 * function, JUMP_OUT_OF_LOOP does NOT terminate the loop. You have to 
 * manually end the loop by either a break or a goto.
 */
#define MARK(n)	(1<<(n))
#define MARK_FUNCTION			MARK(0)
#define MARK_LOOP			MARK(1)
#define SET_EXCEPTION_MARK(n)		p_ex_set_mark(FL n)
#define REMOVE_EXCEPTION_MARK(n)	p_ex_remove_mark(FL n)
#define POP_TO(m)			p_ex_unwind(FL (m))

/*
 * each function using return MUST use EXCEPTION_PROLOGUE at the 
 * beginning, and EXCEPTION_EPILOGUE at the end. Instead of using
 * return, it MUST use ESCAPE.
 */
#define EXCEPTION_PROLOGUE 	SET_EXCEPTION_MARK(MARK_FUNCTION);
#define EXCEPTION_EPILOGUE 	REMOVE_EXCEPTION_MARK(MARK_FUNCTION);
#define ESCAPE			POP_TO(MARK_FUNCTION);return 

/*
 * If you want to use mark on the stack to end loops, use
 * LOOP_BEGIN before the loop and LOOP_END at its end.
 * Note that JUMP_OUT_OF_LOOP must be followed by a break or goto,
 * as it will not end the loop.
 */
#define LOOP_BEGIN		SET_EXCEPTION_MARK(MARK_LOOP);
#define LOOP_END		REMOVE_EXCEPTION_MARK(MARK_LOOP);
#define JUMP_OUT_OF_LOOP	POP_TO(MARK_LOOP)

/*
 * To pop some frames from the exception stack, you may use POP_EXCPETION.
 * This macro can be used instead of LOOP_BEGIN and LOOP_END. The only
 * problem is that you must know exactly how many frames to pop
 * when exiting a loop.
 */
#define POP_EXCEPTION1		p_ex_pop(FL 1)
#define POP_EXCEPTION(n)	p_ex_pop(FL (n))

/* 
 * Before using exception, you must define the clear exception with
 * SET_CLEAR_EXCEPTION(p). This macro accepts a pointer. This is the
 * only exception the runtime may raise, and if this happens,
 * EXCEPTIONS is set to the value defined by SET_CLEAR_EXCEPTION.
 */
#define SET_CLEAR_EXCEPTION(p)		p_ex_set_clear_exception(FL (OB)(p))

/*
 * RAISE(x) raises an exception. The value x (actually a pointer) is passed
 * to the exception handling routine, which can use it through the
 * macro EXCEPTION. If this routine needs to propagate the exception,
 * it can use RAISE(EXCEPTION)
 */
#define RAISE(x)			p_ex_raise(FL (x))
#define EXCEPTION 	   		((OB)p_ex_exception(FLO))

/* 
 * To protect some statements A against exception, use
 * PROTECT_BEGIN
 *	A;
 * PRTOECT_WHEN
 *	if(EXCPETION==....
 * PRTOECT_END
 *
 * FORBID p_n
 */
#define PROTECT_BEGIN	   { struct EXCEPT_ELEMENT p_n;			     \
			     p_n.flags=0;				     \
			     p_n.mark=0;				     \
			     p_n.nlocks=0;				     \
			     if(setjmp(p_n.jmp)==0) {		  	     \
				p_ex_push(FL &p_n);
             
#define PROTECT_WHEN	     p_ex_pop(FL 1); } else {
#define PROTECT_END	     }		   \
			   }

/*
 * PROTECT_R_* is for use in the runtime only
 */
#define PROTECT_R_BEGIN	   { struct EXCEPT_ELEMENT p_n;			     \
			     p_n.flags=0;				     \
			     p_n.mark=0;				     \
			     p_n.nlocks=0;				     \
			     if(setjmp(p_n.jmp)==0) {		  	     \
				p_ex_push(FLP &p_n);
             
#define PROTECT_R_WHEN	     p_ex_pop(FLP 1); } else {
#define PROTECT_R_END	     }		   \
			   }
/*
 * Multiple locks.
 * To lock n locks, use the lock macros as follows:
 *
 * LOCKn(l1, ... ln)
 *     statements;
 * LOCKEND
 * (ENDLOCK is equivalent to LOCKEND, but may not be supported in future
 * versions of this library)
 * WARNING: if PSATHER_CHK is defined arguments to LOCKn() and TRYn() are 
 * evaluated twice, so they may not have any side effects.
 * Example: LOCK1(g++) is an error!
 */
#ifdef PSATHER_CHK
#define CHKLOCK(lock,d,n)	if((lock)==NULL) RFATAL("locking a void lock (arg " #n); else d
#define CHKLOCKR(lock,d,n)	if((lock)==NULL) RFATAL("locking a void lock (arg " #n); else d
#define CHKTRY(lock,d,n)	if((lock)==NULL) RFATAL("trying a void lock (arg " #n); else d
#define CHKUNLOCK(lock)		if((lock)==NULL) RFATAL("unlocking a void lock");
#define CHKGATEp(gate,d)	((gate)==NULL?(void *)RFATALexp("using an uninitialized gate"):(d))
#define CHKGATEi(gate,d)	((gate)==NULL?RFATALexp("using an uninitialized gate"):(d))
#define CHKGATEv(gate,d)	((gate)==NULL?RFATALexp("using an uninitialized gate"):(d),0)
#else /* PSATHER_CHK */
#define CHKLOCK(lock,d,n) 	d
#define CHKLOCKR(lock,d,n) 	d
#define CHKTRY(lock,d,n) 	d
#define CHKUNLOCK(lock) 	
#define CHKGATEp(gate,d)	d
#define CHKGATEi(gate,d)	d
#define CHKGATEv(gate,d)	d
#endif /* PSATHER_CHK */
#define LOCK1(l1)	   { struct EXCEPT_ELEMENT1 p_n; 		    \
			     p_n.flags=0;				    \
			     p_n.mark=0;				    \
			     p_n.nlocks=1;				    \
			     CHKLOCK(l1,p_n.locks[0]= *(DLOCK *)l1,1);	    \
			     if(setjmp(p_n.jmp)==0) {			    \
			    	p_lock(FL (struct EXCEPT_ELEMENT *)&p_n); 
#define LOCK2(l1,l2)	   { struct EXCEPT_ELEMENT2 p_n;		    \
			     p_n.flags=0;				    \
			     p_n.mark=0;				    \
			     p_n.nlocks=2;			            \
			     CHKLOCK(l1,p_n.locks[0]= *(DLOCK *)l1,1);	    \
			     CHKLOCK(l2,p_n.locks[1]= *(DLOCK *)l2,2);	    \
			     qsort((char *)p_n.locks,2,sizeof(DLOCK),(int(*)())p_ex_cmplcks);\
			     if(setjmp(p_n.jmp)==0) {			    \
			    	p_lock(FL (struct EXCEPT_ELEMENT *)&p_n); 
#define LOCK3(l1,l2,l3)	   { struct EXCEPT_ELEMENT3 p_n;		    \
			     p_n.flags=0;				    \
			     p_n.mark=0;				    \
			     p_n.nlocks=3;			            \
			     CHKLOCK(l1,p_n.locks[0]= *(DLOCK *)l1,1);	    \
			     CHKLOCK(l2,p_n.locks[1]= *(DLOCK *)l2,2);	    \
			     CHKLOCK(l3,p_n.locks[2]= *(DLOCK *)l3,3);	    \
			     qsort((char *)p_n.locks,3,sizeof(DLOCK),(int(*)())p_ex_cmplcks);\
			     if(setjmp(p_n.jmp)==0) {			    \
			    	p_lock(FL (struct EXCEPT_ELEMENT *)&p_n); 
#define LOCK4(l1,l2,l3,l4)   { struct EXCEPT_ELEMENT4 p_n; 		    \
			     p_n.flags=0;				    \
			     p_n.mark=0;				    \
			     p_n.nlocks=4;			            \
			     CHKLOCK(l1,p_n.locks[0]= *(DLOCK *)l1,1);	    \
			     CHKLOCK(l2,p_n.locks[1]= *(DLOCK *)l2,2);	    \
			     CHKLOCK(l3,p_n.locks[2]= *(DLOCK *)l3,3);	    \
			     CHKLOCK(l4,p_n.locks[3]= *(DLOCK *)l4,4);	    \
			     qsort((char *)p_n.locks,4,sizeof(DLOCK),(int(*)())p_ex_cmplcks);\
			     if(setjmp(p_n.jmp)==0) {			    \
			    	p_lock(FL (struct EXCEPT_ELEMENT *)&p_n); 
#define LOCK5(l1,l2,l3,l4,l5){ struct EXCEPT_ELEMENT5 p_n;		    \
			     p_n.flags=0;				    \
			     p_n.mark=0;				    \
			     p_n.nlocks=5;			            \
			     CHKLOCK(l1,p_n.locks[0]= *(DLOCK *)l1,1);	    \
			     CHKLOCK(l2,p_n.locks[1]= *(DLOCK *)l2,2);	    \
			     CHKLOCK(l3,p_n.locks[2]= *(DLOCK *)l3,3);	    \
			     CHKLOCK(l4,p_n.locks[3]= *(DLOCK *)l4,4);	    \
			     CHKLOCK(l5,p_n.locks[4]= *(DLOCK *)l5,5);	    \
			     qsort((char *)p_n.locks,5,sizeof(DLOCK),(int(*)())p_ex_cmplcks);\
			     if(setjmp(p_n.jmp)==0) {			    \
			    	p_lock(FL (struct EXCEPT_ELEMENT *)&p_n); 
				
#define LOCKEND		        p_ex_pop(FL 1); 			\
			     } else {					\
				p_ex_raise(FL EXCEPTION);		\
			     }						\
			   }
#define ENDLOCK LOCKEND

/* 
 * LOCK_RESERVE is only meant for the runtime and should
 * never be called by code generated by the compiler.
 */
#define LOCK_RESERVE1(l1)   { struct EXCEPT_ELEMENT1 p_n;		    \
			     p_n.flags=LOCK_RESERVE;			    \
			     p_n.nlocks=1;				    \
			     p_n.mark=0;				    \
			     CHKLOCKR(l1,p_n.locks[0]= *(DLOCK *)l1,1);	    \
			     if(setjmp(p_n.jmp)==0) {			    \
			    	p_lock(FLP (struct EXCEPT_ELEMENT *)&p_n); 
				
#define LOCK_R_END		p_ex_pop(FLP 1); 			\
			     } else {					\
				p_ex_raise(FLP EXCEPTION);		\
			     }						\
			   }
/*
 * If you need to try on n locks, do it as follows:
 * TRYn(l1,... ln)
 *      statements_if_try_succeeds
 * TRYELSE
 *      statements_if_try_fails
 * TRYEND
 */
#define TRY1(l1)	   { struct EXCEPT_ELEMENT1 p_n;		\
			     p_n.flags=0;				\
			     p_n.mark=0;				\
			     p_n.nlocks=1;			        \
			     CHKTRY(l1,p_n.locks[0]= *(DLOCK *)l1,1);	\
			     if(setjmp(p_n.jmp)==0) {			\
			    	if(p_try(FL (struct EXCEPT_ELEMENT *)&p_n)) {
#define TRY2(l1,l2)	   { struct EXCEPT_ELEMENT2 p_n;		\
			     p_n.flags=0;				\
			     p_n.mark=0;				\
			     p_n.nlocks=2;			        \
			     CHKTRY(l1,p_n.locks[0]= *(DLOCK *)l1,1);	\
			     CHKTRY(l2,p_n.locks[1]= *(DLOCK *)l2,2);	\
			     qsort((char *)p_n.locks,2,sizeof(DLOCK),(int(*)())p_ex_cmplcks);\
			     if(setjmp(p_n.jmp)==0) {			\
			    	if(p_try(FL (struct EXCEPT_ELEMENT *)&p_n)) {
#define TRY3(l1,l2,l3)	   { struct EXCEPT_ELEMENT3 p_n;		\
			     p_n.flags=0;				\
			     p_n.mark=0;				\
			     p_n.nlocks=3;			        \
			     CHKTRY(l1,p_n.locks[0]= *(DLOCK *)l1,1);	\
			     CHKTRY(l2,p_n.locks[1]= *(DLOCK *)l2,2);	\
			     CHKTRY(l3,p_n.locks[2]= *(DLOCK *)l3,3);	\
			     qsort((char *)p_n.locks,3,sizeof(DLOCK),(int(*)())p_ex_cmplcks);\
			     if(setjmp(p_n.jmp)==0) {			\
			    	if(p_try(FL (struct EXCEPT_ELEMENT *)&p_n)) {
#define TRY4(l1,l2,l3,l4)  { struct EXCEPT_ELEMENT4 p_n;		\
			     p_n.flags=0;				\
			     p_n.mark=0;				\
			     p_n.nlocks=4;			        \
			     CHKTRY(l1,p_n.locks[0]= *(DLOCK *)l1,1);	\
			     CHKTRY(l2,p_n.locks[1]= *(DLOCK *)l2,2);	\
			     CHKTRY(l3,p_n.locks[2]= *(DLOCK *)l3,3);	\
			     CHKTRY(l4,p_n.locks[3]= *(DLOCK *)l4,4);	\
			     qsort((char *)p_n.locks,4,sizeof(DLOCK),(int(*)())p_ex_cmplcks);\
			     if(setjmp(p_n.jmp)==0) {			\
			    	if(p_try(FL (struct EXCEPT_ELEMENT *)&p_n)) {
#define TRY5(l1,l2,l3,l4,l5){ struct EXCEPT_ELEMENT5 p_n;		\
			     p_n.flags=0;				\
			     p_n.mark=0;				\
			     p_n.nlocks=5;			        \
			     CHKTRY(l1,p_n.locks[0]= *(DLOCK *)l1,1);	\
			     CHKTRY(l2,p_n.locks[1]= *(DLOCK *)l2,2);	\
			     CHKTRY(l3,p_n.locks[2]= *(DLOCK *)l3,3);	\
			     CHKTRY(l4,p_n.locks[3]= *(DLOCK *)l4,5);	\
			     CHKTRY(l5,p_n.locks[4]= *(DLOCK *)l5,4);	\
			     qsort((char *)p_n.locks,5,sizeof(DLOCK),(int(*)())p_ex_cmplcks);\
			     if(setjmp(p_n.jmp)==0) {			\
			    	if(p_try(FL (struct EXCEPT_ELEMENT *)&p_n)) {
				
#define TRYELSE			p_ex_pop(FL 1); } else {
#define TRYEND		        } 					\
			     } else {					\
				p_ex_raise(FL EXCEPTION);		\
			     }						\
			   }

/*
 * You can unlock any lock acquired by either a LOCK or a TRY
 * statement before the LOCKEND (TRYEND)
 */
#define UNLOCK(n)	do { CHKUNLOCK(n);p_unlock(FL (DLOCK*)(n)); } while(0)


/*
 * The rest of this header file is filled with prototype declarations
 * FORBID FORK_FUNC
 * (the other identifier are catched by reading the .o files)
 */

typedef void (*FORK_FUNC)(OB self,void *arg,GATE gate,int pos);

void p_lock(FLA struct EXCEPT_ELEMENT *);
int p_try(FLA struct EXCEPT_ELEMENT *);
void p_endlock(FLA struct EXCEPT_ELEMENT *);
void p_unlock(FLA DLOCK *lck);

struct EXCEPT_ELEMENT *p_ex_save_top(FLAO);
void p_ex_push(FLA struct EXCEPT_ELEMENT *);
void p_ex_pop(FLA int);
void p_ex_unwind(FLA short);
void p_ex_set_mark(FLA short);
void p_ex_remove_mark(FLA short);
DLOCK * p_get_lock(FLA struct LOCK_T_struct *lck,enum LOCK_TYPES type);

int p_ex_cmplcks(DLOCK *a,DLOCK *b);

void psather_start(FLA char *prog_name);
void psather_stop(FLAO);

void p_lock_v(void *);
void p_unlock_v(void *);

#ifndef  IN_PSATHER_LIBRARY
void p_ex_raise(FLA OB);
OB p_ex_exception(FLAO);

OB p_mutex_create(FLAO);

struct GATE_T_struct;
struct LOCK_T_struct *p_gate_create(FLA long size);
long p_gate_size(FLA struct GATE_T_struct *gate);
long p_gate_has_threads(FLA struct GATE_T_struct *gate);
void p_gate_set(FLA struct GATE_T_struct *gate,...);
OB p_gate_get(FLA struct GATE_T_struct *gate,...);
void p_gate_enqueue(FLA struct GATE_T_struct *gate,...);
OB p_gate_dequeue(FLA struct GATE_T_struct *gate,...);
void p_gate_clear(FLA struct GATE_T_struct *gate);
void p_gate_sync(FLA void *gate);


void p_attach(FLA FORK_FUNC function,OB self,void *arg,long size,void * gate,int position);

void *p_arg_alloc(FLA long size);
void p_arg_free(FLA char *ptr);
void *p_cohort(FLAO);

void p_cleared_set(FLA short value);
short p_cleared_get(FLAO);
void p_trap_clear_set(FLA short value);
short p_trap_clear_get(FLAO);
void p_defer(FLAO);

#endif

#endif
/* 
 * Copyright (C) International Computer Science Institute, 1994.  COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the SATHER LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/License" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email comments to "sather-bugs@icsi.berkeley.edu". 
 */

#ifndef _LOCKS_H_
#define _LOCKS_H_

#include <assert.h>
#include <thread.h>
#include <sys/errno.h>
#define IN_PSATHER_LIBRARY

#ifdef DEBUG_PSATHER
#  include <stdio.h>
   extern long seq_count;
#  define SEQ (seq_count++)
#  define PRFLT(file,line,type,what,with)	fprintf(stderr,"%ld\t%ld\t%s(%ld)\t%c_%ld\t%s\n",SEQ,(long)THREAD_GET_ID,file,(long)line,type,*(long *)(with+1),what)
#else /* DEBUG_PSATHER */
#  ifdef PSATHER_STATISTICS
#    include <stdio.h>
     extern long seq_count;
#    define SEQ (seq_count++)
#  endif /* PSATHER_STATISTICS */
#endif /* DEBUG_PSATHER */

#ifdef PSATHER_STATISTICS
#  define PRSTAT(type,what,with)        fprintf(stderr,"%ld\t%ld\t%s(%ld)\t%s_%ld\t%s\n",(long)(SEQ),(long)(THREAD_GET_ID),file,(long)line,type,(long)with,(what))
#else
#  define PRSTAT(type,what,with)
#endif

#ifdef DEBUG_PSATHER
#  define PRMFL(file,line,what,with)	PRFLT(file,line,'M',what,with)
#  define PRM(what,with)	PRMFL(__FILE__,__LINE__,what,with)
#  define LM_INIT		
#  define LM_TYPE		mutex_t *
#  define LM_CREATE	l_create(__FILE__,__LINE__)
LM_TYPE l_create(char *,long);
#  define LM_LOCK(x)    do { \
                                PRM("LOCK",(x));\
                                if(!mutex_trylock(x)) PRM("LOCK DONE IMMEDIATLY",(x));\
                                else {mutex_lock((x));PRM("LOCK DONE LATER",x);} \
                        } while(0)

#  define LM_TRY(x) 	l_try(x,__FILE__,__LINE__)
int l_try(LM_TYPE,char*,long);
#  define LM_UNLOCK(x)	do { PRM("UNLOCK",(x));mutex_unlock((x)); } while(0)
#  define LM_DELETE(x)	mutex_destroy((x))
#else /* DEBUG_PSATHER */
#  define PRMFL(a,b,c,d)
#  define PRM(a,b)
#  define LM_INIT		
#  define LM_TYPE		mutex_t *
#  define LM_CREATE	l_create()
LM_TYPE l_create(void);
#  define LM_LOCK(x) 	while(mutex_lock((x))==EINTR)
#  define LM_TRY(x) 	(!mutex_trylock((x)))
#  define LM_UNLOCK(x)	mutex_unlock((x))
#  define LM_DELETE(x)	mutex_destroy((x))
#endif /* DEBUG_PSATHER */

#define LTYPE		int
#define THREAD_ID	thread_t
#define NO_THREAD	0
#define THREAD_GET_ID	thr_self()

#define CLEAR_THREAD(id)	p_clear_thread(id)
#define CLEAR_ME		CLEAR_THREAD(THREAD_GET_ID)

typedef sema_t *SEMAPHORE_T;
#ifdef DEBUG_PSATHER
#  define PRSFL(file,line,what,with)	PRFLT(file,line,'S',what,with)
#  define PRS(what,with)	PRSFL(__FILE__,__LINE__,what,with)
#  define LSEMA_CREATE(c)	s_create(__FILE__,__LINE__,(c))
#  define LSEMA_DELETE(x) 	s_delete((x))
#  define LSEMA_WAIT(x) 	do { PRS("WAIT",x);while(sema_wait((x))==EINTR);PRS("WAIT DONE",x); } while(0)
#  define LSEMA_SIGNAL(x)	do { PRS("SIGNAL",x);sema_post((x)); } while(0)
SEMAPHORE_T s_create(char *,long,int);
void s_delete(char *,long,SEMAPHORE_T);
void s_post(char *,long,SEMAPHORE_T);
void s_wait(char *,long,SEMAPHORE_T);
#else
#  define PRSFL(a,b,c,d)
#  define PRS(a,b)
#  define LSEMA_CREATE(c)	s_create((c))
#  define LSEMA_DELETE(x) 	s_delete((x))
#  define LSEMA_WAIT(x) 	while(sema_wait((x))==EINTR);
#  define LSEMA_SIGNAL(x)	sema_post((x))
SEMAPHORE_T s_create(int);
void s_delete(SEMAPHORE_T);
void s_post(SEMAPHORE_T);
void s_wait(SEMAPHORE_T);
#endif

#define SET_CLEARED(v)          (get_tdata()->cleared=(v))
#define GET_CLEARED             (get_tdata()->cleared)
#define SET_TRAP_CLEAR(v)       do { (get_tdata()->trap_clear=(v));if((v) && GET_CLEARED) CLEAR_ME; } while(0)
#define GET_TRAP_CLEAR          (get_tdata()->trap_clear)
#define SET_CLEAR_FUNCTION(f)	(get_tdata()->clear_function=(f))
#define DEFER                   thr_yield()

#define SET_JUMP_POINTER(p)	(get_tdata()->jump_buf=(void*)(p)
#define GET_JUMP_POINTER	((jmp_buf *)get_tdata()->jump_buf)

#define STORE_COHORT(cohort) 	(get_tdata()->gate=(cohort))
#define GET_COHORT		(get_tdata()->gate)

#define GET_THREAD_LIST_POS     	(tlist_elem_t *)(get_tdata()->thread_list_pos)
#define SET_THREAD_LIST_POS(p)  	(get_tdata()->thread_list_pos=(p))
#define THREAD_LIST_ELEMENT_TYPE	THREAD_ID
#define GET_THREAD_LIST_ELEMENT         (THREAD_GET_ID)

#define EXCEPTION_STACK			((struct EXCEPT_ELEMENT *)(get_tdata()->exception_stack))

/*
 * BLOCK_CLEAR_SIGNAL creates a block, where no Clear Signal is delivered
 * before it finishes. The end of the block is marked by ENABLE_CLEAR_SIGNAL.
 * These two macros have to be at the same syntactic level, as they may open
 * a new C-block.
 * To prematurly end the block you can call BREAK_ENABLE_CLEAR_SIGNAL, iff
 * you are not in a loop or case statement.
 */
#define START_BLOCK_CLEAR_SIGNAL 	{ int i_C_s_;i_C_s_=get_tdata()->ignore_clear_signal;get_tdata()->ignore_clear_signal=1;
#define END_BLOCK_CLEAR_SIGNAL     	ENABLE_CLEAR_SIGNAL; }
#define ENABLE_CLEAR_SIGNAL	 	get_tdata()->ignore_clear_signal=i_C_s_;if(CLEAR_SIGNAL_ARRIVED && !i_C_s_) p_clear_thread(THREAD_GET_ID)
#define BLOCK_CLEAR_SIGNAL	 	get_tdata()->ignore_clear_signal=1
#define BREAK_ENABLE_CLEAR_SIGNAL	ENABLE_CLEAR_SIGNAL;break
#define CLEAR_SIGNAL_ARRIVED		get_tdata()->got_clear_signal

void p_clear_thread(THREAD_ID);



#endif

/* 
 * Copyright (C) International Computer Science Institute, 1994.  COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the SATHER LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/License" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email comments to "sather-bugs@icsi.berkeley.edu". 
 */
#ifndef _LONG_STACK_H_
#define _LONG_STACK_H_

typedef struct LONG_STACK_struct {
	struct LONG_STACK_struct *next;
	long l;
} *LONG_STACK;

#define ls_push(s,ll) 	ls_push_f(&(s),(ll))
#define ls_pop(s) 	ls_pop_f(&(s))
#define ls_is_empty(s) 	((s)==NULL)
#define ls_clear(s) 	ls_clear_f(&(s))
#define ls_create()	NULL

void ls_push_f(LONG_STACK *s,long ll);
long ls_pop_f(LONG_STACK *s);
void ls_clear_f(LONG_STACK *s);

#endif

/* 
 * Copyright (C) International Computer Science Institute, 1994.  COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the SATHER LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/License" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email comments to "sather-bugs@icsi.berkeley.edu". 
 */
#ifndef _THR_LONG_STACK_H_
#define _THR_LONG_STACK_H_


typedef struct THR_LONG_STACK_struct {
	struct THR_LONG_STACK_struct *next;
	THREAD_ID id;
	LONG_STACK unlocked;
} *THR_LONG_STACK;

#define tst_new_thread(s,id) tst_new_thread_f(&(s),id)
#define tst_del_thread(s,id) tst_del_thread_f(&(s),id)
#define tst_find_thread_stack(s,id) tst_find_thread_stack_f(&(s),id)
#define tst_create()	NULL
#define tst_clear(s)	tst_clear_f(&(s))

void tst_new_thread_f(THR_LONG_STACK *thr,THREAD_ID id);
void tst_del_thread_f(THR_LONG_STACK *thr,THREAD_ID id);
LONG_STACK tst_find_thread_stack_f(THR_LONG_STACK *thr,THREAD_ID id);
void tst_clear_f(THR_LONG_STACK *thr);

#endif

#ifndef _TLIST_LH_
#define _TLIST_LH_
/*
 * a double linked list of threads
 */
typedef struct tlist_elem_t_struct {
	struct  tlist_elem_t_struct *next,*prev;
	THREAD_LIST_ELEMENT_TYPE id;
} tlist_elem_t;

typedef struct tlist_t_struct {
	tlist_elem_t *head,*tail;
} tlist_t;

#define tlist_init(t)		tlist_init_p(&t)
#define tlist_enqueue(t,elem)	tlist_enqueue_p(&t,elem)
#define tlist_remove(t,pos)	tlist_remove_p(&t,pos)
#define tlist_empty(t)		((t).head!=NULL)
#define tlist_dequeue(t)	tlist_dequeue_p(&t)

void tlist_init_p(tlist_t *);
tlist_elem_t *tlist_enqueue_p(tlist_t *,THREAD_LIST_ELEMENT_TYPE);
void tlist_remove_p(tlist_t *,tlist_elem_t *);
THREAD_LIST_ELEMENT_TYPE tlist_dequeue_p(tlist_t *);

#endif
/* 
 * Copyright (C) International Computer Science Institute, 1994.  COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the SATHER LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/License" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email comments to "psather-bugs@icsi.berkeley.edu". 
 */

#ifndef _LOCKS_LH_
#define _LOCKS_LH_

#include <sys/types.h>
#include <sys/time.h>
#include <sys/timeb.h>
#include <stdio.h>


#define DLOCK_TAG	(-1)
#define MUTEX_TAG	(-2)
#define GATE_TAG	(-3)
#define GATEELEMENT_TAG (-6)

#define MAX_LOCKS 5  /* no more than 10 locks at a time */

struct LOCK_T_L {
	OB_HEADER header; /* used for the GC */
				/* a list of functions called when locking,
				 * unlocking or trylocking the lock.
				 * Additionally some functions to create and
				 * delete the lock
				 */
	struct LOCK_F_struct *fnct;
				/* bottom points to the first plate needing the
				 * lock and top to the last disk
				 * needing this lock
				 */
	struct LOCK_HOLE_struct *bottom,*top;
	/*enum LOCK_FLAGS flags;*/  /* some flags, see above		    */
	union {
		struct {
			THREAD_ID id; /* id of the thread holding this lock */
				/* a list of flags. Each flag 'knows' if the 
				 * last lock has been unlocked or not. This 
				 * info is needed when doing an endlock 
				 */
			LONG_STACK unlocked;
		} single;
		THR_LONG_STACK multiple; /* not used currently              */
	} unl;
	LM_TYPE hold;	/* Holding lock					    */
	unsigned long mark;
#ifdef DEBUG_PSATHER
	long number;
#else
#ifdef PSATHER_STATISTICS
	long number;
#endif
#endif
};

typedef struct LOCK_T_struct {
	struct LOCK_T_L l;
} *LOCK_T;

typedef int (*LOCK_FUNC)(void *,LTYPE);
typedef void (*LOCK_PROC)(void *,LTYPE);
typedef LOCK_T (*LOCK_FUNC_V)();

typedef struct LOCK_F_struct {
	LOCK_FUNC lock_free;
	LOCK_FUNC condition_ok;  /* must return one, if the conditions for
	                            this lock hold */
	LOCK_PROC unlock;
	LOCK_FUNC try;
	LOCK_FUNC_V create;
	LOCK_PROC delete;
#ifdef DEBUG_PSATHER
	void (*print)(void *);
#endif
} LOCK_F;

/*
 * GATE_ELEM_T is used to store the values enqueued into the gate
 */
typedef struct GATE_ELEM_T_struct{
	OB_HEADER header;
	struct GATE_ELEM_T_struct *next,*prev;
	void *arg;
} *GATE_ELEM_T;

typedef struct GATE_T_struct {
	struct LOCK_T_L l;	/* the lock structure			   */
	short lck;		/* lock used to lock the gate (1=locked)   */
	int threads;		/* number of threads attached to this gate */
	int sync;		/* number of threads doing a sync	   */
	SEMAPHORE_T sync_sem;	/* needed to synchronize threads	   */
	long enqueued;		/* number of enqueued values		   */
	long elem_size;		/* size of the elements to be enqueued 	   */
	GATE_ELEM_T bot,top;	/* stack of enqueued values 		   */
	tlist_t attached;   	/* list of attached threads 		   */
} *GATE_T;

typedef struct MUTEX_T_struct {
	struct LOCK_T_L l;	/* the lock structure			   */
	short lck;		/* the lock used to lock the mutex	   */
} *MUTEX_T;

typedef struct LOCK_HOLE_struct {
	DLOCK d;
	unsigned long mark;
	struct LOCK_PLATE_struct *myPlate;
	struct LOCK_HOLE_struct *above,*below;
} LOCK_HOLE;

#define L_K l_k
extern double l_k;

enum LOCK_PLATE_FLAGS {
	LOCK_TIMER_SIGNAL=1,
	LOCK_TIMER_RUNNING=2,
	LOCK_PLATE_ABOVE_MOVED=4,
	LOCK_CHANGE=8,
	LOCK_I_AM_PLATE_WITHOUT_TIMER=16,
	LOCK_PLATE_BELOW_MOVED=32,
	LOCK_PLATE_GOT_SIGNAL=64,
	LOCK_JUMP_TO_TOP=128,
	LOCK_TRY=256,
	LOCK_START_TIMER=512
};
typedef struct LOCK_PLATE_struct {
#ifdef DEBUG_PSATHER
	int index;
#endif
	THREAD_ID id;
	long size;
	struct timeval started_at,timer_started_at;
	enum LOCK_PLATE_FLAGS flags;
	SEMAPHORE_T restart_thread;
#ifdef LWP
	LOCK_TIMER ltimer;
	LM_TYPE ltimer_lck;
#endif
#ifdef DEADLOCK_DETECTION
	long check_id; /* used for dead lock detection */
#endif
	LOCK_HOLE hole[1];
} *LOCK_PLATE;

void p_mutex_init();
LOCK_F *p_mutex_functions(void);
void p_gate_init();
LOCK_F *p_gate_functions(void);

void p_ex_raise(FLA void *);
void *p_ex_exception(FLAO);

void p_gate_attach(FLA GATE_T);
void p_gate_detach(FLA GATE_T);
void p_gate_set(FLA GATE_T gate,void *ptr,int size);
void *p_gate_get(FLA GATE_T gate,void *ptr);
void p_gate_enqueue(FLA GATE_T gate,void *ptr,int size);
void *p_gate_dequeue(FLA GATE_T gate,void *ptr);
void p_gate_clear(FLA GATE_T gate);
void p_gate_sync(FLA GATE_T gate);
int p_gate_size(FLA GATE_T gate);
int p_gate_has_threads(FLA GATE_T gate);

void p_init_lock(LOCK_T lck);
	
void p_init_locks(void);
void p_ex_init_thread(FLAO);
extern void *p_clear_exception;

#ifdef DEADLOCK_DETECTION
extern long no_of_threads_running; /* used for deadlock detection */
extern long no_of_plates_enqueued;  
extern long no_of_plates_failed;
extern long check_id;
extern LM_TYPE deadlock;
#  ifdef NO_TIMER_IF_TOP_PLATE
extern long no_of_plates_without_timer;
#  endif
#endif

#ifdef PSATHER_RT_TRACE
extern FILE *rt_file;
extern LM_TYPE trace_lock;
extern unsigned long start_sec;
void print_trace(FLA int type,char start_stop,unsigned long id);
int print_begin_trace_line(FLA int type,char start_stop);
void print_statistics();
#define TR_START		'+'
#define TR_STOP			'-'
#define TR_DONE			'='
#define TR_DEADLOCK		0
#define TR_CHECK		1
#define TR_MUTEX_CREATED	10
#define TR_GATE_CREATED		11
#define TR_GATE_ENQUEUE		20
#define TR_GATE_DEQUEUE		21
#define TR_GATE_SET		22
#define TR_GATE_GET		23
#define TR_GATE_SYNC		24
#define TR_GATE_CLEAR		25
#define TR_GATE_ATTACH		26
#define TR_GATE_DETACH		27
#define TR_THREAD		28
#define TR_ACQUIRED 		30
#define TR_RELEASED 		31
#define TR_WAKEUP		35
#define TR_LOCK			50
#define TR_LOCK_RESERVE		51
#define TR_TRY			52
#define TR_TRY_FAILED		53
#define TR_ENDLOCK		55
#define TR_UNLOCK		58
#define TR_ATTACH		60
#define TR_TIMER_START		70
#define TR_TIMER_STOP		71
#define TR_TIMER_RING		72
#define TR_PLATE_SPEAR		80
#define TR_PLATE_MOVE_TOP	81
#define TR_PLATE_REMOVE		82
#define TR_STICK_RESERVE	90
#define TR_STICK_GIVEUP		91
#define TR_LAST 		100
#else
#define print_trace(type,start_stop,id)
#define print_statistics()
#endif

#endif
#ifndef _TLIST_LH_
#define _TLIST_LH_
/*
 * a double linked list of threads
 */
typedef struct tlist_elem_t_struct {
	struct  tlist_elem_t_struct *next,*prev;
	THREAD_LIST_ELEMENT_TYPE id;
} tlist_elem_t;

typedef struct tlist_t_struct {
	tlist_elem_t *head,*tail;
} tlist_t;

#define tlist_init(t)		tlist_init_p(&t)
#define tlist_enqueue(t,elem)	tlist_enqueue_p(&t,elem)
#define tlist_remove(t,pos)	tlist_remove_p(&t,pos)
#define tlist_empty(t)		((t).head!=NULL)
#define tlist_dequeue(t)	tlist_dequeue_p(&t)

void tlist_init_p(tlist_t *);
tlist_elem_t *tlist_enqueue_p(tlist_t *,THREAD_LIST_ELEMENT_TYPE);
void tlist_remove_p(tlist_t *,tlist_elem_t *);
THREAD_LIST_ELEMENT_TYPE tlist_dequeue_p(tlist_t *);

#endif
#ifndef _TDATA_H_
#define _TDATA_H_


typedef struct TDATA_struct {
	short cleared;
	short trap_clear;
	void *jump_buf;
	GATE_T gate;
	void (*clear_function)(void);
	void *thread_list_pos;
	short ignore_clear_signal;
	short got_clear_signal;
	void *exception_stack;
	struct LOCK_TIMER_struct *ti;
} *TDATA;

extern thread_key_t tdata;

void free_tdata(void);
void create_tdata(void);
TDATA get_tdata(void);

#endif
/* 
 * Copyright (C) International Computer Science Institute, 1994.  COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the SATHER LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/License" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email comments to "sather-bugs@icsi.berkeley.edu". 
 */
#include <assert.h>
#include <signal.h>

#define NEW_THREAD(func,stack_size,arg,cluster) thr_create(NULL,stack_size,(void *(*)(void *))func,arg,0,NULL);
#define START_THREAD(arg) void start_thread(arg)
#define INIT_THREAD	create_tdata();set_signal_handler();
#define END_THREAD	free_tdata()


void clear_handler(void)
{
	signal(SIGUSR1,(void (*)(int))clear_handler);
	if(!get_tdata()->ignore_clear_signal) p_clear_thread(THREAD_GET_ID);
	else get_tdata()->got_clear_signal=1;
}

void set_signal_handler(void)
{
	struct sigaction act;
	act.sa_sigaction=(void (*)(int,siginfo_t *,void *))clear_handler;
	sigemptyset(&act.sa_mask);
	act.sa_flags=SA_SIGINFO|SA_RESTART;
	assert(sigaction(SIGUSR1,&act,NULL)==0);
}


void p_clear_thread(THREAD_ID id)
{
	if(id==THREAD_GET_ID) {
		get_tdata()->got_clear_signal=0;
		if(get_tdata()->cleared==0) {
			get_tdata()->cleared=1;
			if(get_tdata()->trap_clear) RAISE(p_clear_exception);
		}
	} else thr_kill(id,SIGUSR1);
}
/* 
 * Copyright (C) International Computer Science Institute, 1994.  COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the SATHER LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/License" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email comments to "sather-bugs@icsi.berkeley.edu". 
 */

#include <thread.h>
#include <signal.h>
#include <string.h>

#ifdef DEBUG_PSATHER
long seq_count=0;
#else
#ifdef PSATHER_STATISTICS
long seq_count=1;
#endif
#endif

#ifdef DEBUG_PSATHER
static long sema_number=1;
sema_t *s_create(char *file,long line,int c)
{
	sema_t *n;
	n=(sema_t *)GC_malloc(sizeof(sema_t)+sizeof(long));
	sema_init(n,c,USYNC_THREAD,NULL);
	*(long *)(n+1)=sema_number++;
	PRSFL(file,line,"CREATE",n);
	return n;
}
#else /* DEBUG_PSATHER */
sema_t *s_create(int c)
{
	sema_t *n;
	n=(sema_t *)GC_malloc(sizeof(sema_t));
	sema_init(n,c,USYNC_THREAD,NULL);
	return n;
}
#endif /* DEBUG_PSATHER */

void s_delete(sema_t *s)
{
	sema_destroy(s);
	GC_free(s);
}

#ifdef DEBUG_PSATHER
int l_try(mutex_t *x,char *file,long line)
{
	int i;
	PRMFL(file,line,"TRY",x);
	i=(!mutex_trylock((x)));
	if(i) PRMFL(file,line,"TRY SUCC",x);
	else PRMFL(file,line,"TRY FAIL",x);
	return i;
}
#endif

#ifdef DEBUG_PSATHER
static long mutex_number=1;
mutex_t *l_create(char *file,long line)
{
	mutex_t *x;
	x=(mutex_t *)GC_malloc(sizeof(mutex_t)+sizeof(long));
	*(long *)(x+1)=mutex_number++;
	mutex_init(x,USYNC_THREAD,NULL);
	PRMFL(file,line,"CREATE",x);
	return x;
}
#else
mutex_t *l_create()
{
	mutex_t *x;
	x=(mutex_t *)GC_malloc(sizeof(mutex_t));
	mutex_init(x,USYNC_THREAD,NULL);
	return x;
}
#endif

#define addtmv(a,b) 	do { (a).tv_sec+=(b).tv_sec;(a).tv_usec+=(b).tv_usec;if((a).tv_usec>1000000) { (a).tv_usec-=1000000;(a).tv_sec++; } } while(0)
#define subtmv(a,b) 	do { (a).tv_sec-=(b).tv_sec;(a).tv_usec-=(b).tv_usec;if((a).tv_usec<0) { (a).tv_usec+=1000000;(a).tv_sec--; } } while(0)
#define cmptmvi(a,b) 	(((a)>(b))?1:((a)<(b)?-1:0))
#define cmptmv(a,b)  	((a).tv_sec==(b).tv_sec?cmptmvi((a).tv_usec,(b).tv_usec):cmptmvi((a).tv_sec,(b).tv_sec))
#define cptmv(a,b) 	a=b
#define usectmv(a) 	((a).tv_sec*1000000+(a).tv_usec)
#define settmvusec(a,s) do {a.tv_sec=s/100000;a.tv_usec=s%1000000; } while(0)

#define START_TIMER(plate)	p_start_timer(FLP plate)
#define STOP_TIMER(plate)	p_stop_timer(FLP plate)
int p_start_timer(FLA LOCK_PLATE);
void p_stop_timer(FLA LOCK_PLATE);


/* 
 * Copyright (C) International Computer Science Institute, 1994.  COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the SATHER LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/License" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email comments to "sather-bugs@icsi.berkeley.edu". 
 */
#include <assert.h>

thread_key_t tdata;

extern void p_init_locks(void);
extern void p_init_timer_thread(void);
static void init_lock_v();

char *bin_prog_name=NULL;
void psather_start(FLA char *prog_name)
{
	extern void set_signal_handler(void);
	bin_prog_name=prog_name;
	set_signal_handler();
	thr_keycreate(&tdata,(void (*)(void *))free_tdata);

	create_tdata();

	p_init_timer_thread();
	p_init_locks();
	init_lock_v();
	print_trace(FLP TR_THREAD,TR_START,0);
}

void psather_stop(FLAO)
{
	print_trace(FLP TR_THREAD,TR_STOP,0);
	print_statistics();
}

mutex_t vlocks[256];
static void init_lock_v()
{
	int i;
	for(i=0;i<256;i++)
		mutex_init(&vlocks[i],USYNC_THREAD,0);
}

void p_lock_v(void *v)
{
	mutex_lock(&vlocks[((unsigned long)v>>3)&0xff]);
}

void p_unlock_v(void *v)
{
	mutex_unlock(&vlocks[((unsigned long)v>>3)&0xff]);
}

extern thread_key_t tdata;

void free_tdata(void) { GC_free(get_tdata()); }

void create_tdata(void) 
{
	TDATA n;
	n=GC_malloc(sizeof(struct TDATA_struct));
	n->gate=NULL;
	n->cleared=0;
	n->trap_clear=1;
	n->clear_function=NULL;
	n->thread_list_pos=NULL;
	n->ignore_clear_signal=0;
	n->got_clear_signal=0;
	n->exception_stack=NULL;
	n->ti=NULL;
	thr_setspecific(tdata,n);
}

TDATA get_tdata()
{
	TDATA n;
	thr_getspecific(tdata,(void **)&n);
	return n;
}

#include <unistd.h>

#define addtmv(a,b) 	do { (a).tv_sec+=(b).tv_sec;(a).tv_usec+=(b).tv_usec;if((a).tv_usec>1000000) { (a).tv_usec-=1000000;(a).tv_sec++; } } while(0)
#define subtmv(a,b) 	do { (a).tv_sec-=(b).tv_sec;(a).tv_usec-=(b).tv_usec;if((a).tv_usec<0) { (a).tv_usec+=1000000;(a).tv_sec--; } } while(0)
#define cmptmvi(a,b) 	(((a)>(b))?1:((a)<(b)?-1:0))
#define cmptmv(a,b)  	((a).tv_sec==(b).tv_sec?cmptmvi((a).tv_usec,(b).tv_usec):cmptmvi((a).tv_sec,(b).tv_sec))
#define cptmv(a,b) 	a=b
#define usectmv(a) 	((a).tv_sec*1000000+(a).tv_usec)
#define settmvusec(a,s) do {a.tv_sec=s/100000;a.tv_usec=s%1000000; } while(0)

#define START_TIMER(plate)	p_start_timer(FLP plate)
#define STOP_TIMER(plate)	p_stop_timer(FLP plate)

typedef struct TIMER_struct {
	struct TIMER_struct *next;
	LOCK_PLATE plate;
	struct timeval tv;
	THREAD_ID id;
} *TIMER;

TIMER timer_first=NULL;
TIMER timer_done=NULL;
SEMAPHORE_T timer_semaphore;
LM_TYPE timer_lock;
timer_t timer_id;
thread_t timer_thread_id;

extern void WakeupPlate(LOCK_PLATE p,enum LOCK_PLATE_FLAGS why);
extern int gettimeofday(struct timeval *,struct timezone *);
static void handle_timer_call()
{
        struct timeval now;
	struct itimerspec its;
	TIMER t;

	LM_LOCK(timer_lock);

	while(timer_first!=NULL) {
		gettimeofday(&now,NULL);
		if(cmptmv(now,timer_first->tv)<0) break;

		WakeupPlate(timer_first->plate,LOCK_TIMER_SIGNAL);

		/*
		 * as free may not be used in a signal handler, we have
		 * to store obsolete timers and have them deleted elsewhere
		 */
		t=timer_first;
		timer_first=t->next;
		t->next=timer_done;
		timer_done=t;
	}
	
	if(timer_first!=NULL) {
		its.it_interval.tv_sec=0;
		its.it_interval.tv_nsec=0;
		its.it_value.tv_sec=timer_first->tv.tv_sec;
		its.it_value.tv_nsec=timer_first->tv.tv_usec*1000;
		assert(timer_settime(timer_id,TIMER_ABSTIME,&its,NULL)==0);
	}

	LM_UNLOCK(timer_lock);
}

static void timer_thread()
{
	struct sigevent evp;
	LM_TYPE my_lock;	
	struct sigaction act;

	timer_lock=LM_CREATE;
	timer_done=NULL;
	timer_first=NULL;

	evp.sigev_notify=SIGEV_SIGNAL;
	evp.sigev_signo=SIGALRM;
	evp.sigev_value.sival_int=1;
	assert(timer_create(CLOCK_REALTIME,&evp,&timer_id)==0);

	act.sa_sigaction=(void (*)(int,siginfo_t *,void *))handle_timer_call;
	sigemptyset(&act.sa_mask);
	act.sa_flags=SA_SIGINFO;
	assert(sigaction(SIGALRM,&act,NULL)==0);
		        
	LSEMA_SIGNAL(timer_semaphore);

#if 0  /* does not work, as it simple blocks the whole program !!!!! */
	while(1) pause(); /* sleep forever */
#endif 
	my_lock=LM_CREATE;
	LM_LOCK(my_lock);
	LM_LOCK(my_lock);
}

void p_init_timer_thread()
{
	timer_semaphore=LSEMA_CREATE(0);

	thr_create(NULL,0,(void *(*)(void *))timer_thread,NULL,
	     THR_BOUND|THR_DAEMON|THR_NEW_LWP,&timer_thread_id);

	LSEMA_WAIT(timer_semaphore);
	LSEMA_DELETE(timer_semaphore);
}

void p_stop_timer(FLA LOCK_PLATE p)
{
	TIMER t,t1;

	LM_LOCK(timer_lock);
	if(timer_first!=NULL && p->flags&LOCK_TIMER_RUNNING) {
		if(timer_first->id==THREAD_GET_ID) {
			t=timer_first;
			timer_first=t->next;
			free(t);
		} else {
			t=timer_first;
			while(t->next!=NULL) 
				if(t->next->id==THREAD_GET_ID) {
					t1=t->next;
					t->next=t1->next;
					free(t1);
					break;
				} else t=t->next;
		}
	}
	/*
	 * delete all obsolete timers
	 */
	while(timer_done!=NULL) {
		t=timer_done;
		timer_done=timer_done->next;
		free(t);
	}
	LM_UNLOCK(timer_lock);
        p->flags&= ~(LOCK_TIMER_RUNNING);
}

/* 
 * returns 0 if the timer is not started because it has rang in the past.
 * In this case NO signal at all will be sent to the calling thread,  and
 * it is his responsability to hande this case correctly.
 */
int p_start_timer(FLA LOCK_PLATE p)
{
        struct timeval now,cur,t;
        unsigned long age;
	TIMER ti;

	gettimeofday(&now,NULL); 

        /* calculate age in usec */
	cptmv(t,now);
	subtmv(t,p->started_at);
	age=usectmv(t);
        age=(double)(age)*(double)(L_K);
        if(age<=100) age=100; /* do not wait less than 100 */

	settmvusec(t,age);
	cptmv(now,p->started_at);
	addtmv(now,t);
	gettimeofday(&cur,NULL);
	if(cmptmv(now,cur)<0) {
		return 0;
	}
	cptmv(t,now);
	subtmv(t,cur);
	if((usectmv(t))<5000) {
		return 0; /* no timer that will 
			   ring in less than  5msec */
	}

        p_stop_timer(FLP p);
        p->flags|=LOCK_TIMER_RUNNING;
	p->flags&= ~LOCK_TIMER_SIGNAL;

#ifdef PSATHER_RT_TRACE
	if(print_begin_trace_line(FLP TR_TIMER_START,TR_DONE)) {
		fprintf(rt_file," %ld.%06ld\n",now.tv_sec-start_sec,now.tv_usec);
		fflush(rt_file);
	}
	LM_UNLOCK(trace_lock);
#endif
	LM_LOCK(timer_lock);


	ti=(struct TIMER_struct *)malloc(sizeof(struct TIMER_struct));
	ti->plate=p;
	ti->id=THREAD_GET_ID;
	ti->next=NULL;
	ti->tv=now;
	if(timer_first==NULL) {
		timer_first=ti;
		thr_kill(timer_thread_id,SIGALRM);
	} else if(cmptmv(now,timer_first->tv)<0) {
		ti->next=timer_first;
		timer_first=ti;
		thr_kill(timer_thread_id,SIGALRM);
	} else {
		TIMER t1=timer_first;
		while(t1->next!=NULL) {
			if(cmptmv(now,t1->next->tv)<0) {
				ti->next=t1->next;
				t1->next=ti;
				break;
			}
			t1=t1->next;
		}
		if(t1->next==NULL) t1->next=ti;
	}
	
	LM_UNLOCK(timer_lock);
	return 1;
}


/* 
 * Copyright (C) International Computer Science Institute, 1994.  COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the SATHER LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/License" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email comments to "sather-bugs@icsi.berkeley.edu". 
 */
#include <assert.h>
#include <stddef.h>
#include <stdio.h>
#include <setjmp.h>

struct fork_t {
	FORK_FUNC func;
	GATE_T gate;
	LM_TYPE lck;
	void *self;
	int pos;
	void *arg;
#ifdef DEBUG
	char *file;
	long line;
#endif
};

static START_THREAD(struct fork_t *f)
{
	GATE_T gate;
#ifdef DEBUG
	char *file;
	long line;
	file=f->file;
	line=f->line;
#endif
	INIT_THREAD;
	print_trace(FLP TR_THREAD,TR_START,0);
	p_ex_init_thread(FLPO);
	p_gate_attach(FLP gate=f->gate);
	PROTECT_R_BEGIN
#ifdef DEADLOCK_DETECTION
		LM_LOCK(deadlock);
		no_of_threads_running++;
		LM_UNLOCK(deadlock);
#endif
		LM_UNLOCK(f->lck); 
			   /* restart the thread which created this thread,
	                      the other thread will now unlock the gate.
			      There is no way a clear could happen before
			      this statement */
		(*f->func)(f->self,f->arg,(GATE)p_get_lock(FLP (struct LOCK_T_struct *)gate,L_MUTEX),f->pos);
	PROTECT_R_WHEN
	PROTECT_R_END

	p_gate_detach(FLP gate);
#ifdef DEADLOCK_DETECTION
	LM_LOCK(deadlock);
	no_of_threads_running--;
	LM_UNLOCK(deadlock);
#endif
	print_trace(FLP TR_THREAD,TR_STOP,0);
	free(f);
	END_THREAD;
}


void p_attach(FLA FORK_FUNC func,void *self,void *args,long s_args,GATE_T gate,int cid)
{
	struct fork_t *f;
	LM_TYPE lck;

	f=malloc(sizeof(struct fork_t));
	f->func=func;
	f->gate=gate;
	f->self=self;
	f->pos=cid;
	lck=f->lck=LM_CREATE;
	f->arg=args;
#ifdef DEBUG
	f->file=file;
	f->line=line;
#endif
	LM_LOCK(lck);

	LOCK_RESERVE1(p_get_lock(FLP (struct LOCK_T_struct *)gate,L_MUTEX))
		NEW_THREAD((void (*)())start_thread,200000,(char **)f,cid);

		LM_LOCK(lck);
		LM_DELETE(lck);
	LOCK_R_END
}

void p_cleared_set(FLA short value)
{
	SET_CLEARED(value);
}

short p_cleared_get(FLAO)
{
	return GET_CLEARED;
}

void p_trap_clear_set(FLA short value)
{
	SET_TRAP_CLEAR(value);
}

short p_trap_clear_get(FLAO)
{
	return GET_TRAP_CLEAR;
}

void p_defer(FLAO)
{
	DEFER;
}




void tlist_init_p(tlist_t *t) 
{
	t->head=NULL;
	t->tail=NULL;
}
	
tlist_elem_t *tlist_enqueue_p(tlist_t *t,THREAD_LIST_ELEMENT_TYPE id)
{
	tlist_elem_t *n;
	n=malloc(sizeof(tlist_elem_t));
	n->prev=NULL;
	n->next=t->head;
	t->head=n;
	if(n->next==NULL) t->tail=n;
	n->id=id;
	return n;
}

void tlist_remove_p(tlist_t *t,tlist_elem_t *elem)
{
	if(elem->prev==NULL) t->head=elem->next;
	else elem->prev->next=elem->next;
	if(elem->next==NULL) t->tail=elem->prev;
	else elem->next->prev=elem->prev;
}	

THREAD_LIST_ELEMENT_TYPE tlist_dequeue_p(tlist_t *t)
{
	THREAD_LIST_ELEMENT_TYPE id=0;
	if(t->head!=NULL) {
		id=t->head->id;
		tlist_remove(*t,t->head);
	}
	return id;
}

/* 
 * Copyright (C) International Computer Science Institute, 1994.  COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the SATHER LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/License" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email comments to "sather-bugs@icsi.berkeley.edu". 
 */

#include <stdio.h>

static LOCK_F mutex_functions;

#ifdef DEBUG_PSATHER
void p_mutex_print(MUTEX_T mtx)
{
	fprintf(stdout,"LOCK %ld: uses MUTEX_LOCKS M_%ld and M_%ld  (helper)\n",
		mtx->l.number,*(long *)(mtx->lck+1),*(long *)(mtx->l.hold+1));
}
#endif

LOCK_F * p_mutex_functions()
{
	return &mutex_functions;
}


MUTEX_T p_mutex_create(FLAO)
{
	MUTEX_T x;

	x=(MUTEX_T)malloc(sizeof(struct MUTEX_T_struct));
	p_init_lock((LOCK_T)x);
	x->l.fnct=&mutex_functions;
	x->l.header.tag=MUTEX_TAG;
	x->lck=0;
	print_trace(FLP TR_MUTEX_CREATED,TR_DONE,(long)x);
	return x;
}

void p_mutex_lock(MUTEX_T lck)
{
	assert(lck->lck==0);
	lck->lck=1;
}
int p_mutex_try(MUTEX_T lck)
{
	if(lck->lck) return 0;
	lck->lck=1;
	return 1;
}
int p_mutex_free(MUTEX_T lck)
{
	return lck->lck==0;
}
int p_mutex_cond(MUTEX_T lck)
{
	return 1;
}
void p_mutex_unlock(MUTEX_T lck)
{
	assert(lck->lck==1);
	lck->lck=0;
}
void p_mutex_delete(MUTEX_T lck)
{
}

void p_mutex_init()
{
	LM_INIT;
	mutex_functions.lock_free=(LOCK_FUNC)p_mutex_free;
	mutex_functions.condition_ok=(LOCK_FUNC)p_mutex_cond;
	mutex_functions.unlock=(LOCK_PROC)p_mutex_unlock;
	mutex_functions.try=(LOCK_FUNC)p_mutex_try;
	mutex_functions.create=(LOCK_FUNC_V)p_mutex_create;
	mutex_functions.delete=(LOCK_PROC)p_mutex_delete;
#ifdef DEBUG_PSATHER
	mutex_functions.print=(void (*)(void *))p_mutex_print;
#endif
}


/* 
 * Copyright (C) International Computer Science Institute, 1994.  COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the SATHER LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/License" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email comments to "sather-bugs@icsi.berkeley.edu". 
 */

#include <stdio.h>
#include <assert.h>

#ifdef PSATHER_CHK
#define CHKGATEVOID(g)	if((g)==NULL) RFATALp("using a void gate")
#else
#define CHKGATEVOID(g)
#endif

static LOCK_F gate_functions;

#define GATELOCK(gate,type)	LOCK_RESERVE1(p_get_lock(FL (struct LOCK_T_struct *)gate,type))

#ifdef DEBUG_PSATHER
void p_gate_print(GATE_T mtx)
{
	fprintf(stderr,"LOCK %ld: uses GATE_LOCKS M_%ld and M_%ld  (helper)\n",
		(long)mtx->l.number,*(long *)(mtx->lck+1),*(long *)(mtx->l.hold+1));
}
#endif

LOCK_F * p_gate_functions()
{
	return &gate_functions;
}


GATE_T p_gate_create(FLA long size)
{
	GATE_T x;

	x=(GATE_T)malloc(sizeof(struct GATE_T_struct));
	p_init_lock((LOCK_T)x);
	x->l.fnct=&gate_functions;
	x->l.header.tag=GATE_TAG;
	x->lck=0;
	x->threads=0;
	x->enqueued=0;
	x->sync=0;
	x->sync_sem=LSEMA_CREATE(0);
	x->elem_size=size;
	x->bot=NULL;
	x->top=NULL;
	tlist_init(x->attached);
	print_trace(FLP TR_GATE_CREATED,TR_DONE,(long)x);
	return x;
}

static int p_gate_cond(GATE_T lck,LTYPE type)
{
	if(type&L_NO_THREADS && lck->threads)    return 0;
	if(type&L_THREADS && lck->threads==0)    return 0;
	if(type&L_NOT_EMPTY && lck->enqueued==0) return 0;
	if(type&L_EMPTY && lck->enqueued) 	 return 0;
	return 1;
}
static int p_gate_free(GATE_T lck,LTYPE type)
{
	return lck->lck==0;
}

static int p_gate_try(GATE_T lck,LTYPE type)
{
	if(p_gate_cond(lck,type) && p_gate_free(lck,type)) {
		lck->lck=1;
		return 1;
	} else return 0;
}

static void p_gate_unlock(GATE_T lck,LTYPE type)
{
	assert(lck->lck==1);
	lck->lck=0;
}

void p_gate_delete(GATE_T lck,LTYPE type)
{
	free(lck);
}
void p_gate_init()
{
	LM_INIT;
	gate_functions.condition_ok=(LOCK_FUNC)p_gate_cond;
	gate_functions.lock_free=(LOCK_FUNC)p_gate_free;
	gate_functions.unlock=(LOCK_PROC)p_gate_unlock;
	gate_functions.try=(LOCK_FUNC)p_gate_try;
	gate_functions.create=(LOCK_FUNC_V)p_gate_create;
	gate_functions.delete=(LOCK_PROC)p_gate_delete;
#ifdef DEBUG_PSATHER
	gate_functions.print=(void (*)(void *))p_gate_print;
#endif
}


void *p_gate_get(FLA GATE_T gate,void *rarg)
{
	void *volatile ret=NULL;
	
	CHKGATEVOID(gate);
	print_trace(FLP TR_GATE_GET,TR_START,(long)gate);
	PRSTAT("GATE","GET",gate->l.number);

	GATELOCK(gate,L_NOT_EMPTY)
		START_BLOCK_CLEAR_SIGNAL
			ret=NULL;
			if(rarg==NULL && gate->elem_size>0 && gate->elem_size<=sizeof(void *)) {
				ret=gate->bot->arg;
			} else if(gate->elem_size>0) {
				assert(rarg!=NULL);
				memcpy(rarg,&(gate->bot->arg),gate->elem_size);
			}
		END_BLOCK_CLEAR_SIGNAL
	LOCK_R_END
	PRSTAT("GATE","GET DONE",gate->l.number);
	print_trace(FLP TR_GATE_GET,TR_STOP,(long)gate);
	return (void *)ret;
}

void *p_gate_dequeue(FLA GATE_T gate,void *rarg)
{
	void *volatile ret=NULL;
	CHKGATEVOID(gate);
	PRSTAT("GATE","DEQUEUE",gate->l.number);
	print_trace(FLP TR_GATE_DEQUEUE,TR_START,(long)gate);

	GATELOCK(gate,L_NOT_EMPTY)
		START_BLOCK_CLEAR_SIGNAL
			ret=NULL;
			gate->enqueued--;
			if(rarg==NULL && gate->elem_size>0 && gate->elem_size<=sizeof(void *)) {
				GATE_ELEM_T o;
				o=gate->bot;
				gate->bot=o->next;
				if(gate->bot==NULL) gate->top=NULL;
				else gate->bot->prev=NULL;
				ret=o->arg;
				free(o);
			} else if(gate->elem_size>0) {
				GATE_ELEM_T o;
				assert(rarg!=NULL);
				o=gate->bot;
				gate->bot=o->next;
				if(gate->bot==NULL) gate->top=NULL;
				else gate->bot->prev=NULL;
				memcpy(rarg,&(o->arg),gate->elem_size);
				free(o);
			}
		END_BLOCK_CLEAR_SIGNAL
	LOCK_R_END
	print_trace(FLP TR_GATE_DEQUEUE,TR_STOP,(long)gate);
	PRSTAT("GATE","DEQUEUE DONE",gate->l.number);
	return (void *)ret;
}

void p_gate_set(FLA GATE_T gate,void *elem,int size)
{
	CHKGATEVOID(gate);
	PRSTAT("GATE","ENQUEUE",gate->l.number);
	print_trace(FLP TR_GATE_SET,TR_START,(long)gate);
	GATELOCK(gate,L_MUTEX)
		START_BLOCK_CLEAR_SIGNAL
			if(gate->enqueued==0) {
				gate->enqueued++;
				if(!size && gate->elem_size>0 && gate->elem_size<=sizeof(void *)) {
					GATE_ELEM_T n;
					n=malloc(sizeof(struct GATE_ELEM_T_struct));
					n->header.tag=GATEELEMENT_TAG;
					n->arg=elem;
					n->next=n->prev=NULL;
					gate->bot=gate->top=n;
				} else if(gate->elem_size>0) {
					GATE_ELEM_T n;
					assert(size==gate->elem_size);
					n=malloc(sizeof(struct GATE_ELEM_T_struct)-sizeof(void *)+gate->elem_size);
					memcpy(&(n->arg),elem,gate->elem_size);
					n->header.tag=GATEELEMENT_TAG;
					n->next=n->prev=NULL;
					gate->bot=gate->top=n;
				}
			} else {
				if(!size && gate->elem_size>0 && gate->elem_size<=sizeof(void *)) {
					gate->bot->arg=elem;
				} else if(gate->elem_size>0) {
					assert(gate->elem_size==size);
					memcpy(&(gate->bot->arg),elem,gate->elem_size);
				}
			}
		END_BLOCK_CLEAR_SIGNAL
	LOCK_R_END
	print_trace(FLP TR_GATE_SET,TR_STOP,(long)gate);
	PRSTAT("GATE","ENQUEUE DONE",gate->l.number);
}

void p_gate_enqueue(FLA GATE_T gate,void *elem,int size)
{
	CHKGATEVOID(gate);
	PRSTAT("GATE","ENQUEUE",gate->l.number);
	print_trace(FLP TR_GATE_ENQUEUE,TR_START,(long)gate);
	GATELOCK(gate,L_MUTEX)
		START_BLOCK_CLEAR_SIGNAL
			gate->enqueued++;
			if(!size && gate->elem_size>0 && gate->elem_size<=sizeof(void *)) {
				GATE_ELEM_T n;
				n=malloc(sizeof(struct GATE_ELEM_T_struct));
				if(gate->top==NULL) {
					n->next=n->prev=NULL;
					gate->bot=gate->top=n;
				} else {
					n->prev=gate->top;
					n->next=NULL;
					gate->top=n;
					n->prev->next=n;
				}
				n->header.tag=GATEELEMENT_TAG;
				n->arg=elem;
			} else if(gate->elem_size>0) {
				GATE_ELEM_T n;
				assert(size==gate->elem_size);
				n=malloc(sizeof(struct GATE_ELEM_T_struct)-sizeof(void *)+gate->elem_size);
				memcpy(&(n->arg),elem,gate->elem_size);
				if(gate->top==NULL) {
					n->next=n->prev=NULL;
					gate->bot=gate->top=n;
				} else {
					n->prev=gate->top;
					n->next=NULL;
					gate->top=n;
					n->prev->next=n;
				}
				n->header.tag=GATEELEMENT_TAG;
			}
		END_BLOCK_CLEAR_SIGNAL
	LOCK_R_END
	print_trace(FLP TR_GATE_ENQUEUE,TR_STOP,(long)gate);
	PRSTAT("GATE","ENQUEUE DONE",gate->l.number);
}

void p_gate_attach(FLA GATE_T gate)
{
	/* this function is exclusiv, but there are no locks used.
	 * This must be done this way, because the thread which forked
	 * this thread is still locking the gate and will lock it, until
	 * p_gate_attach() returns (see start_thread()).
	 */
	/* p_lock(1,gate,L_MUTEX); */
	CHKGATEVOID(gate);
	PRSTAT("GATE","ATTACH",gate->l.number);
	print_trace(FLP TR_GATE_ATTACH,TR_START,(long)gate);
	START_BLOCK_CLEAR_SIGNAL
		gate->threads++;
		SET_THREAD_LIST_POS(tlist_enqueue(gate->attached,GET_THREAD_LIST_ELEMENT));
		STORE_COHORT(gate);
		PRSTAT("GATE","ATTACH DONE",gate->l.number);
		print_trace(FLP TR_GATE_ATTACH,TR_STOP,(long)gate);
	END_BLOCK_CLEAR_SIGNAL
}

void p_gate_detach(FLA GATE_T gate)
{
	CHKGATEVOID(gate);
	print_trace(FLP TR_GATE_DETACH,TR_START,(long)gate);
	PRSTAT("GATE","DETACH",gate->l.number);
	GATELOCK(gate,L_MUTEX)
		START_BLOCK_CLEAR_SIGNAL
			gate->threads--;
			tlist_remove(gate->attached,GET_THREAD_LIST_POS);
			if(gate->sync==gate->threads) {
				while(gate->sync--) LSEMA_SIGNAL(gate->sync_sem);
				gate->sync=0;
			}
		END_BLOCK_CLEAR_SIGNAL
	LOCK_R_END
	print_trace(FLP TR_GATE_DETACH,TR_STOP,(long)gate);
	PRSTAT("GATE","DETACH DONE",gate->l.number);
}

void p_gate_sync(FLA GATE_T gate)
{
	volatile int do_wait=0;
	CHKGATEVOID(gate);
	print_trace(FLP TR_GATE_SYNC,TR_START,(long)gate);
	PRSTAT("GATE","SYNC",gate->l.number);
	GATELOCK(gate,L_MUTEX)
		START_BLOCK_CLEAR_SIGNAL
			if(gate->sync==gate->threads-1) {
				while(gate->sync--) LSEMA_SIGNAL(gate->sync_sem);
				gate->sync=0;
			} else {
				gate->sync++;
				do_wait=1;
			}
		END_BLOCK_CLEAR_SIGNAL
	LOCK_R_END
	if(do_wait) {
#ifdef DEADLOCK_DETECTION
		LM_LOCK(deadlock);
		no_of_threads_running--;
		LM_UNLOCK(deadlock);
#endif
		LSEMA_WAIT(gate->sync_sem);
#ifdef DEADLOCK_DETECTION
		LM_LOCK(deadlock);
		no_of_threads_running++;
		LM_UNLOCK(deadlock);
#endif
	}
	PRSTAT("GATE","SYNC DONE",gate->l.number);
	print_trace(FLP TR_GATE_SYNC,TR_STOP,(long)gate);
}

int p_gate_size(FLA GATE_T gate) { return gate->enqueued; }
int p_gate_has_threads(FLA GATE_T gate) { return gate->threads; }
GATE_T p_cohort(FLAO) { return (GATE_T)GET_COHORT; }

void p_gate_clear(FLA GATE_T gate)
{
	volatile int kill_me=0;
	CHKGATEVOID(gate);
	PRSTAT("GATE","CLEAR",gate->l.number);
	print_trace(FLP TR_GATE_CLEAR,TR_START,(long)gate);

	GATELOCK(gate,L_MUTEX)
		START_BLOCK_CLEAR_SIGNAL
			gate->enqueued=0;
			while(gate->bot!=NULL) {
				GATE_ELEM_T n;
				n=gate->bot;
				gate->bot=n->next;
			}
			gate->top=NULL;
			
			while(tlist_empty(gate->attached)) {
				THREAD_LIST_ELEMENT_TYPE id;
				id=tlist_dequeue(gate->attached);
				if(id==GET_THREAD_LIST_ELEMENT) kill_me=1;
				else CLEAR_THREAD(id);
			}
		END_BLOCK_CLEAR_SIGNAL
	LOCK_R_END

	print_trace(FLP TR_GATE_CLEAR,TR_STOP,(long)gate);
	PRSTAT("GATE","CLEAR DONE",gate->l.number);
	if(kill_me) CLEAR_THREAD(GET_THREAD_LIST_ELEMENT);
}



#include <stdio.h>
#include <stdarg.h>
#include <assert.h>
#include <string.h>
#include <setjmp.h>

double l_k=0.001;

#ifdef DEADLOCK_DETECTION
LM_TYPE deadlock;
long no_of_threads_running=1;
long no_of_plates_enqueued=0;
long no_of_plates_failed=0;
long check_id=1;
#  ifdef NO_TIMER_IF_TOP_PLATE
long no_of_plates_without_timer=0;
#  endif

#endif

enum T_OR_L { TL_LOCK, TL_TRY };

/* static LM_TYPE mtx; not needed anymore */

void p_init_locks()
{
	p_mutex_init();
	p_gate_init();
#ifdef DEADLOCK_DETECTION
	deadlock=LM_CREATE;
#endif
#ifdef PSATHER_RT_TRACE
	{
		char p[1000];
		extern char *bin_prog_name;
		strcpy(p,bin_prog_name);
		strcat(p,".stat");
		rt_file=fopen(p,"w");
		if(rt_file==NULL) {
			fprintf(stderr,"cannot open trace/statistic file `%s'\n",p);
			abort();
		}
		trace_lock=LM_CREATE;
	}
#endif
}

#ifdef PSATHER_RT_TRACE
FILE *rt_file;
LM_TYPE trace_lock;
unsigned long trace_stat[TR_LAST];
static void  trace_type(int type)
{
	switch(type) {
	case TR_MUTEX_CREATED        : fprintf(rt_file,"MUTEX_CREATED");break;
	case TR_GATE_CREATED         : fprintf(rt_file,"GATE_CREATED");break;
	case TR_GATE_ENQUEUE         : fprintf(rt_file,"GATE_ENQUEUE");break;
	case TR_GATE_DEQUEUE         : fprintf(rt_file,"GATE_DEQUEUE");break;
	case TR_GATE_SET             : fprintf(rt_file,"GATE_SET");break;
	case TR_GATE_GET             : fprintf(rt_file,"GATE_GET");break;
	case TR_GATE_SYNC            : fprintf(rt_file,"GATE_SYNC");break;
	case TR_GATE_CLEAR           : fprintf(rt_file,"GATE_CLEAR");break;
	case TR_GATE_ATTACH          : fprintf(rt_file,"GATE_ATTACH");break;
	case TR_GATE_DETACH          : fprintf(rt_file,"GATE_DETACH");break;
	case TR_THREAD               : fprintf(rt_file,"THREAD");break;
	case TR_LOCK                 : fprintf(rt_file,"LOCK");break;
	case TR_LOCK_RESERVE         : fprintf(rt_file,"LOCK_RESERVE");break;
	case TR_TRY                  : fprintf(rt_file,"TRY");break;
	case TR_TRY_FAILED           : fprintf(rt_file,"TRY_FAILED");break;
	case TR_ENDLOCK              : fprintf(rt_file,"ENDLOCK");break;
	case TR_UNLOCK               : fprintf(rt_file,"UNLOCK");break;
	case TR_ATTACH               : fprintf(rt_file,"ATTACH");break;
	case TR_PLATE_SPEAR          : fprintf(rt_file,"PLATE_SPEAR");break;
	case TR_PLATE_MOVE_TOP       : fprintf(rt_file,"PLATE_MOVE_TOP");break;
	case TR_PLATE_REMOVE         : fprintf(rt_file,"PLATE_REMOVE");break;
	case TR_STICK_RESERVE        : fprintf(rt_file,"STICK_RESERVE");break;
	case TR_STICK_GIVEUP         : fprintf(rt_file,"STICK_GIVEUP");break;
	case TR_TIMER_START          : fprintf(rt_file,"TIMER_START");break;
	case TR_TIMER_STOP           : fprintf(rt_file,"TIMER_STOP");break;
	case TR_TIMER_RING           : fprintf(rt_file,"TIMER_RING");break;
	case TR_DEADLOCK             : fprintf(rt_file,"DEADLOCK");break;
	case TR_ACQUIRED             : fprintf(rt_file,"ACQUIRED");break;
	case TR_RELEASED             : fprintf(rt_file,"RELEASED");break;
	case TR_CHECK                : fprintf(rt_file,"CHECK");break;
	case TR_WAKEUP               : fprintf(rt_file,"WAKEUP");break;

	default: fprintf(rt_file,"???(%d)",type);
	}
}

#define MAX_LOCKS_FOR_STAT 50 /* max 50 locks per lock, try, reserve */
long n_locks[MAX_LOCKS_FOR_STAT];
long n_try[MAX_LOCKS_FOR_STAT];
long n_reserve[MAX_LOCKS_FOR_STAT];
long n_move[MAX_LOCKS_FOR_STAT];

unsigned long start_sec=0;
extern int gettimeofday(struct timeval *,struct timezone *);
void print_statistics()
{
	int i,tot;
	struct timeval tv;
	LM_LOCK(trace_lock);
	gettimeofday(&tv,NULL);
	fprintf(rt_file,"%ld.%06ld ",tv.tv_sec-start_sec,tv.tv_usec);
	fprintf(rt_file,"END\n----------------------------------------------------------------------\n");
	fprintf(rt_file,"Statistics:\n");
	for(i=0;i<TR_LAST;i++) {
		if(trace_stat[i]) {
			trace_type(i);
			fprintf(rt_file,"\t%ld\n",trace_stat[i]);
		}
	}
	fprintf(rt_file,"\n");
	for(i=tot=0;i<MAX_LOCKS_FOR_STAT;i++) 
		if(n_locks[i]) fprintf(rt_file,"LOCK%d\t%ld\n",i,n_locks[i]);
	for(i=tot=0;i<MAX_LOCKS_FOR_STAT;i++) 
		if(n_reserve[i]) fprintf(rt_file,"LOCK_RESERVE%d\t%ld\n",i,n_reserve[i]);
	for(i=tot=0;i<MAX_LOCKS_FOR_STAT;i++) 
		if(n_try[i]) fprintf(rt_file,"TRY%d\t%ld\n",i,n_try[i]);
	for(i=tot=0;i<MAX_LOCKS_FOR_STAT;i++) 
		if(n_move[i]) fprintf(rt_file,"MOVE%d\t%ld\n",i,n_move[i]);
#ifdef LWP
	PRINT_IDLE(rt_file);
#endif
	fflush(rt_file);
	LM_UNLOCK(trace_lock);
}
int print_begin_trace_line(FLA int type,char startstop) 
{
#ifndef PSATHER_RT_STATISTICS_ONLY
	static unsigned long seq=0;
#endif
	struct timeval tv;
	LM_LOCK(trace_lock);
	if(start_sec==0) {
		while(1) {
			gettimeofday(&tv,NULL);
			if(tv.tv_usec<1000) break;
		}
		start_sec=tv.tv_sec;
	} else gettimeofday(&tv,NULL);

	if(startstop!=TR_STOP) trace_stat[type]++;
#ifndef PSATHER_RT_STATISTICS_ONLY
#ifdef DEBUG
	fprintf(rt_file,"%s:%ld:",file,line);
#endif
	fprintf(rt_file,"%ld ",seq++);
	fprintf(rt_file,"%ld.%06ld ",tv.tv_sec-start_sec,tv.tv_usec);
	fprintf(rt_file,"%ld ",(long)THREAD_GET_ID);
	fprintf(rt_file,"%c ",startstop);
	trace_type(type);
	return 1;
#else
	return 0;
#endif
}
void print_trace(FLA int type,char startstop,unsigned long id) 
{
	if(!print_begin_trace_line(FLP type,startstop)) { LM_UNLOCK(trace_lock);return; }
	fprintf(rt_file," %ld\n",id);
	fflush(rt_file);
	LM_UNLOCK(trace_lock);
}
static void prlocks(FLA int type,char st,struct EXCEPT_ELEMENT *nn)
{
	int i;
	if(st!=TR_STOP) {
		switch(type) {
		case TR_LOCK: n_locks[nn->nlocks]++;break;
		case TR_TRY: n_try[nn->nlocks]++;break;
		case TR_LOCK_RESERVE: n_reserve[nn->nlocks]++;break;
		case TR_PLATE_MOVE_TOP: n_move[nn->nlocks]++;break;
		}
	}
	if(!print_begin_trace_line(FLP type,st)) { LM_UNLOCK(trace_lock);return; }
	fprintf(rt_file," %d ",nn->nlocks);
	for(i=0;i<nn->nlocks;i++) {
		fprintf(rt_file,"%ld ",(long)nn->locks[i].lck);
		if(nn->locks[i].type&L_MUTEX) fprintf(rt_file,"L");
		if(nn->locks[i].type&L_THREADS) fprintf(rt_file,"=");
		if(nn->locks[i].type&L_NO_THREADS) fprintf(rt_file,"/");
		if(nn->locks[i].type&L_EMPTY) fprintf(rt_file,"O");
		if(nn->locks[i].type&L_NOT_EMPTY) fprintf(rt_file,"@");
		if(i!=nn->nlocks-1) fprintf(rt_file," ");
	}

	fprintf(rt_file,"\n");
	fflush(rt_file);
	LM_UNLOCK(trace_lock);
}
static void prlocksp(FLA int type,char st,LOCK_PLATE nn)
{
	int i;
	if(st!=TR_STOP) {
		switch(type) {
		case TR_LOCK: n_locks[nn->size]++;break;
		case TR_TRY: n_try[nn->size]++;break;
		case TR_LOCK_RESERVE: n_reserve[nn->size]++;break;
		case TR_PLATE_MOVE_TOP: n_move[nn->size]++;break;
		}
	}
	if(!print_begin_trace_line(FLP type,st)) { LM_UNLOCK(trace_lock);return;}
	fprintf(rt_file," %ld ",nn->size);
	for(i=0;i<nn->size;i++) {
		fprintf(rt_file,"%ld ",(long)nn->hole[i].d.lck);
		if(nn->hole[i].d.type&L_MUTEX) fprintf(rt_file,"L");
		if(nn->hole[i].d.type&L_THREADS) fprintf(rt_file,"=");
		if(nn->hole[i].d.type&L_NO_THREADS) fprintf(rt_file,"/");
		if(nn->hole[i].d.type&L_EMPTY) fprintf(rt_file,"O");
		if(nn->hole[i].d.type&L_NOT_EMPTY) fprintf(rt_file,"@");
		if(i!=nn->size-1) fprintf(rt_file," ");
	}

	fprintf(rt_file,"\n");
	fflush(rt_file);
	LM_UNLOCK(trace_lock);
}

#endif


void p_init_lock(LOCK_T lck)
{
#ifdef DEBUG_PSATHER
	static long lock_number=1;
	lck->l.number=lock_number++;
#endif
	lck->l.bottom=lck->l.top=NULL;
	lck->l.unl.single.id=NO_THREAD;
	lck->l.unl.single.unlocked=ls_create();
	lck->l.hold=LM_CREATE;
	lck->l.mark=1;
}

/* 
 * returns 1 if all conditions are met to lock this lock 
 * (includes NO check to see if it is free, though!!!)
 */
static int p_check_conditions_only(FLA DLOCK *d)
{
	return (*d->lck->l.fnct->condition_ok)(d->lck,d->type);
}

/*
 * returns 1 if the lock is free AND the conditions
 * are met
 */
int p_check_lock_and_conditions(FLA DLOCK *d)
{
	return (*d->lck->l.fnct->condition_ok)(d->lck,d->type) &&
	       (*d->lck->l.fnct->lock_free)(d->lck,d->type);
}


/* 
 * returns 1, if the lock is open, 0 otherwise
 */
static int p_check_lock(FLA DLOCK *d)
{
       return (*d->lck->l.fnct->lock_free)(d->lck,d->type);
}

/*
 * returns 0 if a try by thread id would fail on this lock
 */
static int p_check_one_for(FLA DLOCK *d,THREAD_ID id)
{
	assert(d->count);
	if(!p_check_conditions_only(FLP d)) return 0;
	if(d->lck->l.unl.single.id==id) return 1;
	else return (*d->lck->l.fnct->lock_free)(d->lck,d->type);
}

/*
 * return 0 if a try would fail, 1 otherwise
 * The answer is ONLY correct for the current thread, it cannot be used
 * to predict results for other threads.
 */
static int p_check_one(FLA DLOCK *d)
{
	return p_check_one_for(FLP d,THREAD_GET_ID);
}

/*
 * returns 1 if the current thread already owns that lock
 */
static int p_is_my_lock(FLA DLOCK *d)
{
	return d->lck->l.unl.single.id==THREAD_GET_ID;
}


/* 
 * returns 0 if the try failed, and 1 otherwise
 */

static int p_try_one(FLA DLOCK *d)
{
	assert(d->count);
	if(!p_check_conditions_only(FLP d)) return 0;
	if(d->lck->l.unl.single.id==THREAD_GET_ID) {
		ls_push(d->lck->l.unl.single.unlocked,d->count);
		return 1;
	} else {
		if((*d->lck->l.fnct->try)(d->lck,d->type)) {
			d->lck->l.unl.single.id=THREAD_GET_ID;
			ls_push(d->lck->l.unl.single.unlocked,d->count);
			print_trace(FLP TR_ACQUIRED,TR_DONE,(long)d->lck);
			return 1;
		} else return 0;
	}
}

static int p_unlock_one(FLA DLOCK *d)
{
	int r;
#ifndef NO_ERROR_CHECKS
	if(d->lck->l.unl.single.id!=THREAD_GET_ID 
	   || ls_is_empty(d->lck->l.unl.single.unlocked)) {
		RFATALp("a thread tries to unlock a lock it did not acquire\n");
	}
#endif
	r=ls_pop(d->lck->l.unl.single.unlocked);
#ifndef NO_ERROR_CHECKS
	if(r==0) {
		RFATALp("a thread tries to unlock a lock it did already unlock\n");
		exit(-1);
	} else 
#endif
	if(r>1) {
		ls_push(d->lck->l.unl.single.unlocked,r-1);
		r=0;
	} else if(ls_is_empty(d->lck->l.unl.single.unlocked)) {
		d->lck->l.unl.single.id=NO_THREAD;
		print_trace(FLP TR_RELEASED,TR_DONE,(long)d->lck);
		(*d->lck->l.fnct->unlock)(d->lck,d->type);
		r=1;
	} else {
		ls_push(d->lck->l.unl.single.unlocked,0);
		r=0;
	}
	return r;
}

/*
 * this function assumes that the lock is holded by the current thread
 */
static int p_endlock_one(FLA DLOCK *d)
{
	int i;
	 /* the following condition is true IF the thread lost
	    the lock due to an unlock */
	if(d->lck->l.unl.single.id!=THREAD_GET_ID) return(0);
	i=ls_pop(d->lck->l.unl.single.unlocked);
	if(ls_is_empty(d->lck->l.unl.single.unlocked)) {
		d->lck->l.unl.single.id=NO_THREAD;
		print_trace(FLP TR_RELEASED,TR_DONE,(long)d->lck);
		(*d->lck->l.fnct->unlock)(d->lck,d->type);
		return 1;
	} 
	return 0;
}

void WakeupPlate(LOCK_PLATE p,enum LOCK_PLATE_FLAGS why)
{
	p->flags|=why;

#ifdef PSATHER_RT_TRACE
	if(why!=LOCK_TIMER_SIGNAL) {
#ifdef DEBUG
		if(print_begin_trace_line("???",0, TR_WAKEUP,TR_DONE)) {
#else
		if(print_begin_trace_line(TR_WAKEUP,TR_DONE)) {
#endif
			fprintf(rt_file," %ld ",(long)p->id);
			switch(why) {
			case LOCK_TIMER_SIGNAL: fprintf(rt_file,"TIMER");break;
			case LOCK_PLATE_ABOVE_MOVED:  fprintf(rt_file,"PLATE_ABOVE_MOVED");break;
			case LOCK_PLATE_BELOW_MOVED:  fprintf(rt_file,"PLATE_BELOW_MOVED");break;
			case LOCK_CHANGE: 	fprintf(rt_file,"CHANGE");break;
			case LOCK_JUMP_TO_TOP: 	fprintf(rt_file,"JUMP_TO_TOP");break;
			case LOCK_START_TIMER: 	fprintf(rt_file,"START_TIMER");break;
			default:		fprintf(rt_file,"???");break;
			}
			fprintf(rt_file,"\n");
		}
		LM_UNLOCK(trace_lock);
	}
#endif
	if(!(p->flags&LOCK_PLATE_GOT_SIGNAL)) {
		p->flags|=LOCK_PLATE_GOT_SIGNAL;
		LSEMA_SIGNAL(p->restart_thread); 
	}
}

#ifdef UNUSED
/*
 * Checks if all plates speared upon an open lock do not have the correct
 * conditions. If at least one of the plat does have the correct condition,
 * the bottom plate will get a LOCK_JUMP_TO_TOP signal.
 * It returns 1 (WCJ_DONE) if such a signal has been sent, 2 (WCJ_NO_PLATE)
 * if no plate * fullfills the conditions, 3 (WCJ_BOTTOM_OK) if the bottom 
 * plate fullfills the conditions, and 0 (WCJ_LOCKED) if the lock is locked otherwise.
 */
#define WCJ_LOCKED    0
#define WCJ_DONE      1
#define WCJ_NO_PLATE  2
#define WCJ_BOTTOM_OK 3

static int MakePlateWithWrongConditionsJump(DLOCK *lck)
{
	LOCK_HOLE *h;
	if(p_check_lock(FL lck)) {
		h=lck->lck->l.bottom;
		while(h!=NULL) {
			if(p_check_one_for(&h->d,h->myPlate->id)) break;
			h=h->above;
		}
		if(h==NULL) return WCJ_NO_PLATE;
		h=h->below;
		if(h==NULL) return WCJ_BOTTOM_OK;
		while(h->below!=NULL) h=h->below;
		WakeupPlate(h->myPlate,LOCK_JUMP_TO_TOP);
		return WCJ_DONE;
	}
	return WCJ_LOCKED;
}
#endif

#ifdef DEBUG_PSATHER
#define MAX_PLATES 400  /* no more than 100 locks at the same time during debugging */
#define MAX_DLOCKS 400  /* no more than 100 different locks */
static LOCK_PLATE plates[MAX_PLATES];
static int nplate=0;
static int locknr[MAX_PLATES];
LM_TYPE plate_lck=NULL;

void PrintPlates()
{
	LOCK_T cur;
	int did_print;
	LOCK_T lcks[MAX_DLOCKS];
	int last_lock;
	int n;

	if(plate_lck==NULL) plate_lck=LM_CREATE;

	LM_LOCK(plate_lck);
	fprintf(stderr,"PLATE:");
	for(n=0;n<MAX_PLATES;n++) {
		if(plates[n]!=NULL) fprintf(stderr," %-2d",n);
		locknr[n]=0;
	}
	fprintf(stderr,"\n");

	fprintf(stderr,"TH_ID:");
	for(n=0;n<MAX_PLATES;n++) {
		if(plates[n]!=NULL) fprintf(stderr," %-2d",plates[n]->id);
		locknr[n]=0;
	}
	fprintf(stderr,"\n");

	did_print=1;
	last_lock=0;
	while(did_print) {
		cur=NULL;
		for(n=0;n<MAX_PLATES;n++)
			if(plates[n]!=NULL) 
				if(locknr[n]<plates[n]->size) 
					if(cur==NULL || cur>plates[n]->hole[locknr[n]].lck) 
						cur=plates[n]->hole[locknr[n]].lck;

		if(cur==NULL) break;
		assert(cur->l.top!=NULL && cur->l.bottom!=NULL);
		lcks[last_lock++]=cur;
		fprintf(stderr,"  %3ld:",cur->l.number);
		
		for(n=0;n<MAX_PLATES;n++) {
			if(plates[n]!=NULL) {
				if(plates[n]->hole[locknr[n]].lck==cur) {
					fprintf(stderr,"-*-");
					locknr[n]++;
				} else {
					fprintf(stderr,"---");
				}
			}
		}
		fprintf(stderr,"\n");
	}
	fprintf(stderr,"\n");
	for(n=0;n<last_lock;n++) {
		LOCK_HOLE *l;
		fprintf(stderr,"  -");
		l=lcks[n]->l.bottom;
		assert(l!=NULL);
		while(l) {
			if(l->below!=NULL) assert(l->below->above==l);
			else assert(lcks[n]->l.bottom==l);
			if(l->above!=NULL) assert(l->above->below==l);
			else assert(lcks[n]->l.top==l);
			fprintf(stderr,"%d-",l->myPlate->index);
			l=l->above;
		}
		fprintf(stderr,"   ");
		(*lcks[n]->l.fnct->print)(lcks[n]);
	
	}
	
	fprintf(stderr,"\n");

	LM_UNLOCK(plate_lck);
}

#endif

static void RemovePlate(FLA LOCK_PLATE n)
{
	int i;
#ifdef DEBUG_PSATHER
	if(plate_lck==NULL) plate_lck=LM_CREATE;
	LM_LOCK(plate_lck);
	fprintf(stderr,"Deleting plate %d now\n",n->index);
#endif
#ifdef PSATHER_RT_TRACE
	prlocksp(FLP TR_PLATE_REMOVE,TR_DONE,n);
#endif
	STOP_TIMER(n);
	LSEMA_DELETE(n->restart_thread);
	for(i=0;i<n->size;i++) {
		if(n->hole[i].d.lck->l.bottom==&(n->hole[i])) 
			n->hole[i].d.lck->l.bottom=n->hole[i].above;
		if(n->hole[i].d.lck->l.top==&(n->hole[i])) 
			n->hole[i].d.lck->l.top=n->hole[i].below;
		if(n->hole[i].above!=NULL) 
			n->hole[i].above->below=n->hole[i].below;
		if(n->hole[i].below!=NULL) 
			n->hole[i].below->above=n->hole[i].above;
	}
#ifdef DEADLOCK_DETECTION
	LM_LOCK(deadlock);
	no_of_plates_enqueued--;
	no_of_plates_failed=0;
	check_id++;
	LM_UNLOCK(deadlock);
#endif

#ifdef DEBUG_PSATHER
	plates[n->index]=NULL;
	fprintf(stderr,"Deleting plate %d done\n",n->index);
	LM_UNLOCK(plate_lck);
	PrintPlates();
#endif
	free(n);
}

#ifdef PSATHER_RT_TRACE
#define RESERVE_STICKS	do { int i;\
	for(i=0;i<nn->nlocks;i++) { \
		print_trace(FLP TR_STICK_RESERVE,TR_START,(long)nn->locks[i].lck);\
		LM_LOCK(nn->locks[i].lck->l.hold);\
		print_trace(FLP TR_STICK_RESERVE,TR_STOP,(long)nn->locks[i].lck);\
	} } while(0)
#define GIVEUP_STICKS	do { int i;\
	for(i=0;i<nn->nlocks;i++) {\
		print_trace(FLP TR_STICK_GIVEUP,TR_START,(long)nn->locks[i].lck); \
		LM_UNLOCK(nn->locks[i].lck->l.hold); 			      \
		print_trace(FLP TR_STICK_GIVEUP,TR_STOP,(long)nn->locks[i].lck);  \
	} } while(0)
#else
#define RESERVE_STICKS	do { int i;\
	for(i=0;i<nn->nlocks;i++) \
		LM_LOCK(nn->locks[i].lck->l.hold); \
	} while(0)
#define GIVEUP_STICKS	do { int i;\
	for(i=0;i<nn->nlocks;i++) \
		LM_UNLOCK(nn->locks[i].lck->l.hold); \
	} while(0)
#endif

static int some_closed(LOCK_PLATE n)
{
	int i;
	for(i=0;i<n->size;i++) 
		if(!p_check_one(FL &n->hole[i].d)) return 1;
	return 0;
}

/*
 * This function checks if all locks are visible and lockable by 
 * the current thread. (It does no longer lock them as 
 * previous versions did).
 */
static int visible_plus_ok(FLA LOCK_PLATE n) 
{
	int i;
	print_trace(FLP TR_CHECK,TR_DONE,0);
	for(i=0;i<n->size;i++) {
#ifdef IGNORE_WRONG_CONDITION_PLATES
		LOCK_HOLE *h;
		if(!p_check_one(FL &n->hole[i].d)) return 0;
		h= n->hole[i].below;
		while(h!=NULL) {
			if(p_check_one_for(FL &h->d,h->myPlate->id)) return 0;
			h=h->below;
		}
#else
		if(n->hole[i].below==NULL) {
			if(!p_check_one(FL &n->hole[i].d)) return 0;
		} else return 0;
#endif
	}
	return 1;
}

static void remove_all_marks(LOCK_PLATE n)
{
	int i;
	for(i=0;i<n->size;i++)
		n->hole[i].mark=0;
}

static void note_marks(LOCK_PLATE n)
{
	int i;
	for(i=0;i<n->size;i++) 
		n->hole[i].mark=n->hole[i].d.lck->l.mark;
}

static void note_time(LOCK_PLATE n)
{
	gettimeofday(&n->timer_started_at,NULL);
}


static void adjust_marks(LOCK_T lck)
{
	LOCK_HOLE *h,*ch=NULL;
	h=lck->l.bottom;
	lck->l.mark++;

	while(h!=NULL) {
		if(p_check_one_for(FL &h->d,h->myPlate->id)) {
			if(ch==NULL) ch=h;
		} else h->mark=lck->l.mark;
		h=h->above;
	}
	if(ch!=NULL) {
		if(ch->below==NULL) WakeupPlate(ch->myPlate,LOCK_CHANGE);
		else WakeupPlate(lck->l.bottom->myPlate,LOCK_JUMP_TO_TOP);
	}
}

/*
 * Lock all locks. It does not check if the current thread may lock
 * theses locks. If a lock failes, the program aborts
 */
static void lock_up(FLA struct EXCEPT_ELEMENT *nn)
{
	int i,bottom;
	LOCK_HOLE *h;
	for(i=0;i<nn->nlocks;i++) {
		assert(p_try_one(FLP &nn->locks[i]));
		h=nn->locks[i].lck->l.bottom;
		nn->locks[i].lck->l.mark++;
		bottom=1;
		if(h==NULL) 
			nn->locks[i].lck->l.mark=1;
		else
			nn->locks[i].lck->l.mark++;
		while(h!=NULL) {
			h->mark= nn->locks[i].lck->l.mark;
			if(h->myPlate->flags&LOCK_TRY) 
				WakeupPlate(h->myPlate,LOCK_CHANGE);
			else if(bottom && p_check_one_for(FLP &h->d,h->myPlate->id)) {
				bottom=0;
				WakeupPlate(h->myPlate,LOCK_START_TIMER);
			}
			h=h->above;
		}
	}
}

#ifdef UNUSED
static int on_bottom_of_some_sticks(LOCK_PLATE n)
{
	int i;
	for(i=0;i<n->size;i++) if(n->hole[i].below==NULL) return 1;
	return 0;
}
#endif

#ifdef UNUSED
/*
 * This function returns 1, if the plate is at the bottom of all sticks
 */
static int on_bottom_of_all_sticks(LOCK_PLATE n)
{
	int i;
	for(i=0;i<n->size;i++)
		if(n->hole[i].below!=NULL) return 0;
	return 1;
}
#endif

#ifdef UNUSED
static int some_plates_are_above(LOCK_PLATE n)
{
	int i;
	for(i=0;i<n->size;i++)
		if(n->hole[i].above!=NULL) return 1;
	return 0;
}
#endif

#ifdef UNUSED
#ifdef DEADLOCK_DETECTION
static void CheckDeadLock(FLAO)
{
	   if(no_of_plates_failed
#  ifdef NO_TIMER_IF_TOP_PLATE
		   +no_of_plates_without_timer
#  endif
	      ==no_of_threads_running) {
		   fprintf(stderr,"DEADLOCK SITUATION, ABORTING\n");
		   print_trace(FLP TR_DEADLOCK,TR_START,0);
		   print_statistics();
		   exit(-1);
	   }
}
#endif
#endif

static void MovePlateToTopPos(LOCK_PLATE n)
{
	int i,moved=0;
	int sticks_at_bottom=0;
	LOCK_HOLE *h;

	for(i=0;i<n->size;i++) {
		h= &n->hole[i];
		if(h->below==NULL) sticks_at_bottom++;
		if(h->above!=NULL) {
			moved++;
			if(h->below!=NULL) h->below->above=h->above;
			else h->d.lck->l.bottom=h->above;
			h->above->below=h->below;
			h->below=h->d.lck->l.top;
			h->below->above=h;
			h->above=NULL;
			h->d.lck->l.top=h;
		}
	}

#ifdef PSATHER_RT_TRACE
	if(moved) prlocksp(FLD TR_PLATE_MOVE_TOP,TR_DONE,n);
#endif
}

#ifdef UNUSED
static void NotifyPlateAbove(LOCK_PLATE n)
{
	int i;
	LOCK_HOLE *h;
	for(i=0;i<n->size;i++) {
		h= &n->hole[i];
		if(h->above!=NULL && h->below==NULL) {
#ifdef MAKE_PLATE_WITH_WRONG_CONDITION_JUMP
			/* temporarly remove bottom plate */
			h->d.lck->l.bottom=h->above;
			h->above->below=NULL;
			switch(MakePlateWithWrongConditionsJump(&h->d)) {
			case WCJ_LOCKED   : 
			case WCJ_BOTTOM_OK: WakeupPlate(h->above->myPlate,LOCK_PLATE_BELOW_MOVED);
			case WCJ_DONE     : break;
			case WCJ_NO_PLATE : break;
			}
			h->above->below=h;
			h->d.lck->l.bottom=h;
#else
			WakeupPlate(h->above->myPlate,LOCK_PLATE_BELOW_MOVED);
#endif
		}
	}
}
#endif

#ifdef UNUSED
static void NotifyPlateBelow(LOCK_PLATE n,char *stick_moved)
{
#ifdef NO_TIMER_IF_TOP_PLATE
	int i;
	LOCK_HOLE *h;
	for(i=0;i<n->size;i++) {
		h= &n->hole[i];
		if(h->below!=NULL && (stick_moved==NULL?1:stick_moved[i])) {
			if(!(h->below->myPlate->flags&(LOCK_TIMER_RUNNING))) {
#ifdef NO_WAKEUP_FOR_1_PLATES_IF_NOT_AT_BOTTOM
				if(h->below->myPlate->size==1?h->below->below==NULL:1) {
#endif
					WakeupPlate(h->below->myPlate,LOCK_PLATE_ABOVE_MOVED);
#ifdef NO_WAKEUP_FOR_1_PLATES_IF_NOT_AT_BOTTOM
				}
#endif
			}
		}
	}
#endif
}
#endif

static void NotifyBottomPlates(LOCK_PLATE plate)
{
	LOCK_HOLE *h;
	int i;
	for(i=0;i<plate->size;i++) {
		h=&plate->hole[i];
		if(p_check_lock(FL &h->d)) {
			h=h->d.lck->l.bottom;
			while(h->above!=NULL) { /* do not check top plate */
				if(p_check_one_for(FL &h->d,h->myPlate->id) && h->mark!=h->d.lck->l.mark) {
					if(h->below==NULL) WakeupPlate(h->myPlate,LOCK_START_TIMER);
					else WakeupPlate(h->d.lck->l.bottom->myPlate,LOCK_JUMP_TO_TOP);
					break;
				} else h->mark=h->d.lck->l.mark;
				h=h->above;
			}
		}
	}
}

static int all_locks_could_be_acquired(LOCK_PLATE n)
{
	int i;
	for(i=0;i<n->size;i++)
		if(!p_check_lock_and_conditions(FL &n->hole[i].d)) return 0;
	return 1;
}

static void jump_to_top(LOCK_PLATE n)
{
	if(!all_locks_could_be_acquired(n)) note_marks(n);
	note_time(n);
	MovePlateToTopPos(n);
	NotifyBottomPlates(n);
}

static void PlacePlate(LOCK_PLATE plate)
{
	int i;
	LOCK_HOLE *h;

#ifdef DEBUG_PSATHER
	if(plate_lck==NULL) plate_lck=LM_CREATE;
	LM_LOCK(plate_lck);
	fprintf(stderr,"Adding plate %d now\n",plate->index);
#endif
	for(i=0;i<plate->size;i++) {
		h=&(plate->hole[i]);
		if(h->d.lck->l.top==NULL) {
			h->d.lck->l.top=h->d.lck->l.bottom=h;
			h->above=NULL;
			h->below=NULL;
		} else {
			h->below=h->d.lck->l.top;
			h->d.lck->l.top=h;
			h->above=NULL;
			h->below->above=h;
		}
	}
#ifdef DEBUG_PSATHER
	plates[plate->index]=plate;
	fprintf(stderr,"Adding plate %d done\n",plate->index);
	LM_UNLOCK(plate_lck);
	PrintPlates();
#endif
	/* LM_UNLOCK(mtx); not needed, see above */
}

static LOCK_PLATE MakeNSpearPlate(enum T_OR_L tl,struct EXCEPT_ELEMENT *nn)
{
	LOCK_PLATE n;
	int i,j;
#ifdef DEADLOCK_DETECTION
	LM_LOCK(deadlock);
	no_of_plates_enqueued++;
	LM_UNLOCK(deadlock);
#endif
#ifdef PSATHER_RT_TRACE
	prlocks(FLD TR_PLATE_SPEAR,TR_DONE,nn);
#endif
	n=(LOCK_PLATE)malloc(sizeof(struct LOCK_PLATE_struct)+(nn->nlocks-1)*sizeof(LOCK_HOLE));
	n->restart_thread=LSEMA_CREATE(0);
	n->size=nn->nlocks;
	for(i=0;i<nn->nlocks;i++) if(nn->locks[i].type&L_IGNORE) n->size--;
	n->flags=0;
	gettimeofday(&n->started_at,NULL);
#ifdef LWP
	n->ltimer=NULL;
	n->ltimer_lck=NULL;
#endif
	if(tl==TL_TRY) n->flags|=LOCK_TRY;
#ifdef DEADLOCK_DETECTION
	n->check_id=0;
#endif
	for(i=j=0;j<n->size;i++,j++) {
		while(nn->locks[i].type&L_IGNORE) i++;
		n->hole[j].myPlate=n;
		n->hole[j].d=nn->locks[i];
		n->hole[j].mark=0;
	}
#ifdef DEBUG_PSATHER
	while(plates[nplate]!=NULL) {
		if(nplate==MAX_PLATES-1) nplate=0;
		else nplate++;
	}
	n->index=nplate++;
#endif
	n->id=THREAD_GET_ID;
	PlacePlate(n);
	return n;
}

static void CombineLocks(struct EXCEPT_ELEMENT *nn)
{
	int i,j;
	
	if(nn->nlocks==1) return;
	for(j=1,i=0;j<nn->nlocks;j++) {
		if(nn->locks[i].lck==nn->locks[j].lck) {
			nn->locks[i].type|=nn->locks[j].type;
			nn->locks[i].count+=nn->locks[j].count;
		} else {
			i++;
			if(j!=i) nn->locks[i]=nn->locks[j];
		}
	}
	nn->nlocks=i+1;
}

static int CheckForDoubleLocksWhichFail(FLA enum T_OR_L tl,struct EXCEPT_ELEMENT *nn)
{
	int i;
	for(i=0;i<nn->nlocks;i++) {
		if(p_is_my_lock(FL &nn->locks[i])) {
		   	if(!p_check_conditions_only(FL &nn->locks[i])) {
				if(tl==TL_TRY) {
					return 1;
				} else {
					RFATALp("A thread has a lock and tried\nto acquire it again under some conditions.\nExample: g.no_threads, g.dequeue, ....\nThis condition is currently wrong, and as the thread already owns\nthat lock, the condition will never change.\n---> DEADLOCK\n");
				}
			} else nn->locks[i].type|=L_IGNORE;
		}
	}
	return 0;
}

/*
 * Returns one if the plate has on one stick with an open lock
 * conditions which fail, and all plates below on this sticks also
 * have conditions which fail.
 * Otherwise it returns 0
 */
static int some_conditions_wrong_on_free_lock(LOCK_PLATE n)
{
	int i;
	LOCK_HOLE *h;
	for(i=0;i<n->size;i++) {
		h=&(n->hole[i]);
		if(p_check_lock(FL &h->d) && !p_check_conditions_only(FL &h->d)) {
			h=h->below;
			while(h!=NULL) {
				if(p_check_one_for(FL &h->d,h->myPlate->id))
					break;
				h=h->below;
			} 
			if(h==NULL) return 1;
		}
	}
	return 0;
}

static int some_plate_above_needs_lock(LOCK_PLATE n)
{
	int i;
	LOCK_HOLE *h;
	for(i=0;i<n->size;i++) {
		h=&(n->hole[i]);
		if(p_check_lock(FL &h->d)) {
			h=h->above;
			while(h!=NULL) {
				if(p_check_one_for(FL &h->d,h->myPlate->id) &&
				   h->mark!=h->d.lck->l.mark)
					return 1;
				h=h->above;
			}
		}
	}
	return 0;
}

static void make_plates_below_move(LOCK_PLATE n)
{
	int i;
	for(i=0;i<n->size;i++) {
		LOCK_HOLE *h= &(n->hole[i]);
		if(h->d.lck->l.bottom!=h) {
			LOCK_HOLE *b=h->d.lck->l.bottom;
			if(p_check_one_for(FL &b->d,b->myPlate->id) && (b->mark!=b->d.lck->l.mark)) 
				WakeupPlate(b->myPlate,LOCK_START_TIMER);
			else 
				WakeupPlate(b->myPlate,LOCK_JUMP_TO_TOP);
		}
	}
}
static void SendSignalsForNewPlate(LOCK_PLATE n)
{
	if(all_locks_could_be_acquired(n)) {
		make_plates_below_move(n);
	} else {
		note_marks(n);
		note_time(n);
	}
}

static int plate_should_jump_to_top(LOCK_PLATE n)
{
	int i;
	LOCK_HOLE *h;
	for(i=0;i<n->size;i++) {
		h=&n->hole[i];

		if(p_check_lock(FL &h->d) && !p_check_conditions_only(FL &h->d)) {
			h=h->below;
			while(h!=NULL) {
				if(p_check_one_for(FL &h->d,h->myPlate->id))
					return 0;
				h=h->below;
			}
		}
	}
	return 1;
}	

static int p_lock_or_try(FLA enum T_OR_L tl,struct EXCEPT_ELEMENT *nn)
{
	LOCK_PLATE n;
	DEFER;
	START_BLOCK_CLEAR_SIGNAL
		RESERVE_STICKS;
		if(CheckForDoubleLocksWhichFail(FLP tl,nn)) { GIVEUP_STICKS;return 0; }
		n=MakeNSpearPlate(tl,nn);
		if(visible_plus_ok(FLP n)) {
			RemovePlate(FLP n);
			lock_up(FLP nn);
			p_ex_push(FL nn);
			ENABLE_CLEAR_SIGNAL;
			PRSTAT("multi_lock","done",nn->nlocks);
			if(!(nn->flags&LOCK_RESERVE)) GIVEUP_STICKS;
			return 1;
		} else if(tl==TL_TRY && some_closed(n)) {
			RemovePlate(FLP n);
			GIVEUP_STICKS;
			return 0;
		}
		SendSignalsForNewPlate(n);

		while(1) {
			GIVEUP_STICKS;
			LSEMA_WAIT(n->restart_thread);
			RESERVE_STICKS;

			n->flags&= ~LOCK_PLATE_GOT_SIGNAL;

			if(CLEAR_SIGNAL_ARRIVED) {
				RemovePlate(FLP n);
				GIVEUP_STICKS;
				ENABLE_CLEAR_SIGNAL;
				CLEAR_ME;
			}
			if(n->flags&(LOCK_CHANGE|LOCK_START_TIMER)) {
				if(visible_plus_ok(FLP n)) {
					RemovePlate(FLP n);
					lock_up(FLP nn);
					p_ex_push(FL nn);
					ENABLE_CLEAR_SIGNAL;
					PRSTAT("multi_lock","done",nn->nlocks);
					if(!(nn->flags&LOCK_RESERVE)) GIVEUP_STICKS;
					return 1;
				} else if(tl==TL_TRY && some_closed(n)) {
					RemovePlate(FLP n);
					GIVEUP_STICKS;
					return 0;
				}
				if(all_locks_could_be_acquired(n)) {
					note_time(n);
					remove_all_marks(n);
					make_plates_below_move(n);
					n->flags&= ~LOCK_START_TIMER;
				} else
					note_marks(n);
			}
			if(n->flags&LOCK_CHANGE) {
				note_time(n);

				if(some_plate_above_needs_lock(n))
					if(some_conditions_wrong_on_free_lock(n))
						jump_to_top(n);
					else
						if(!START_TIMER(n)) jump_to_top(n);
				n->flags&= ~LOCK_CHANGE;
			}

			if(n->flags&(LOCK_JUMP_TO_TOP|LOCK_TIMER_SIGNAL)) {
				STOP_TIMER(n);
				jump_to_top(n);
				n->flags&= ~(LOCK_JUMP_TO_TOP|LOCK_START_TIMER|LOCK_TIMER_SIGNAL);
			}

			if(n->flags&LOCK_START_TIMER) {
				if(n->flags&LOCK_TIMER_RUNNING) {
					if(plate_should_jump_to_top(n)) {
						STOP_TIMER(n);
						jump_to_top(n);
					}
				} else {
					STOP_TIMER(n);
					if(plate_should_jump_to_top(n)) 
						jump_to_top(n);
					else
						if(!START_TIMER(n)) jump_to_top(n);
				}
				n->flags&= ~LOCK_START_TIMER;
			}
		}
	END_BLOCK_CLEAR_SIGNAL;
}

void p_lock(FLA  struct EXCEPT_ELEMENT *nn)
{
	CombineLocks(nn); /* the runtime does not like locks like lock a,a,
	                     but prefers them to be combined in one lock.
			     CombineLocks solves this problem. */
#ifdef PSATHER_RT_TRACE
	if(nn->flags&LOCK_RESERVE) prlocks(FLP TR_LOCK_RESERVE,TR_START,nn);
	else prlocks(FLP TR_LOCK,TR_START,nn);
#endif
	assert(p_lock_or_try(FLP TL_LOCK,nn));
#ifdef PSATHER_RT_TRACE
	if(nn->flags&LOCK_RESERVE) prlocks(FLP TR_LOCK_RESERVE,TR_STOP,nn);
	else prlocks(FLP TR_LOCK,TR_STOP,nn);
#endif
}

int p_try(FLA struct EXCEPT_ELEMENT *nn)
{
	int i;
	CombineLocks(nn); /* the runtime does not like locks like lock a,a,
	                     but prefers them to be combined in one lock.
			     CombineLocks resolves this problem. */
#ifdef PSATHER_RT_TRACE
	prlocks(FLP TR_TRY,TR_START,nn);
#endif
	i= p_lock_or_try(FLP TL_TRY,nn);
#ifdef PSATHER_RT_TRACE
	if(i) prlocks(FLP TR_TRY,TR_STOP,nn);
	else prlocks(FLP TR_TRY_FAILED,TR_STOP,nn);
#endif
	return i;
}

void p_unlock(FLA DLOCK *lck)
{
	PRSTAT("unlock","start",1);

#ifdef PSATHER_RT_TRACE
	print_trace(FLP TR_UNLOCK,TR_START,(unsigned long)lck->lck);
#endif
	START_BLOCK_CLEAR_SIGNAL
	/* Should be reserve STICKS, but we do not have the correct data structure here ! */
		LM_LOCK(lck->lck->l.hold);
		if(p_unlock_one(FL lck)) adjust_marks(lck->lck);
		LM_UNLOCK(lck->lck->l.hold);
	END_BLOCK_CLEAR_SIGNAL;
	PRSTAT("multi_unlock","done",1);
#ifdef PSATHER_RT_TRACE
	print_trace(FLP TR_UNLOCK,TR_STOP,(unsigned long)lck->lck);
#endif
}

void p_endlock(FLA struct EXCEPT_ELEMENT *nn)
{
	int i;
#ifdef PSATHER_RT_TRACE
	prlocks(FLP TR_ENDLOCK,TR_START,nn);
#endif
	DEFER;
	PRSTAT("multi_endlock","start",nn->nlocks);
	START_BLOCK_CLEAR_SIGNAL
		if(!(nn->flags&LOCK_RESERVE)) RESERVE_STICKS;
		for(i=0;i<nn->nlocks;i++) {
			if(p_endlock_one(FLP &nn->locks[i])) {
				adjust_marks(nn->locks[i].lck);
			}
		}
		GIVEUP_STICKS;
	END_BLOCK_CLEAR_SIGNAL;
	PRSTAT("multi_endlock","done",nn->nlocks);
#ifdef PSATHER_RT_TRACE
	prlocks(FLP TR_ENDLOCK,TR_STOP,nn);
#endif
}

#ifdef PSATHER_STATISTICS
char *lock_type_str(char *w,enum LOCK_TYPES type)
{
	static char ret[30];
	strcpy(ret,w);strcat(ret," ");
#define LTP(n) case n: strcat(ret,#n);break
	switch(type) {
	LTP(L_MUTEX);
	LTP(L_NO_THREADS);
	LTP(L_THREADS);
	LTP(L_EMPTY);
	LTP(L_NOT_EMPTY);
	default: return "LOCK_TYPE_????";
	}
	return ret;
}
#endif
DLOCK *p_get_lock(FLA LOCK_T lck,enum LOCK_TYPES type)
{
	DLOCK* n;
	n=malloc(sizeof(struct DLOCK_struct));
	n->lck=lck;
	n->header.tag=DLOCK_TAG;
	n->type=type;
	n->count=1;
	return n;
}
#include <setjmp.h>


void *p_clear_exception;

void p_ex_set_clear_exception(FLA void *e)
{
	PRSTAT("exc","set_clear_exc",0);
	p_clear_exception=e;
}

void *p_ex_exception(FLAO)
{
	if(EXCEPTION_STACK==NULL) {
		PRSTAT("exc","reading NULL",0);
		return(NULL);
	} else {
		PRSTAT("exc","reading",EXCEPTION_STACK->id);
		return(EXCEPTION_STACK->exception);
	}
}

void p_ex_init_thread(FLAO)
{
	PRSTAT("exc","init",0);
	EXCEPTION_STACK=NULL;
}

struct EXCEPT_ELEMENT *p_ex_save_top(FLAO)
{	
	PRSTAT("exc","save_top",EXCEPTION_STACK->id);
	return (struct EXCEPT_ELEMENT *)EXCEPTION_STACK;
}

void p_ex_unwind(FLA short mark)
{
	struct EXCEPT_ELEMENT *n;
	n=EXCEPTION_STACK;
	while((n->mark&mark)==0) {
		PRSTAT("exc","unwind",EXCEPTION_STACK->id);
		p_endlock(FLP n);
		n=n->next;
		assert(n!=NULL);
	}
	p_ex_remove_mark(FL mark);
	EXCEPTION_STACK=n;
}

void p_ex_push(FLA struct EXCEPT_ELEMENT *n)
{
	n->next=EXCEPTION_STACK;
	n->mark=0;
	EXCEPTION_STACK=n;
#ifdef PSATHER_STATISTICS
	if(n->next==NULL) n->id=1;
	else n->id=n->next->id++;
#endif
	PRSTAT("exc","push",EXCEPTION_STACK->id);
}

void p_ex_pop(FLA int num)
{
	struct EXCEPT_ELEMENT *n=EXCEPTION_STACK;
	while(num-->0) {
		PRSTAT("exc","pop",EXCEPTION_STACK->id);
		p_endlock(FLP n);
		n=n->next;
	}
	EXCEPTION_STACK=n;
}

void p_ex_raise(FLA void *ex)
{
	struct EXCEPT_ELEMENT *n;
	if(EXCEPTION_STACK==NULL) {
		PRSTAT("exc","raise NULL",0);
		return;
	}
	PRSTAT("exc","raise",EXCEPTION_STACK->id);
	n=EXCEPTION_STACK;
	p_endlock(FLP n);
	EXCEPTION_STACK=n->next;
	if(n->next!=NULL) EXCEPTION_STACK->exception=ex;
	longjmp(n->jmp,1);
}

int p_ex_cmplcks(DLOCK *a,DLOCK *b)
{
	if((a)->lck<(b)->lck) return -1;
	else if((a)->lck>(b)->lck) return 1;
	else return 0;
}

void p_ex_set_mark(FLA short mark)
{
	EXCEPTION_STACK->mark|=mark;
}

void p_ex_remove_mark(FLA short mark)
{
	EXCEPTION_STACK->mark&= ~mark;
}


/* 
 * Copyright (C) International Computer Science Institute, 1994.  COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the SATHER LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/License" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email comments to "sather-bugs@icsi.berkeley.edu". 
 */
#include <stddef.h>
#include <malloc.h>

void ls_push_f(LONG_STACK *s,long ll)							\
{
	LONG_STACK n;
	n=(LONG_STACK)malloc(sizeof(struct LONG_STACK_struct));
	n->next= *s;
	n->l=ll;
	*s=n;
}

long ls_pop_f(LONG_STACK *s)
{
	long l;
	LONG_STACK o;
	if(*s==NULL) return(0);
	o=*s;
	*s=o->next;
	l=o->l;
	return l;
}

void ls_clear_f(LONG_STACK *s)
{
	while(*s!=NULL) ls_pop_f(s);
}



void tst_new_thread_f(THR_LONG_STACK *thr,THREAD_ID id)
{
	THR_LONG_STACK n;
	n=(THR_LONG_STACK)malloc(sizeof(struct THR_LONG_STACK_struct));
	n->next= *thr;
	n->id=id;
	n->unlocked=ls_create();
	*thr=n;
}

void tst_del_thread_f(THR_LONG_STACK *thr,THREAD_ID id)
{
	THR_LONG_STACK o;
	if((*thr)==NULL) return;
	else if((*thr)->id==id) {
		o= *thr;
		*thr=o->next;
		ls_clear(o->unlocked);
	} else
		tst_del_thread((*thr)->next,id);
}

LONG_STACK tst_find_thread_stack_f(THR_LONG_STACK *thr,THREAD_ID id)
{
	if((*thr)==NULL) return(NULL);
	else if((*thr)->id==id) return((*thr)->unlocked);
	else return tst_find_thread_stack((*thr)->next,id);
}

void tst_clear_f(THR_LONG_STACK *thr)
{
	while((*thr)!=NULL) tst_del_thread(*thr,(*thr)->id);
}

