/* 
 * Copyright (C) International Computer Science Institute, 1994.  COPYRIGHT
 * NOTICE: This code is provided "AS IS" WITHOUT ANY WARRANTY and is subject
 * to the terms of the SATHER LIBRARY GENERAL PUBLIC LICENSE contained in
 * the file "Doc/License" of the Sather distribution.  The license is also
 * available from ICSI, 1947 Center St., Suite 600, Berkeley CA 94704, USA.
 * ---> Please email comments to "sather-bugs@icsi.berkeley.edu". 
 */
#ifndef _PSATHER_H_
#define _PSATHER_H_

#include <stddef.h>
#include <stdlib.h>
#include <setjmp.h>

/*
 * PSATHER is only defined when the cs compiler compiles this
 * file. 
 */
#ifndef PSATHER 
#define PSATHER    
typedef void *OB;
typedef struct { int tag; } OB_HEADER;
#endif

/*
 * To get good error messages, you need to define DEBUG and PSATHER_CHK
 * when compiling your program and the psather library.
 * FORBID FLO
 * FORBID FL
 * FORBID FLA
 * FORBID FLAO
 * FORBID FLP
 * FORBID FLPO
 */
#ifdef DEBUG
#define FLO	__FILE__,__LINE__
#define FL	FLO,
#define FLAO	char*file,long line
#define FLA	FLAO,
#define FLPO	file,line
#define FLP	FLPO,
#define FLDO	"?",0
#define FLD	FLDO,
#define RFATAL(s) do { fprintf(stderr,"%s:%ld:%s\n",__FILE__,(long)__LINE__,s);abort(); } while(0)
#define RFATALexp(s) (fprintf(stderr,"%s:%ld:%s\n",__FILE__,(long)__LINE__,s),abort(),0)
#define RFATALp(s) do { fprintf(stderr,"%s:%ld:%s\n",file,(long)line,s);abort(); } while(0)
#else
#define FLO
#define FL
#define FLAO	void
#define FLA
#define FLPO	
#define FLP	
#define FLDO
#define FLD
#define RFATAL(s) do { fprintf(stderr,"%s\n",s);abort(); } while(0)
#define RFATALexp(s) (fprintf(stderr,"%s\n",s),abort(),0)
#define RFATALp(s) do { fprintf(stderr,"%s\n",s);abort(); } while(0)
#endif

/*
 * HERE should return the cluster id (an integer).
 * For one cluster machines like solaris threads it defaults
 * to 0
 */
#define HERE		0
#define ANY		0

/*
 * CLUSTERS defines the number of clusters for the current
 * run. This must be a constant during run time, but for
 * each run it may have a different value.
 */
#define CLUSTERS	1

/*
 * PSATHER_START: must be called before any of the functions defined
 *               in this header files are used.
 * PSATHER_STOP:  should be called before ending the program.
 */
#define PSATHER_START(progname)	psather_start(FL progname)
#define PSATHER_STOP		psather_stop(FLO)

/*
 * the SYS class
 */
#define SYS_CLEARED_SET(value)		p_cleared_set(FL value)
#define SYS_CLEARED_GET			p_cleared_get(FLO)

#define SYS_TRAP_CLEAR_SET(value)	p_trap_clear_set(FL value)
#define SYS_TRAP_CLEAR_GET		p_trap_clear_get(FLO)

#define SYS_DEFER			p_defer(FLO)

#define SYS_IMPORT
#define SYS_EXPORT

/*
 * VASSIGN guarantees atomicity for value type assignment
 */
#define VASSIGN(a,b) do { p_lock_v(&a);a=b;p_unlock_v(&a); } while(0)

/*
 * ATTACH, uses four arguments
 * function: the function to execute
 * gate:     each thread is attached to a gate, and here you pass the
 *           gate for the new process.
 * argument: a structure passed as argument to function.
 * position: the cluster on which this function should be executed.
 */
#define ATTACH(func,self,arg,gate,pos) p_attach(FL (FORK_FUNC)func,self,(void *)arg,sizeof(*arg),((GATE)gate)->lck,pos)

/*
 * LOCKS, MUTEX and GATES
 * First, we define the type of the different locks. 
 * Note that each lock type must use one bit. These bits are | together,
 *
 * Note: the pSather Compiler should not generate identifiers with the
 *       following names. Forbid is used to extract them automagically.
 *	 FORBID LOCK_TYPES
 *	 FORBID L_IGNORE
 *	 FORBID L_MUTEX
 *	 FORBID L_NO_THREADS
 *	 FORBID L_THREADS
 *	 FORBID L_EMPTY
 *	 FORBID L_NOT_EMPTY
 *	 FORBID DLOCK
 *	 FORBID LOCK
 *	 FORBID dLOCK
 *	 FORBID DLOCK_struct
 */
enum LOCK_TYPES { 
	L_IGNORE=0x8000, 
	L_MUTEX=1,
	L_NO_THREADS=2,
	L_THREADS=4,
	L_EMPTY=8,
	L_NOT_EMPTY=16
};
typedef struct DLOCK_struct { 
	OB_HEADER header;
	struct LOCK_T_struct *lck; 
	enum LOCK_TYPES type; 
	char count; 
} DLOCK;

typedef DLOCK *LOCK; /* anachronism */
typedef DLOCK *dLOCK;
typedef DLOCK *GATE;
typedef DLOCK *GATEV;
typedef DLOCK *GATET;
typedef DLOCK *MUTEX;

/* 
 * MUTEX
 */
#define MUTEX_CREATE			p_get_lock(FL (struct LOCK_T_struct *)p_mutex_create(FLO),L_MUTEX)

/*
 * Gates
 */
#define GATE_CREATE			(GATE)p_get_lock(FL p_gate_create(FL 0),L_MUTEX)
#define GATET_CREATE			(GATET)p_get_lock(FL p_gate_create(FL 4),L_MUTEX)
#define GATEV_CREATE(V)			(GATEV)p_get_lock(FL p_gate_create(FL sizeof(V)),L_MUTEX)
 
#define GATE_SIZE(gate)			CHKGATEi(gate,p_gate_size(FL (struct GATE_T_struct *)((GATE)gate)->lck))
#define GATET_SIZE(gate)		CHKGATEi(gate,p_gate_size(FL (struct GATE_T_struct *)((GATET)gate)->lck))
#define GATEV_SIZE(gate)		CHKGATEi(gate,p_gate_size(FL (struct GATE_T_struct *)((GATEV)gate)->lck))
 
#define GATE_HAS_THREADS(gate)		CHKGATEi(gate,p_gate_has_threads(FL (struct GATE_T_struct *)((GATE)gate)->lck))
#define GATET_HAS_THREADS(gate)		CHKGATEi(gate,p_gate_has_threads(FL (struct GATE_T_struct *)((GATET)gate)->lck))
#define GATEV_HAS_THREADS(gate)		CHKGATEi(gate,p_gate_has_threads(FL (struct GATE_T_struct *)((GATEV)gate)->lck))
 
#define GATE_SET(gate)			CHKGATEv(gate,p_gate_set(FL (struct GATE_T_struct *)((GATE)gate)->lck,0))
#define GATET_SET(gate,pointer)		CHKGATEv(gate,p_gate_set(FL (struct GATE_T_struct *)((GATET)gate)->lck,pointer,0))
#define GATEV_SET(gate,str)		CHKGATEv(gate,p_gate_set(FL (struct GATE_T_struct *)((GATEV)gate)->lck,&str,sizeof(str)))
 
#define GATE_GET(gate)			CHKGATEp(gate,p_gate_get(FL (struct GATE_T_struct *)((GATE)gate)->lck,NULL))
#define GATET_GET(gate,pointer)		CHKGATEp(gate,pointer=p_gate_get(FL (struct GATE_T_struct *)((GATET)gate)->lck,NULL))
#define GATEV_GET(gate,str)		CHKGATEp(gate,p_gate_get(FL (struct GATE_T_struct *)((GATEV)gate)->lck,&str))

#define GATE_ENQUEUE(gate)		CHKGATEv(gate,p_gate_enqueue(FL (struct GATE_T_struct *)((GATE)gate)->lck))
#define GATET_ENQUEUE(gate,pointer)	CHKGATEv(gate,p_gate_enqueue(FL (struct GATE_T_struct *)((GATET)gate)->lck,pointer,0))
#define GATEV_ENQUEUE(gate,str)		CHKGATEv(gate,p_gate_enqueue(FL (struct GATE_T_struct *)((GATEV)gate)->lck,&str,sizeof(str)))

#define GATE_DEQUEUE(gate)		CHKGATEp(gate,p_gate_dequeue(FL (struct GATE_T_struct *)((GATE)gate)->lck,NULL))
#define GATET_DEQUEUE(gate,pointer)	CHKGATEp(gate,pointer=(void*)p_gate_dequeue(FL (struct GATE_T_struct *)((GATET)gate)->lck,NULL))
#define GATEV_DEQUEUE(gate,str) 	CHKGATEp(gate,p_gate_dequeue(FL (struct GATE_T_struct *)((GATEV)gate)->lck,&str))

#define GATE_CLEAR(gate)		CHKGATEv(gate,p_gate_clear(FL (struct GATE_T_struct *)((GATE)gate)->lck))
#define GATET_CLEAR(gate)		CHKGATEv(gate,p_gate_clear(FL (struct GATE_T_struct *)((GATET)gate)->lck))
#define GATEV_CLEAR(gate)		CHKGATEv(gate,p_gate_clear(FL (struct GATE_T_struct *)((GATEV)gate)->lck))

#define GATE_SYNC(gate)			CHKGATEv(gate,p_gate_sync(FL (struct GATE_T_struct *)((GATE)gate)->lck))
#define GATET_SYNC(gate)		CHKGATEv(gate,p_gate_sync(FL (struct GATE_T_struct *)((GATET)gate)->lck))
#define GATEV_SYNC(gate)		CHKGATEv(gate,p_gate_sync(FL (struct GATE_T_struct *)((GATEV)gate)->lck))

#define GATE_MUTEX(n)			CHKGATEp(n,p_get_lock(FL ((GATE)n)->lck,L_MUTEX))
#define GATET_MUTEX(n)			CHKGATEp(n,p_get_lock(FL ((GATET)n)->lck,L_MUTEX))
#define GATEV_MUTEX(n)			CHKGATEp(n,p_get_lock(FL ((GATEV)n)->lck,L_MUTEX))

#define GATE_EMPTY(n)			CHKGATEp(n,p_get_lock(FL ((GATE)n)->lck,L_EMPTY))
#define GATET_EMPTY(n)			CHKGATEp(n,p_get_lock(FL ((GATET)n)->lck,L_EMPTY))
#define GATEV_EMPTY(n)			CHKGATEp(n,p_get_lock(FL ((GATEV)n)->lck,L_EMPTY))

#define GATE_NOT_EMPTY(n)		CHKGATEp(n,p_get_lock(FL ((GATE)n)->lck,L_NOT_EMPTY))
#define GATET_NOT_EMPTY(n)		CHKGATEp(n,p_get_lock(FL ((GATET)n)->lck,L_NOT_EMPTY))
#define GATEV_NOT_EMPTY(n)		CHKGATEp(n,p_get_lock(FL ((GATEV)n)->lck,L_NOT_EMPTY))

#define GATE_THREADS(n)			CHKGATEp(n,p_get_lock(FL ((GATE)n)->lck,L_THREADS))
#define GATET_THREADS(n)		CHKGATEp(n,p_get_lock(FL ((GATET)n)->lck,L_THREADS))
#define GATEV_THREADS(n)		CHKGATEp(n,p_get_lock(FL ((GATEV)n)->lck,L_THREADS))

#define GATE_NO_THREADS(n)		CHKGATEp(n,p_get_lock(FL ((GATE)n)->lck,L_NO_THREADS))
#define GATET_NO_THREADS(n)		CHKGATEp(n,p_get_lock(FL ((GATET)n)->lck,L_NO_THREADS))
#define GATEV_NO_THREADS(n)		CHKGATEp(n,p_get_lock(FL ((GATEV)n)->lck,L_NO_THREADS))

/*
 * WARNING: you can get the gate to which a thread is attached, BUT
 *          you must know what kind of gate this is.
 * There is currently no way in pSather to get this gate, as the
 * cohort is implemented by directly using the specific gate.
 */
#define GATE_COHORT			(p_get_lock(FL p_cohort(FLO),L_MUTEX))
#define GATET_COHORT			(p_get_lock(FL p_cohort(FLO),L_MUTEX))
#define GATEV_COHORT			(p_get_lock(FL p_cohort(FLO),L_MUTEX))

/*
 * And now the protect, lock, try and unlock
 * a maximum of 5 arguments are currently supported for lock and try
 */

/*
 * the structure EXCEPT_ELEMENT should not be used anywhere in the code
 * FORBID EXCEPT_ELEMENT
 * FORBID EXCEPT_ELEMENT0
 * FORBID EXCEPT_ELEMENT1
 * FORBID EXCEPT_ELEMENT2
 * FORBID EXCEPT_ELEMENT3
 * FORBID EXCEPT_ELEMENT4
 * FORBID EXCEPT_ELEMENT5
 */
struct EXCEPT_ELEMENT {
        struct EXCEPT_ELEMENT *next;
        jmp_buf jmp;
        OB exception;
#ifdef DEBUG
	long id;
#endif
	short flags;
	short mark;
        short nlocks;
        DLOCK locks[1];
};
#ifdef DEBUG
#define EXCEPT_ELEMENTn(n) struct EXCEPT_ELEMENT ## n { \
        struct EXCEPT_ELEMENT *next;			\
        jmp_buf jmp;					\
        OB exception;					\
	long id;					\
	short flags;					\
	short mark;					\
        short nlocks;					\
        DLOCK locks[n];					\
}
#else
#define EXCEPT_ELEMENTn(n) struct EXCEPT_ELEMENT ## n { \
        struct EXCEPT_ELEMENT *next;			\
        jmp_buf jmp;					\
        OB exception;					\
	short flags;					\
	short mark;					\
        short nlocks;					\
        DLOCK locks[n];					\
}
#endif
EXCEPT_ELEMENTn(1);
EXCEPT_ELEMENTn(2);
EXCEPT_ELEMENTn(3);
EXCEPT_ELEMENTn(4);
EXCEPT_ELEMENTn(5);
#define LOCK_RESERVE 1

/*
 * While executing a lock, try or protect statement, an exception stack
 * is build. You can mark specific stack frames and unwind the stack to the
 * first mark at any time. You have 16 different mark (0-15), but MARK(0)
 * and MARK(1) are predefined. The first is used to mark the stack at the
 * beginning of a function, while the second can be used for loops.
 * Be warned however, that while ESCAPE unwinds the stack end ends the
 * function, JUMP_OUT_OF_LOOP does NOT terminate the loop. You have to 
 * manually end the loop by either a break or a goto.
 */
#define MARK(n)	(1<<(n))
#define MARK_FUNCTION			MARK(0)
#define MARK_LOOP			MARK(1)
#define SET_EXCEPTION_MARK(n)		p_ex_set_mark(FL n)
#define REMOVE_EXCEPTION_MARK(n)	p_ex_remove_mark(FL n)
#define POP_TO(m)			p_ex_unwind(FL (m))

/*
 * each function using return MUST use EXCEPTION_PROLOGUE at the 
 * beginning, and EXCEPTION_EPILOGUE at the end. Instead of using
 * return, it MUST use ESCAPE.
 */
#define EXCEPTION_PROLOGUE 	SET_EXCEPTION_MARK(MARK_FUNCTION);
#define EXCEPTION_EPILOGUE 	REMOVE_EXCEPTION_MARK(MARK_FUNCTION);
#define ESCAPE			POP_TO(MARK_FUNCTION);return 

/*
 * If you want to use mark on the stack to end loops, use
 * LOOP_BEGIN before the loop and LOOP_END at its end.
 * Note that JUMP_OUT_OF_LOOP must be followed by a break or goto,
 * as it will not end the loop.
 */
#define LOOP_BEGIN		SET_EXCEPTION_MARK(MARK_LOOP);
#define LOOP_END		REMOVE_EXCEPTION_MARK(MARK_LOOP);
#define JUMP_OUT_OF_LOOP	POP_TO(MARK_LOOP)

/*
 * To pop some frames from the exception stack, you may use POP_EXCPETION.
 * This macro can be used instead of LOOP_BEGIN and LOOP_END. The only
 * problem is that you must know exactly how many frames to pop
 * when exiting a loop.
 */
#define POP_EXCEPTION1		p_ex_pop(FL 1)
#define POP_EXCEPTION(n)	p_ex_pop(FL (n))

/* 
 * Before using exception, you must define the clear exception with
 * SET_CLEAR_EXCEPTION(p). This macro accepts a pointer. This is the
 * only exception the runtime may raise, and if this happens,
 * EXCEPTIONS is set to the value defined by SET_CLEAR_EXCEPTION.
 */
#define SET_CLEAR_EXCEPTION(p)		p_ex_set_clear_exception(FL (OB)(p))

/*
 * RAISE(x) raises an exception. The value x (actually a pointer) is passed
 * to the exception handling routine, which can use it through the
 * macro EXCEPTION. If this routine needs to propagate the exception,
 * it can use RAISE(EXCEPTION)
 */
#define RAISE(x)			p_ex_raise(FL (x))
#define EXCEPTION 	   		((OB)p_ex_exception(FLO))

/* 
 * To protect some statements A against exception, use
 * PROTECT_BEGIN
 *	A;
 * PRTOECT_WHEN
 *	if(EXCPETION==....
 * PRTOECT_END
 *
 * FORBID p_n
 */
#define PROTECT_BEGIN	   { struct EXCEPT_ELEMENT p_n;			     \
			     p_n.flags=0;				     \
			     p_n.mark=0;				     \
			     p_n.nlocks=0;				     \
			     if(setjmp(p_n.jmp)==0) {		  	     \
				p_ex_push(FL &p_n);
             
#define PROTECT_WHEN	     p_ex_pop(FL 1); } else {
#define PROTECT_END	     }		   \
			   }

/*
 * PROTECT_R_* is for use in the runtime only
 */
#define PROTECT_R_BEGIN	   { struct EXCEPT_ELEMENT p_n;			     \
			     p_n.flags=0;				     \
			     p_n.mark=0;				     \
			     p_n.nlocks=0;				     \
			     if(setjmp(p_n.jmp)==0) {		  	     \
				p_ex_push(FLP &p_n);
             
#define PROTECT_R_WHEN	     p_ex_pop(FLP 1); } else {
#define PROTECT_R_END	     }		   \
			   }
/*
 * Multiple locks.
 * To lock n locks, use the lock macros as follows:
 *
 * LOCKn(l1, ... ln)
 *     statements;
 * LOCKEND
 * (ENDLOCK is equivalent to LOCKEND, but may not be supported in future
 * versions of this library)
 * WARNING: if PSATHER_CHK is defined arguments to LOCKn() and TRYn() are 
 * evaluated twice, so they may not have any side effects.
 * Example: LOCK1(g++) is an error!
 */
#ifdef PSATHER_CHK
#define CHKLOCK(lock,d,n)	if((lock)==NULL) RFATAL("locking a void lock (arg " #n); else d
#define CHKLOCKR(lock,d,n)	if((lock)==NULL) RFATAL("locking a void lock (arg " #n); else d
#define CHKTRY(lock,d,n)	if((lock)==NULL) RFATAL("trying a void lock (arg " #n); else d
#define CHKUNLOCK(lock)		if((lock)==NULL) RFATAL("unlocking a void lock");
#define CHKGATEp(gate,d)	((gate)==NULL?(void *)RFATALexp("using an uninitialized gate"):(d))
#define CHKGATEi(gate,d)	((gate)==NULL?RFATALexp("using an uninitialized gate"):(d))
#define CHKGATEv(gate,d)	((gate)==NULL?RFATALexp("using an uninitialized gate"):(d),0)
#else /* PSATHER_CHK */
#define CHKLOCK(lock,d,n) 	d
#define CHKLOCKR(lock,d,n) 	d
#define CHKTRY(lock,d,n) 	d
#define CHKUNLOCK(lock) 	
#define CHKGATEp(gate,d)	d
#define CHKGATEi(gate,d)	d
#define CHKGATEv(gate,d)	d
#endif /* PSATHER_CHK */
#define LOCK1(l1)	   { struct EXCEPT_ELEMENT1 p_n; 		    \
			     p_n.flags=0;				    \
			     p_n.mark=0;				    \
			     p_n.nlocks=1;				    \
			     CHKLOCK(l1,p_n.locks[0]= *(DLOCK *)l1,1);	    \
			     if(setjmp(p_n.jmp)==0) {			    \
			    	p_lock(FL (struct EXCEPT_ELEMENT *)&p_n); 
#define LOCK2(l1,l2)	   { struct EXCEPT_ELEMENT2 p_n;		    \
			     p_n.flags=0;				    \
			     p_n.mark=0;				    \
			     p_n.nlocks=2;			            \
			     CHKLOCK(l1,p_n.locks[0]= *(DLOCK *)l1,1);	    \
			     CHKLOCK(l2,p_n.locks[1]= *(DLOCK *)l2,2);	    \
			     qsort((char *)p_n.locks,2,sizeof(DLOCK),(int(*)())p_ex_cmplcks);\
			     if(setjmp(p_n.jmp)==0) {			    \
			    	p_lock(FL (struct EXCEPT_ELEMENT *)&p_n); 
#define LOCK3(l1,l2,l3)	   { struct EXCEPT_ELEMENT3 p_n;		    \
			     p_n.flags=0;				    \
			     p_n.mark=0;				    \
			     p_n.nlocks=3;			            \
			     CHKLOCK(l1,p_n.locks[0]= *(DLOCK *)l1,1);	    \
			     CHKLOCK(l2,p_n.locks[1]= *(DLOCK *)l2,2);	    \
			     CHKLOCK(l3,p_n.locks[2]= *(DLOCK *)l3,3);	    \
			     qsort((char *)p_n.locks,3,sizeof(DLOCK),(int(*)())p_ex_cmplcks);\
			     if(setjmp(p_n.jmp)==0) {			    \
			    	p_lock(FL (struct EXCEPT_ELEMENT *)&p_n); 
#define LOCK4(l1,l2,l3,l4)   { struct EXCEPT_ELEMENT4 p_n; 		    \
			     p_n.flags=0;				    \
			     p_n.mark=0;				    \
			     p_n.nlocks=4;			            \
			     CHKLOCK(l1,p_n.locks[0]= *(DLOCK *)l1,1);	    \
			     CHKLOCK(l2,p_n.locks[1]= *(DLOCK *)l2,2);	    \
			     CHKLOCK(l3,p_n.locks[2]= *(DLOCK *)l3,3);	    \
			     CHKLOCK(l4,p_n.locks[3]= *(DLOCK *)l4,4);	    \
			     qsort((char *)p_n.locks,4,sizeof(DLOCK),(int(*)())p_ex_cmplcks);\
			     if(setjmp(p_n.jmp)==0) {			    \
			    	p_lock(FL (struct EXCEPT_ELEMENT *)&p_n); 
#define LOCK5(l1,l2,l3,l4,l5){ struct EXCEPT_ELEMENT5 p_n;		    \
			     p_n.flags=0;				    \
			     p_n.mark=0;				    \
			     p_n.nlocks=5;			            \
			     CHKLOCK(l1,p_n.locks[0]= *(DLOCK *)l1,1);	    \
			     CHKLOCK(l2,p_n.locks[1]= *(DLOCK *)l2,2);	    \
			     CHKLOCK(l3,p_n.locks[2]= *(DLOCK *)l3,3);	    \
			     CHKLOCK(l4,p_n.locks[3]= *(DLOCK *)l4,4);	    \
			     CHKLOCK(l5,p_n.locks[4]= *(DLOCK *)l5,5);	    \
			     qsort((char *)p_n.locks,5,sizeof(DLOCK),(int(*)())p_ex_cmplcks);\
			     if(setjmp(p_n.jmp)==0) {			    \
			    	p_lock(FL (struct EXCEPT_ELEMENT *)&p_n); 
				
#define LOCKEND		        p_ex_pop(FL 1); 			\
			     } else {					\
				p_ex_raise(FL EXCEPTION);		\
			     }						\
			   }
#define ENDLOCK LOCKEND

/* 
 * LOCK_RESERVE is only meant for the runtime and should
 * never be called by code generated by the compiler.
 */
#define LOCK_RESERVE1(l1)   { struct EXCEPT_ELEMENT1 p_n;		    \
			     p_n.flags=LOCK_RESERVE;			    \
			     p_n.nlocks=1;				    \
			     p_n.mark=0;				    \
			     CHKLOCKR(l1,p_n.locks[0]= *(DLOCK *)l1,1);	    \
			     if(setjmp(p_n.jmp)==0) {			    \
			    	p_lock(FLP (struct EXCEPT_ELEMENT *)&p_n); 
				
#define LOCK_R_END		p_ex_pop(FLP 1); 			\
			     } else {					\
				p_ex_raise(FLP EXCEPTION);		\
			     }						\
			   }
/*
 * If you need to try on n locks, do it as follows:
 * TRYn(l1,... ln)
 *      statements_if_try_succeeds
 * TRYELSE
 *      statements_if_try_fails
 * TRYEND
 */
#define TRY1(l1)	   { struct EXCEPT_ELEMENT1 p_n;		\
			     p_n.flags=0;				\
			     p_n.mark=0;				\
			     p_n.nlocks=1;			        \
			     CHKTRY(l1,p_n.locks[0]= *(DLOCK *)l1,1);	\
			     if(setjmp(p_n.jmp)==0) {			\
			    	if(p_try(FL (struct EXCEPT_ELEMENT *)&p_n)) {
#define TRY2(l1,l2)	   { struct EXCEPT_ELEMENT2 p_n;		\
			     p_n.flags=0;				\
			     p_n.mark=0;				\
			     p_n.nlocks=2;			        \
			     CHKTRY(l1,p_n.locks[0]= *(DLOCK *)l1,1);	\
			     CHKTRY(l2,p_n.locks[1]= *(DLOCK *)l2,2);	\
			     qsort((char *)p_n.locks,2,sizeof(DLOCK),(int(*)())p_ex_cmplcks);\
			     if(setjmp(p_n.jmp)==0) {			\
			    	if(p_try(FL (struct EXCEPT_ELEMENT *)&p_n)) {
#define TRY3(l1,l2,l3)	   { struct EXCEPT_ELEMENT3 p_n;		\
			     p_n.flags=0;				\
			     p_n.mark=0;				\
			     p_n.nlocks=3;			        \
			     CHKTRY(l1,p_n.locks[0]= *(DLOCK *)l1,1);	\
			     CHKTRY(l2,p_n.locks[1]= *(DLOCK *)l2,2);	\
			     CHKTRY(l3,p_n.locks[2]= *(DLOCK *)l3,3);	\
			     qsort((char *)p_n.locks,3,sizeof(DLOCK),(int(*)())p_ex_cmplcks);\
			     if(setjmp(p_n.jmp)==0) {			\
			    	if(p_try(FL (struct EXCEPT_ELEMENT *)&p_n)) {
#define TRY4(l1,l2,l3,l4)  { struct EXCEPT_ELEMENT4 p_n;		\
			     p_n.flags=0;				\
			     p_n.mark=0;				\
			     p_n.nlocks=4;			        \
			     CHKTRY(l1,p_n.locks[0]= *(DLOCK *)l1,1);	\
			     CHKTRY(l2,p_n.locks[1]= *(DLOCK *)l2,2);	\
			     CHKTRY(l3,p_n.locks[2]= *(DLOCK *)l3,3);	\
			     CHKTRY(l4,p_n.locks[3]= *(DLOCK *)l4,4);	\
			     qsort((char *)p_n.locks,4,sizeof(DLOCK),(int(*)())p_ex_cmplcks);\
			     if(setjmp(p_n.jmp)==0) {			\
			    	if(p_try(FL (struct EXCEPT_ELEMENT *)&p_n)) {
#define TRY5(l1,l2,l3,l4,l5){ struct EXCEPT_ELEMENT5 p_n;		\
			     p_n.flags=0;				\
			     p_n.mark=0;				\
			     p_n.nlocks=5;			        \
			     CHKTRY(l1,p_n.locks[0]= *(DLOCK *)l1,1);	\
			     CHKTRY(l2,p_n.locks[1]= *(DLOCK *)l2,2);	\
			     CHKTRY(l3,p_n.locks[2]= *(DLOCK *)l3,3);	\
			     CHKTRY(l4,p_n.locks[3]= *(DLOCK *)l4,5);	\
			     CHKTRY(l5,p_n.locks[4]= *(DLOCK *)l5,4);	\
			     qsort((char *)p_n.locks,5,sizeof(DLOCK),(int(*)())p_ex_cmplcks);\
			     if(setjmp(p_n.jmp)==0) {			\
			    	if(p_try(FL (struct EXCEPT_ELEMENT *)&p_n)) {
				
#define TRYELSE			p_ex_pop(FL 1); } else {
#define TRYEND		        } 					\
			     } else {					\
				p_ex_raise(FL EXCEPTION);		\
			     }						\
			   }

/*
 * You can unlock any lock acquired by either a LOCK or a TRY
 * statement before the LOCKEND (TRYEND)
 */
#define UNLOCK(n)	do { CHKUNLOCK(n);p_unlock(FL (DLOCK*)(n)); } while(0)


/*
 * The rest of this header file is filled with prototype declarations
 * FORBID FORK_FUNC
 * (the other identifier are catched by reading the .o files)
 */

typedef void (*FORK_FUNC)(OB self,void *arg,GATE gate,int pos);

void p_lock(FLA struct EXCEPT_ELEMENT *);
int p_try(FLA struct EXCEPT_ELEMENT *);
void p_endlock(FLA struct EXCEPT_ELEMENT *);
void p_unlock(FLA DLOCK *lck);

struct EXCEPT_ELEMENT *p_ex_save_top(FLAO);
void p_ex_push(FLA struct EXCEPT_ELEMENT *);
void p_ex_pop(FLA int);
void p_ex_unwind(FLA short);
void p_ex_set_mark(FLA short);
void p_ex_remove_mark(FLA short);
DLOCK * p_get_lock(FLA struct LOCK_T_struct *lck,enum LOCK_TYPES type);

int p_ex_cmplcks(DLOCK *a,DLOCK *b);

void psather_start(FLA char *prog_name);
void psather_stop(FLAO);

void p_lock_v(void *);
void p_unlock_v(void *);

#ifndef  IN_PSATHER_LIBRARY
void p_ex_raise(FLA OB);
OB p_ex_exception(FLAO);

OB p_mutex_create(FLAO);

struct GATE_T_struct;
struct LOCK_T_struct *p_gate_create(FLA long size);
long p_gate_size(FLA struct GATE_T_struct *gate);
long p_gate_has_threads(FLA struct GATE_T_struct *gate);
void p_gate_set(FLA struct GATE_T_struct *gate,...);
OB p_gate_get(FLA struct GATE_T_struct *gate,...);
void p_gate_enqueue(FLA struct GATE_T_struct *gate,...);
OB p_gate_dequeue(FLA struct GATE_T_struct *gate,...);
void p_gate_clear(FLA struct GATE_T_struct *gate);
void p_gate_sync(FLA void *gate);


void p_attach(FLA FORK_FUNC function,OB self,void *arg,long size,void * gate,int position);

void *p_arg_alloc(FLA long size);
void p_arg_free(FLA char *ptr);
void *p_cohort(FLAO);

void p_cleared_set(FLA short value);
short p_cleared_get(FLAO);
void p_trap_clear_set(FLA short value);
short p_trap_clear_get(FLAO);
void p_defer(FLAO);

#endif

#endif
